
interface Module {
  /**
  * Allocates memory on the Emscripten heap.
  * @param size - The size of the memory to allocate.
  * @returns number - The pointer to the allocated memory.
  * @see https://emscripten.org/docs/api_reference/preamble.js.html#_malloc
  */
  _malloc(size: number): number;
  /**
  * Reads a value from the Emscripten heap.
  * @param ptr - The pointer to the value.
  * @param type - The type of the value.
  * @returns number - The value.
  * @see https://emscripten.org/docs/api_reference/preamble.js.html#getValue
  */
  getValue(ptr: number, type: string): number;
  /**
  * Writes a value to the Emscripten heap.
  * @param ptr - The pointer to the value.
  * @param type - The type of the value.
  * @param value - The value.
  * @returns void - undefined
  * @see https://emscripten.org/docs/api_reference/preamble.js.html#setValue
  */
  setValue(ptr: number, type: string, value: number): void;
  /**
  * Reads a null-terminated string from the Emscripten heap.
  * @param ptr - The pointer to the string.
  * @returns string - The string.
  * @see https://emscripten.org/docs/api_reference/preamble.js.html#UTF8ToString
  */
  UTF8ToString(ptr: number): string;
  /**
  * Writes a string to the Emscripten heap.
  * @param str - The string.
  * @param ptr - The pointer to the string.
  * @param maxLength - The maximum length of the string.
  * @returns void - undefined
  * @see https://emscripten.org/docs/api_reference/preamble.js.html#stringToUTF8
  */
  stringToUTF8(str: string, ptr: number, maxLength: number): void;
}

interface geos {
  /**
  * Sets the notice handler for a GEOS context.
  * @param context - A pointer to the GEOS context handle.
  * @param handler - The notice handler function pointer.
  * @returns number - The previous notice handler function pointer.
    */
  GEOSContext_setNoticeHandler_r(context: number, handler: number): number;
  /**
  * Sets the error handler for a GEOS context.
  * @param context - A pointer to the GEOS context handle.
  * @param handler - The error handler function pointer.
  * @returns number - The previous error handler function pointer.
    */
  GEOSContext_setErrorHandler_r(context: number, handler: number): number;
  /**
  * Registers a callback function to be called when an interrupt request is detected.
  * @param callback - The callback function pointer.
  * @returns void - undefined
    */
  GEOS_interruptRegisterCallback(callback: number): void;
  /**
  * Requests an interrupt of the current GEOS operation.
  * @returns void - undefined
    */
  GEOS_interruptRequest(): void;
  /**
  * Cancels an interrupt request of the current GEOS operation.
  * @returns void - undefined
    */
  GEOS_interruptCancel(): void;
  /**
  * Frees a GEOS object allocated by the library.
  * @param obj - The GEOS object pointer.
  * @returns void - undefined
    */
  GEOSFree(obj: number): void;
  /**
  * Frees a GEOS object allocated by the library in a given context.
  * @param context - A pointer to the GEOS context handle.
  * @param obj - The GEOS object pointer.
  * @returns void - undefined
    */
  GEOSFree_r(context: number, obj: number): void;
  /**
  * Tests if two geometries are disjoint.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSDisjoint(g1: number, g2: number): number;
  /**
  * Tests if two geometries are disjoint in a given context.
  * @param context - A pointer to the GEOS context handle.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSDisjoint_r(context: number, g1: number, g2: number): number;
  /**
  * Tests if two geometries touch at one or more points.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSTouches(g1: number, g2: number): number;
  /**
  * Tests if two geometries touch at one or more points in a given context.
  * @param context - A pointer to the GEOS context handle.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSTouches_r(context: number, g1: number, g2: number): number;
  /**
  * Tests if two geometries intersect in interior or boundary points.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSIntersects(g1: number, g2: number): number;
  /**
  * Tests if two geometries intersect in interior or boundary points in a given context.
  * @param context - A pointer to the GEOS context handle.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSIntersects_r(context: number, g1: number, g2: number): number;
  /**
  * Tests if two geometries cross.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSCrosses(g1: number, g2: number): number;
  /**
  * Tests if two geometries cross in a given context.
  * @param context - A pointer to the GEOS context handle.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSCrosses_r(context: number, g1: number, g2: number): number;
  /**
  * Tests if the first geometry is within the second geometry.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSWithin(g1: number, g2: number): number;
  /**
  * Tests if the first geometry is within the second geometry in a given context.
  * @param context - A pointer to the GEOS context handle.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSWithin_r(context: number, g1: number, g2: number): number;
  /**
  * Tests if the first geometry contains the second geometry.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSContains(g1: number, g2: number): number;
  /**
  * Tests if the first geometry contains the second geometry in a given context.
  * @param context - A pointer to the GEOS context handle.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSContains_r(context: number, g1: number, g2: number): number;
  /**
  * Tests if two geometries overlap in their interior points.
  * @param g1 - The first geometry pointer.
  * @param g2 - The second geometry pointer.
  * @returns number - = 1 if true, 0 if false, -1 on error.
    */
  GEOSOverlaps(g1: number, g2: number): number;
  /**
  * Tests whether two geometries overlap.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @returns number - 1 if the geometries overlap, 0 if not, or -1 on error.
    */
  GEOSOverlaps_r(handle: number, g1: number, g2: number): number;
  /**
  * Tests whether the first geometry covers the second geometry.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @returns number - 1 if the first geometry covers the second geometry, 0 if not, or -1 on error.
    */
  GEOSCovers(g1: number, g2: number): number;
  /**
  * Tests whether the first geometry covers the second geometry in a given context.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @returns number - 1 if the first geometry covers the second geometry, 0 if not, or -1 on error.
    */
  GEOSCovers_r(handle: number, g1: number, g2: number): number;
  /**
  * Tests whether the first geometry is covered by the second geometry.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @returns number - 1 if the first geometry is covered by the second geometry, 0 if not, or -1 on error.
    */
  GEOSCoveredBy(g1: number, g2: number): number;
  /**
  * Tests whether the first geometry is covered by the second geometry in a given context.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @returns number - 1 if the first geometry is covered by the second geometry, 0 if not, or -1 on error.
    */
  GEOSCoveredBy_r(handle: number, g1: number, g2: number): number;
  /**
  * Tests whether two geometries are related according to a given DE-9IM intersection matrix pattern.
  * @param g1 - A pointer to the first geometry
  * @param g2 - A pointer to the second geometry
  * @param pat - A string representing the DE-9IM pattern
  * @returns number - 1 if the geometries are related, 0 if not, 2 on exception
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3a6f4e9a7f0b8c8d5b4a0c5f6c7e4d3f
  */
  GEOSRelatePattern(g1: number, g2: number, pat: string): number;
  /**
  * Tests whether two geometries are related according to a given DE-9IM intersection matrix pattern in a thread-safe manner.
  * @param handle - A GEOS context handle
  * @param g1 - A pointer to the first geometry
  * @param g2 - A pointer to the second geometry
  * @param pat - A string representing the DE-9IM pattern
  * @returns number - 1 if the geometries are related, 0 if not, 2 on exception
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3a6f4e9a7f0b8c8d5b4a0c5f6c7e4d3f
  */
  GEOSRelatePattern_r(handle: number, g1: number, g2: number, pat: string): number;
  /**
  * Tests whether a DE-9IM intersection matrix matches a given pattern.
  * @param mat - A string representing the DE-9IM matrix
  * @param pat - A string representing the DE-9IM pattern
  * @returns number - 1 if the matrix matches the pattern, 0 if not, 2 on exception
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3a6f4e9a7f0b8c8d5b4a0c5f6c7e4d3f
  */
  GEOSRelatePatternMatch(mat: string, pat: string): number;
  /**
  * Tests whether a DE-9IM intersection matrix matches a given pattern in a thread-safe manner.
  * @param handle - A GEOS context handle
  * @param mat - A string representing the DE-9IM matrix
  * @param pat - A string representing the DE-9IM pattern
  * @returns number - 1 if the matrix matches the pattern, 0 if not, 2 on exception
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3a6f4e9a7f0b8c8d5b4a0c5f6c7e4d3f
  */
  GEOSRelatePatternMatch_r(handle: number, mat: string, pat: string): number;
  /**
  * Computes a DE-9IM intersection matrix for two geometries.
  * @param g1 - A pointer to the first geometry
  * @param g2 - A pointer to the second geometry
  * @returns string - A string representing the DE-9IM matrix, or NULL on exception
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3a6f4e9a7f0b8c8d5b4a0c5f6c7e4d3f
  */
  GEOSRelate(g1: number, g2: number): string;
  /**
  * Computes the DE-9IM intersection matrix for two geometries.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @returns string - The DE-9IM intersection matrix as a string, or null if an error occurred.
  * @see https://docs.rs/geos-sys/2.0.0/geos_sys/fn.GEOSRelate_r.html
  */
  GEOSRelate_r(handle: number, g1: number, g2: number): string;
  /**
  * Computes the DE-9IM intersection matrix for two geometries using a specified Boundary Node Rule.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @param bnr - The Boundary Node Rule to use. One of: 1 (Mod2), 2 (EndPoint), 3 (MultiValentEndPoint), 4 (MonoValentEndPoint).
  * @returns string - The DE-9IM intersection matrix as a string, or null if an error occurred.
  * @see https://docs.rs/geos-sys/2.0.1/geos_sys/fn.GEOSRelateBoundaryNodeRule.html
  */
  GEOSRelateBoundaryNodeRule(handle: number, g1: number, g2: number, bnr: number): string;
  /**
  * Computes the DE-9IM intersection matrix for two geometries using a specified Boundary Node Rule.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @param bnr - The Boundary Node Rule to use. One of: 1 (Mod2), 2 (EndPoint), 3 (MultiValentEndPoint), 4 (MonoValentEndPoint).
  * @returns string - The DE-9IM intersection matrix as a string, or null if an error occurred.
  * @see https://docs.rs/geos-sys/2.0.1/geos_sys/fn.GEOSRelateBoundaryNodeRule_r.html
  */
  GEOSRelateBoundaryNodeRule_r(handle: number, g1: number, g2: number, bnr: number): string;
  /**
  * Tests whether a geometry is valid according to the OGC rules for geometry validity.
  * @param geom - The geometry to test.
  * @returns number - 1 if the geometry is valid, 0 if it is invalid, or -1 if an error occurred.
  * @see https://docs.rs/geos-sys/1.0.10/geos_sys/fn.GEOSisValid.html
  */
  GEOSisValid(geom: number): number;
  /**
  * Tests whether a geometry is valid according to the OGC rules for geometry validity.
  * @param handle - A pointer to the GEOS context handle.
  * @param geom - The geometry to test.
  * @returns number - 1 if the geometry is valid, 0 if it is invalid, or -1 if an error occurred.
  * @see https://docs.rs/geos-sys/1.0.10/geos_sys/fn.GEOSisValid_r.html
  */
  GEOSisValid_r(handle: number, geom: number): number;
  /**
  * Returns text stating if a geometry is valid, or if invalid a reason why.
  * @param g - A pointer to a GEOSGeometry object
  * @returns string - A null-terminated string describing the validity of the geometry
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a0f3a7b9f9c6b5a1d0c2f6c8e3a4d4b5e|GEOSisValidReason}
  */
  GEOSisValidReason(g: number): string;
  /**
  * Returns text stating if a geometry is valid, or if invalid a reason why. Thread-safe version.
  * @param handle - A GEOS context handle
  * @param g - A pointer to a GEOSGeometry object
  * @returns string - A null-terminated string describing the validity of the geometry
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a0f3a7b9f9c6b5a1d0c2f6c8e3a4d4b5e|GEOSisValidReason_r}
  */
  GEOSisValidReason_r(handle: number, g: number): string;
  /**
  * Returns a valid_detail row, containing a boolean stating if a geometry is valid,
* a varchar stating a reason why it is invalid and a geometry pointing out where it is invalid.
  * @param g - A pointer to a GEOSGeometry object
  * @param flags - A bitfield for setting options. See {@link https://postgis.net/docs/ST_IsValidDetail.html|ST_IsValidDetail} for details.
  * @param reason - A pointer to store the null-terminated string describing the validity of the geometry
  * @param location - A pointer to store the GEOSGeometry object indicating the location of invalidity
  * @returns number - 1 if the geometry is valid, 0 if it is invalid, or -1 on error
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a1f7b2d9e0a7d2e9d7e6f4a7a2f3e5b9d|GEOSisValidDetail}
  */
  GEOSisValidDetail(g: number, flags: number, reason: string, location: number): number;
  /**
  * Returns a valid_detail row, containing a boolean stating if a geometry is valid,
* a varchar stating a reason why it is invalid and a geometry pointing out where it is invalid. Thread-safe version.
  * @param handle - A GEOS context handle
  * @param g - A pointer to a GEOSGeometry object
  * @param flags - A bitfield for setting options. See {@link https://postgis.net/docs/ST_IsValidDetail.html|ST_IsValidDetail} for details.
  * @param reason - A pointer to store the null-terminated string describing the validity of the geometry
  * @param location - A pointer to store the GEOSGeometry object indicating the location of invalidity
  * @returns number - 1 if the geometry is valid, 0 if it is invalid, or -1 on error
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a1f7b2d9e0a7d2e9d7e6f4a7a2f3e5b9d|GEOSisValidDetail_r}
  */
  GEOSisValidDetail_r(handle: number, g: number, flags: number, reason: string, location: number): number;
  /**
  * Tests whether two geometries are topologically equal.
  * @param g1 - A pointer to a GEOSGeometry object
  * @param g2 - A pointer to another GEOSGeometry object
  * @returns number - 1 if they are equal, 0 if they are not, or -1 on error
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a5b6b7a4a8a0c5f0c1a2e3a9d9b4f6c7f|GEOSEquals}
  */
  GEOSEquals(g1: number, g2: number): number;
  /**
  * Tests whether two geometries are topologically equal.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry to compare.
  * @param g2 - The second geometry to compare.
  * @returns number - 1 if equal, 0 if not, -1 on exception.
    */
  GEOSEquals_r(handle: number, g1: number, g2: number): number;
  /**
  * Tests whether two geometries are exactly equal, up to a specified tolerance.
  * @param g1 - The first geometry to compare.
  * @param g2 - The second geometry to compare.
  * @param tolerance - The maximum distance between any pair of points that are considered equal.
  * @returns number - 1 if equal, 0 if not, -1 on exception.
    */
  GEOSEqualsExact(g1: number, g2: number, tolerance: number): number;
  /**
  * Tests whether two geometries are exactly equal, up to a specified tolerance.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry to compare.
  * @param g2 - The second geometry to compare.
  * @param tolerance - The maximum distance between any pair of points that are considered equal.
  * @returns number - 1 if equal, 0 if not, -1 on exception.
    */
  GEOSEqualsExact_r(handle: number, g1: number, g2: number, tolerance: number): number;
  /**
  * Computes the distance between the closest points of two geometries.
  * @param g1 - The first geometry to measure from.
  * @param g2 - The second geometry to measure to.
  * @param dist - A pointer to a double variable where the distance will be stored.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSDistance(g1: number, g2: number, dist: number): number;
  /**
  * Computes the distance between the closest points of two geometries.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry to measure from.
  * @param g2 - The second geometry to measure to.
  * @param dist - A pointer to a double variable where the distance will be stored.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSDistance_r(handle: number, g1: number, g2: number, dist: number): number;
  /**
  * Computes the distance between two geometries using an indexed algorithm.
* Useful when one or both geometries is very large.
  * @param g1 - Input geometry pointer.
  * @param g2 - Input geometry pointer.
  * @param dist - Pointer to be filled in with distance result.
  * @returns number - 1 on success, 0 on exception.
  * @see geos::operation::distance::IndexedFacetDistance
  */
  GEOSDistanceIndexed(g1: number, g2: number, dist: number): number;
  /**
  * Computes the distance between two geometries using an indexed algorithm.
* Useful when one or both geometries is very large.
* This is a thread-safe version of GEOSDistanceIndexed.
  * @param handle - The context handle.pointer
  * @param g1 - Input geometry pointer.
  * @param g2 - Input geometry pointer.
  * @param dist - Pointer to be filled in with distance result.
  * @returns number - 1 on success, 0 on exception.
  * @see geos::operation::distance::IndexedFacetDistance
  */
  GEOSDistanceIndexed_r(handle: number, g1: number, g2: number, dist: number): number;
  /**
  * Computes the Hausdorff distance between two geometries.
* The Hausdorff distance is a measure of how similar or dissimilar two geometries are.
* It is the greatest distance between any point in one geometry and the closest point in the other geometry.
* The smaller the Hausdorff distance, the more similar the geometries are.
  * @param g1 - Input geometry pointer.
  * @param g2 - Input geometry pointer.
  * @param dist - Pointer to be filled in with distance result.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSHausdorffDistance(g1: number, g2: number, dist: number): number;
  /**
  * Computes the Hausdorff distance between two geometries.
* The Hausdorff distance is a measure of how similar or dissimilar two geometries are.
* It is the greatest distance between any point in one geometry and the closest point in the other geometry.
* The smaller the Hausdorff distance, the more similar the geometries are.
* This is a thread-safe version of GEOSHausdorffDistance.
  * @param handle - The context handle.pointer
  * @param g1 - Input geometry pointer.
  * @param g2 - Input geometry pointer.
  * @param dist - Pointer to be filled in with distance result.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSHausdorffDistance_r(handle: number, g1: number, g2: number, dist: number): number;
  /**
  * Computes an approximation to the Hausdorff distance between two geometries by densifying them before computing their discrete Hausdorff distance.
* The densification fraction parameter controls how much densification occurs.
* A fraction of 0.0 results in no densification, while a fraction of 1.0 results in effectively computing the exact Hausdorff distance.
* The fraction is relative to the size of the input geometries: a fraction of 1.0 means that each input segment will be divided into a number of equal-length segments such that the minimum segment length in either input is equal to that length.
* Increasing values of fraction reduce performance but increase accuracy of the calculated distance.
* A reasonable fraction value is 0.25 (which corresponds to segmenting each input segment into quarters).
  * @param g1 - Input geometry pointer.
  * @param g2 - Input geometry pointer.
  * @param densifyFrac - A fraction between 0 and 1 denoting the degree to which edges should be subdivided (smaller value means more subdivisions).
  * @param dist - Pointer to be filled in with distance result.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSHausdorffDistanceDensify(g1: number, g2: number, densifyFrac: number, dist: number): number;
  /**
  * Computes the Hausdorff distance between two geometries, densifying them by a given fraction.
* The Hausdorff distance is a measure of how similar or dissimilar two geometries are.
* It is the greatest distance between any point in one geometry and the closest point in the other geometry.
* Densifying the geometries involves inserting extra vertices along every line segment so that
* the total length of the segments is increased by a given fraction.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @param densifyFrac - The fraction by which to densify each segment. Must be between 0 and 1.
  * @param distance - A pointer to a double to store the computed distance.
  * @returns number - 1 on success, 0 on exception.
  * @see https://libgeos.org/doxygen/classgeos_1_1algorithm_1_1distance_1_1DiscreteHausdorffDistance.html
  */
  GEOSHausdorffDistanceDensify_r(handle: number, g1: number, g2: number, densifyFrac: number, distance: number): number;
  /**
  * Computes the Fréchet distance between two geometries.
* The Fréchet distance is a measure of similarity between curves that takes into account
* the location and ordering of the points along the curves.
* It is defined as the minimum length of a leash that connects a dog and its owner who walk along their respective curves,
* without backtracking, from one end to the other end.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @param distance - A pointer to a double to store the computed distance.
  * @returns number - 1 on success, 0 on exception.
  * @see https://en.wikipedia.org/wiki/Fr%C3%A9chet_distance
  */
  GEOSFrechetDistance(g1: number, g2: number, distance: number): number;
  /**
  * Computes the Fréchet distance between two geometries using a GEOS context handle.
* The Fréchet distance is a measure of similarity between curves that takes into account
* the location and ordering of the points along the curves.
* It is defined as the minimum length of a leash that connects a dog and its owner who walk along their respective curves,
* without backtracking, from one end to the other end.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @param distance - A pointer to a double to store the computed distance.
  * @returns number - 1 on success, 0 on exception.
  * @see https://en.wikipedia.org/wiki/Fr%C3%A9chet_distance
  */
  GEOSFrechetDistance_r(handle: number, g1: number, g2: number, distance: number): number;
  /**
  * Computes the Fréchet distance between two geometries, densifying them by a given fraction.
* The Fréchet distance is a measure of similarity between curves that takes into account
* the location and ordering of the points along the curves.
* It is defined as the minimum length of a leash that connects a dog and its owner who walk along their respective curves,
* without backtracking, from one end to the other end.
* Densifying the geometries involves inserting extra vertices along every line segment so that
* the total length of the segments is increased by a given fraction.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @param densifyFrac - The fraction by which to densify each segment. Must be between 0 and 1.
  * @param distance - A pointer to a double to store the computed distance.
  * @returns number - 1 on success, 0 on exception.
  * @see https://en.wikipedia.org/wiki/Fr%C3%A9chet_distance
  */
  GEOSFrechetDistanceDensify(g1: number, g2: number, densifyFrac: number, distance: number): number;
  /**
  * Computes the Fréchet distance between two geometries using a GEOS context handle, densifying them by a given fraction.
* The Fréchet distance is a measure of similarity between curves that takes into account
* the location and ordering of the points along the curves.
* It is defined as the minimum length of a leash that connects a dog and its owner who walk along their respective curves,
* without backtracking, from one end to the other end.
* Densifying the geometries involves inserting extra vertices along every line segment so that
* the total length of the segments is increased by a given fraction.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry.
  * @param g2 - The second geometry.
  * @param densifyFrac - The fraction by which to densify each segment. Must be between 0 and 1.
  * @param distance - A pointer to a double to store the computed distance.
  * @returns number - 1 on success, 0 on exception.
  * @see https://en.wikipedia.org/wiki/Fr%C3%A9chet_distance
  */
  GEOSFrechetDistanceDensify_r(handle: number, g1: number, g2: number, densifyFrac: number, distance: number): number;
  /**
  * Calculate the area of a geometry.
  * @param g - A pointer to a GEOSGeometry object
  * @param area - A pointer to be filled in with area result
  * @returns number - 1 on success, 0 on exception
  * @see https://libgeos.org/doxygen/geos__c_8h.html#af5eec7e8796b756cf007fd21bb3a88ad
  */
  GEOSArea(g: number, area: number): number;
  /**
  * Calculate the area of a geometry using a custom context handle.
  * @param handle - A GEOSContextHandle_t object
  * @param g - A pointer to a GEOSGeometry object
  * @param area - A pointer to be filled in with area result
  * @returns number - 1 on success, 0 on exception
    */
  GEOSArea_r(handle: number, g: number, area: number): number;
  /**
  * Calculate the length of a geometry.
  * @param g - A pointer to a GEOSGeometry object
  * @param length - A pointer to be filled in with length result
  * @returns number - 1 on success, 0 on exception
    */
  GEOSLength(g: number, length: number): number;
  /**
  * Calculate the length of a geometry using a custom context handle.
  * @param handle - A GEOSContextHandle_t object
  * @param g - A pointer to a GEOSGeometry object
  * @param length - A pointer to be filled in with length result
  * @returns number - 1 on success, 0 on exception
    */
  GEOSLength_r(handle: number, g: number, length: number): number;
  /**
  * Compute the nearest points between two geometries.
  * @param g1 - A JSON string representing a GEOSGeometry object
  * @param g2 - A JSON string representing a GEOSGeometry object
  * @returns string - A JSON string representing a GEOSGeometry object containing two points, or null on exception
    */
  GEOSNearestPoints(g1: string, g2: string): string;
  /**
  * Returns a GEOSCoordSequence of the nearest points between two geometries.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first geometry as a WKT string.
  * @param g2 - The second geometry as a WKT string.
  * @returns string - A WKT string representing the GEOSCoordSequence of the nearest points.
    */
  GEOSNearestPoints_r(handle: number, g1: string, g2: string): string;
  /**
  * Returns a GEOSGeometry from a Well-Known Text (WKT) representation.
  * @param wkt - The WKT representation of the geometry.
  * @returns number - A pointer to the GEOSGeometry object.
    */
  GEOSGeomFromWKT(wkt: number): number;
  /**
  * Returns a GEOSGeometry from a Well-Known Text (WKT) representation.
  * @param handle - A pointer to the GEOS context handle.
  * @param wkt - The WKT representation of the geometry.
  * @returns number - A pointer to the GEOSGeometry object.
    */
  GEOSGeomFromWKT_r(handle: number, wkt: number): number;
  /**
  * Returns a Well-Known Text (WKT) representation of a GEOSGeometry.
  * @param geom - A pointer to the GEOSGeometry object.
  * @returns number - A WKT representation of the geometry.
    */
  GEOSGeomToWKT(geom: number): number;
  /**
  * Returns a Well-Known Text (WKT) representation of a GEOSGeometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param geom - A pointer to the GEOSGeometry object.
  * @returns number - A WKT representation of the geometry.
    */
  GEOSGeomToWKT_r(handle: number, geom: number): number;
  /**
  * Converts a GEOSGeometry to a WKB representation.
  * @param g - The geometry pointer to convert.
  * @param size - A pointer to a size_t where the size of the returned buffer will be stored.
  * @returns number - A pointer to a buffer containing the WKB representation of the geometry, or NULL on exception. The caller is responsible for freeing the buffer with GEOSFree().
    */
  GEOSGeomToWKB_buf(g: number, size: number): number;
  /**
  * Converts a GEOSGeometry to a WKB representation using a context handle.
  * @param handle - The context handle.pointer
  * @param g - The geometry pointer to convert.
  * @param size - A pointer to a size_t where the size of the returned buffer will be stored.
  * @returns number - A pointer to a buffer containing the WKB representation of the geometry, or NULL on exception. The caller is responsible for freeing the buffer with GEOSFree_r().
    */
  GEOSGeomToWKB_buf_r(handle: number, g: number, size: number): number;
  /**
  * Converts a WKB representation to a GEOSGeometry.
  * @param wkb - A pointer to a buffer containing the WKB representation of the geometry.
  * @param size - The size of the buffer.
  * @returns number - A pointer to a newly allocated GEOSGeometry, or NULL on exception. The caller is responsible for destroying it with GEOSGeom_destroy().
    */
  GEOSGeomFromWKB_buf(wkb: number, size: number): number;
  /**
  * Converts a WKB representation to a GEOSGeometry using a context handle.
  * @param handle - The context handle.pointer
  * @param wkb - A pointer to a buffer containing the WKB representation of the geometry.
  * @param size - The size of the buffer.
  * @returns number - A pointer to a newly allocated GEOSGeometry, or NULL on exception. The caller is responsible for destroying it with GEOSGeom_destroy_r().
    */
  GEOSGeomFromWKB_buf_r(handle: number, wkb: number, size: number): number;
  /**
  * Converts a GEOSGeometry to a HEX-encoded WKB representation.
  * @param g - The geometry pointer to convert.
  * @param size - A pointer to a size_t where the length of the returned string will be stored.
  * @returns string - A null-terminated string containing the HEX-encoded WKB representation of the geometry, or NULL on exception. The caller is responsible for freeing the string with GEOSFree().
    */
  GEOSGeomToHEX_buf(g: number, size: number): string;
  /**
  * Converts a geometry to a hexadecimal representation of its WKB.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The geometry to convert.
  * @param size - A pointer to a size_t where the size of the returned buffer will be stored.
  * @returns string - A null-terminated hexadecimal string, or NULL on exception. The caller is responsible for freeing it.
    */
  GEOSGeomToHEX_buf_r(handle: number, g: number, size: number): string;
  /**
  * Converts a hexadecimal representation of WKB to a geometry.
  * @param hex - The hexadecimal string to convert.
  * @param size - The size of the hexadecimal string.
  * @returns number - A geometry, or NULL on exception. The caller is responsible for destroying it.
    */
  GEOSGeomFromHEX_buf(hex: string, size: number): number;
  /**
  * Converts a hexadecimal representation of WKB to a geometry with a GEOS context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param hex - The hexadecimal string to convert.
  * @param size - The size of the hexadecimal string.
  * @returns number - A geometry, or NULL on exception. The caller is responsible for destroying it.
    */
  GEOSGeomFromHEX_buf_r(handle: number, hex: string, size: number): number;
  /**
  * Tests whether a geometry is empty.
  * @param g - The geometry to test.
  * @returns number - 1 if the geometry is empty, 0 if it is not, or 2 on exception.
    */
  GEOSisEmpty(g: number): number;
  /**
  * Tests whether a geometry is empty with a GEOS context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The geometry to test.
  * @returns number - 1 if the geometry is empty, 0 if it is not, or 2 on exception.
    */
  GEOSisEmpty_r(handle: number, g: number): number;
  /**
  * Tests whether a geometry is simple.
* A geometry is simple if it has no points of self-tangency, self-intersection or other anomalous points.
  * @param g - a pointer to a GEOSGeometry object
  * @returns number - 1 if the geometry is simple, 0 if not, or -1 on error
    */
  GEOSisSimple(g: number): number;
  /**
  * Tests whether a geometry is simple (thread-safe version).
* A geometry is simple if it has no points of self-tangency, self-intersection or other anomalous points.
  * @param handle - a pointer to a GEOS context handle
  * @param g - a pointer to a GEOSGeometry object
  * @returns number - 1 if the geometry is simple, 0 if not, or -1 on error
    */
  GEOSisSimple_r(handle: number, g: number): number;
  /**
  * Tests whether a geometry is a ring.
* A ring is a closed and simple linestring.
  * @param g - a pointer to a GEOSGeometry object
  * @returns number - 1 if the geometry is a ring, 0 if not, or -1 on error
    */
  GEOSisRing(g: number): number;
  /**
  * Tests whether a geometry is a ring (thread-safe version).
* A ring is a closed and simple linestring.
  * @param handle - a pointer to a GEOS context handle
  * @param g - a pointer to a GEOSGeometry object
  * @returns number - 1 if the geometry is a ring, 0 if not, or -1 on error
    */
  GEOSisRing_r(handle: number, g: number): number;
  /**
  * Returns the type name of a geometry as a string.
* The possible values are: "Point", "LineString", "LinearRing", "Polygon", "MultiPoint", "MultiLineString", "MultiPolygon", "GeometryCollection".
  * @param g - a pointer to a GEOSGeometry object
  * @returns string - the type name of the geometry, or NULL on error
    */
  GEOSGeomType(g: number): string;
  /**
  * Returns the name of a geometry's type.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The geometry pointer to get the type name of.
  * @returns string - The name of the geometry's type, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9f3a3f0d6a8c7e9a0b4f2b7c6d0e5f1a
  */
  GEOSGeomType_r(handle: number, g: number): string;
  /**
  * Returns the type ID of a geometry.
  * @param g - The geometry pointer to get the type ID of.
  * @returns number - The type ID of the geometry, or -1 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a6e4b5f2c7d0b6a9a3e7b2d5f8d7f9e4c
  */
  GEOSGeomTypeId(g: number): number;
  /**
  * Returns the type ID of a geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The geometry pointer to get the type ID of.
  * @returns number - The type ID of the geometry, or -1 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3b9d5a4f2e9b4a0c2e8a6f3b1c1d2e0b
  */
  GEOSGeomTypeId_r(handle: number, g: number): number;
  /**
  * Computes an envelope (bounding box) of a geometry.
  * @param g - The geometry pointer to compute the envelope of.
  * @returns number - A newly allocated geometry pointer representing the envelope, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3a5d7f5dcbdb489edddbb43374ece97c
  */
  GEOSEnvelope(g: number): number;
  /**
  * Computes an envelope (bounding box) of a geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The geometry pointer to compute the envelope of.
  * @returns number - A newly allocated geometry pointer representing the envelope, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a4b1b77abcf0b320bc625129ff5e6d1d6
  */
  GEOSEnvelope_r(handle: number, g: number): number;
  /**
  * Computes a geometry that represents the point set intersection of the input geometries.
  * @param g1 - A pointer to the first geometry.
  * @param g2 - A pointer to the second geometry.
  * @returns number - A pointer to the resulting geometry, or NULL if an exception was thrown.
    */
  GEOSIntersection(g1: number, g2: number): number;
  /**
  * Computes a geometry that represents the point set intersection of the input geometries.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - A pointer to the first geometry.
  * @param g2 - A pointer to the second geometry.
  * @returns number - A pointer to the resulting geometry, or NULL if an exception was thrown.
    */
  GEOSIntersection_r(handle: number, g1: number, g2: number): number;
  /**
  * Computes a geometry that represents the point set intersection of the input geometries using a precision model.
  * @param g1 - A pointer to the first geometry.
  * @param g2 - A pointer to the second geometry.
  * @param gridSize - The grid size for the precision model.
  * @returns number - A pointer to the resulting geometry, or NULL if an exception was thrown.
    */
  GEOSIntersectionPrec(g1: number, g2: number, gridSize: number): number;
  /**
  * Computes a geometry that represents the point set intersection of the input geometries using a precision model.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - A pointer to the first geometry.
  * @param g2 - A pointer to the second geometry.
  * @param gridSize - The grid size for the precision model.
  * @returns number - A pointer to the resulting geometry, or NULL if an exception was thrown.
    */
  GEOSIntersectionPrec_r(handle: number, g1: number, g2: number, gridSize: number): number;
  /**
  * Computes a geometry that represents the point set difference of the input geometries.
  * @param ga - A pointer to the first geometry.
  * @param gb - A pointer to the second geometry.
  * @returns number - A pointer to the resulting geometry, or NULL if an exception was thrown.
    */
  GEOSDifference(ga: number, gb: number): number;
  /**
  * Computes a geometry that represents the point set difference of the input geometries.
  * @param handle - A pointer to the GEOS context handle.
  * @param ga - A pointer to the first geometry.
  * @param gb - A pointer to the second geometry.
  * @returns number - A pointer to the resulting geometry, or NULL if an exception was thrown.
    */
  GEOSDifference_r(handle: number, ga: number, gb: number): number;
  /**
  * Computes a geometry that represents the point set difference of the input geometries using a precision model.
  * @param ga - A pointer to the first geometry.
  * @param gb - A pointer to the second geometry.
  * @param gridSize - The grid size for the precision model.
  * @returns number - A pointer to the resulting geometry, or NULL if an exception was thrown.
    */
  GEOSDifferencePrec(ga: number, gb: number, gridSize: number): number;
  /**
  * Computes a geometry that represents the point set difference of the input geometries using a precision model.
  * @param handle - A pointer to the GEOS context handle.
  * @param ga - A pointer to the first geometry.
  * @param gb - A pointer to the second geometry.
  * @param gridSize - The grid size for the precision model.
  * @returns number - A pointer to the resulting geometry, or NULL if an exception was thrown.
    */
  GEOSDifferencePrec_r(handle: number, ga: number, gb: number, gridSize: number): number;
  /**
  * Computes a buffer area around this geometry having the given width and with a specified accuracy of approximation for circular arcs.
  * @param g - A pointer to the input geometry.
  * @param width - The width of the buffer area around this geometry. Negative values create a buffer inside this geometry.
  * @param quadsegs - The number of segments used to approximate a quarter circle (defaults to 8).
  * @returns number - A pointer to the resulting geometry, or NULL if an exception was thrown.
    */
  GEOSBuffer(g: number, width: number, quadsegs: number): number;
  /**
  * Creates a buffer around a geometry with a specified width, number of segments per quadrant, end cap style, join style and mitre limit.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The input geometry pointer.
  * @param width - The buffer width.
  * @param quadsegs - The number of segments per quadrant.
  * @param endCapStyle - The end cap style. One of GEOSBUF_CAP_ROUND, GEOSBUF_CAP_FLAT or GEOSBUF_CAP_SQUARE.
  * @param joinStyle - The join style. One of GEOSBUF_JOIN_ROUND, GEOSBUF_JOIN_MITRE or GEOSBUF_JOIN_BEVEL.
  * @param mitreLimit - The mitre limit. A positive number or zero.
  * @returns number - The buffered geometry pointer or NULL if an exception occurred.
    */
  GEOSBuffer_r(handle: number, g: number, width: number, quadsegs: number, endCapStyle: number, joinStyle: number, mitreLimit: number): number;
  /**
  * Creates a buffer around a geometry with a specified width, number of segments per quadrant, end cap style, join style and mitre limit.
  * @param g - The input geometry pointer.
  * @param width - The buffer width.
  * @param quadsegs - The number of segments per quadrant.
  * @param endCapStyle - The end cap style. One of GEOSBUF_CAP_ROUND, GEOSBUF_CAP_FLAT or GEOSBUF_CAP_SQUARE.
  * @param joinStyle - The join style. One of GEOSBUF_JOIN_ROUND, GEOSBUF_JOIN_MITRE or GEOSBUF_JOIN_BEVEL.
  * @returns number - The buffered geometry pointer or NULL if an exception occurred.
    */
  GEOSBufferWithStyle(g: number, width: number, quadsegs: number, endCapStyle: number, joinStyle: number): number;
  /**
  * Creates a buffer around a geometry with a specified width, number of segments per quadrant, end cap style, join style and mitre limit using a given context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The input geometry pointer.
  * @param width - The buffer width.
  * @param quadsegs - The number of segments per quadrant.
  * @param endCapStyle - The end cap style. One of GEOSBUF_CAP_ROUND, GEOSBUF_CAP_FLAT or GEOSBUF_CAP_SQUARE.
  * @param joinStyle - The join style. One of GEOSBUF_JOIN_ROUND, GEOSBUF_JOIN_MITRE or GEOSBUF_JOIN_BEVEL.
  * @param mitreLimit - The mitre limit. A positive number or zero.
  * @returns number - The buffered geometry pointer or NULL if an exception occurred.
    */
  GEOSBufferWithStyle_r(handle: number, g: number, width: number, quadsegs: number, endCapStyle: number, joinStyle: number, mitreLimit: number): number;
  /**
  * Creates a single-sided buffer around a geometry with a specified width, number of segments per quadrant, side (left or right), end cap style and join style using a given context handle.
* A positive distance will create a buffer on the left side of the geometry (as viewed from the start point to the end point), while a negative distance will create a buffer on the right side.
* A single-sided buffer is constructed by offsetting the input line by the buffer distance then forming a buffer around the offset line by the buffer distance (using round ends).
* This produces a buffer polygon with one straight side which is parallel to the input line at the given offset distance.
* If the input line is oriented counter-clockwise then the left side is inside the polygon; otherwise it is outside.
* If the input line is not simple (i.e. it loops back on itself) then there may be multiple left sides and right sides; in this case it is not clearly defined which side should be buffered.
* In this case the operation may either fail with an exception or return an approximate result which may not be valid.
* This function does not support buffering MultiLineStrings (the result will be an empty polygon).
  * @param g - The input geometry pointer.
  * @param width - The buffer width.
  * @param quadsegs - The number of segments per quadrant.
  * @param joinStyle - The join style. One of 1 = GEOSBUF_JOIN_ROUND, 2 = GEOSBUF_JOIN_MITRE or 3 = GEOSBUF_JOIN_BEVEL.
  * @param mitreLimit - The mitre limit. A positive number or zero.
  * @param side - The side to buffer. 1 for left, -1 for right.
  * @returns number - The single-sided buffered geometry pointer or NULL if an exception occurred.
    */
  GEOSSingleSidedBuffer(g: number, width: number, quadsegs: number, joinStyle: number, mitreLimit: number, side: number): number;
  /**
  * Creates a single-sided buffer around a geometry with a specified width, number of segments per quadrant, side (left or right), end cap style and join style using a given context handle.
* A positive distance will create a buffer on the left side of the geometry (as viewed from the start point to the end point), while a negative distance will create a buffer on the right side.
* A single-sided buffer is constructed by offsetting the input line by the buffer distance then forming a buffer around the offset line by the buffer distance (using round ends).
* This produces a buffer polygon with one straight side which is parallel to the input line at the given offset distance.
* If the input line is oriented counter-clockwise then the left side is inside the polygon; otherwise it is outside.
* If the input line is not simple (i.e. it loops back on itself) then there may be multiple left sides and right sides; in this case it is not clearly defined which side should be buffered.
* In this case the operation may either fail with an exception or return an approximate result which may not be valid.
* This function does not support buffering MultiLineStrings (the result will be an empty polygon).
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The input geometry pointer.
  * @param width - The buffer width.
  * @param quadsegs - The number of segments per quadrant.
  * @param side - The side to buffer. 1 for left, -1 for right.
  * @param endCapStyle - The end cap style. One of 1 = GEOSBUF_CAP_ROUND, 2 = GEOSBUF_CAP_FLAT or 3 =GEOSBUF_CAP_SQUARE.
  * @param joinStyle - The join style. One of 1 = GEOSBUF_JOIN_ROUND, 2 = GEOSBUF_JOIN_MITRE or 3 = GEOSBUF_JOIN_BEVEL.
  * @returns number - The single-sided buffered geometry pointer or NULL if an exception occurred.
    */
  GEOSSingleSidedBuffer_r(handle: number, g: number, width: number, quadsegs: number, side: number, endCapStyle: number, joinStyle: number): number;
  /**
  * Computes an offset curve from a geometry. An offset curve is a linear geometry which is offset a given distance from the input. If the offset distance is positive the curve lies on the left side of the input; if it is negative the curve is on the right side. The curve(s) have the same direction as the input line(s).
  * @param geom - A pointer to a GEOSGeometry object.
  * @param width - The offset distance.
  * @param quadsegs - The number of quadrant segments used to approximate a quarter circle (defaults to 8).
  * @param joinStyle - The join style for line joins (defaults to BufferParameters::JOIN_STYLE).
  * @param mitreLimit - The mitre ratio limit (only applies for mitre join style, defaults to 5.0).
  * @returns number - A pointer to a new GEOSGeometry object representing the offset curve, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a06557cad6153f56703b6875d4f21dc18
  */
  GEOSOffsetCurve(geom: number, width: number, quadsegs: number, joinStyle: number, mitreLimit: number): number;
  /**
  * Computes an offset curve from a geometry using a handle. An offset curve is a linear geometry which is offset a given distance from the input. If the offset distance is positive the curve lies on the left side of the input; if it is negative the curve is on the right side. The curve(s) have the same direction as the input line(s).
  * @param handle - A pointer to a GEOSContextHandle_t object.
  * @param geom - A pointer to a GEOSGeometry object.
  * @param width - The offset distance.
  * @param quadsegs - The number of quadrant segments used to approximate a quarter circle (defaults to 8).
  * @param joinStyle - The join style for line joins (defaults to BufferParameters::JOIN_STYLE).
  * @param mitreLimit - The mitre ratio limit (only applies for mitre join style, defaults to 5.0).
  * @returns number - A pointer to a new GEOSGeometry object representing the offset curve, or NULL on exception.
  * @see https://libgeos.org/doxygen/classgeos_1_1operation_1_1buffer_1_1OffsetCurve.html
  */
  GEOSOffsetCurve_r(handle: number, geom: number, width: number, quadsegs: number, joinStyle: number, mitreLimit: number): number;
  /**
  * Computes the convex hull of a geometry. The convex hull is the smallest convex geometry that contains all the points in the input geometry.
  * @param geom - A pointer to a GEOSGeometry object.
  * @returns number - A pointer to a new GEOSGeometry object representing the convex hull, or NULL on exception.
  * @see https://libgeos.org/doxygen/classgeos_1_1algorithm_1_1ConvexHull.html
  */
  GEOSConvexHull(geom: number): number;
  /**
  * Computes the convex hull of a geometry using a handle. The convex hull is the smallest convex geometry that contains all the points in the input geometry.
  * @param handle - A pointer to a GEOSContextHandle_t object.
  * @param geom - A pointer to a GEOSGeometry object.
  * @returns number - A pointer to a new GEOSGeometry object representing the convex hull, or NULL on exception.
  * @see https://libgeos.org/doxygen/classgeos_1_1algorithm_1_1ConvexHull.html
  */
  GEOSConvexHull_r(handle: number, geom: number): number;
  /**
  * Returns a "concave hull" of a geometry. A concave hull is a polygon which contains all the points of the input, but is a better approximation than the convex hull to the area occupied by the input. Frequently used to convert a multi-point into a polygonal area. that contains all the points in the input Geometry.
* 
* A set of points has a sequence of hulls of increasing concaveness, determined by a numeric target parameter. The concave hull is constructed by removing the longest outer edges of the Delaunay Triangulation of the space between the polygons, until the target criterion parameter is reached. This can be expressed as a ratio between the lengths of the longest and shortest edges. 1 produces the convex hull; 0 produces a hull with maximum concaveness
  * @param geom - A pointer to a GEOSGeometry object.
  * @param ratio - The edge length ratio value, between 0 and 1.
  * @param allowHoles - When non-zero, the polygonal output may contain holes.
  * @returns number - A newly allocated geometry of the concave hull. NULL on exception.
    */
  GEOSConcaveHull(geom: number, ratio: number, allowHoles: number): number;
  /**
  * Returns a "concave hull" of a geometry. A concave hull is a polygon which contains all the points of the input, but is a better approximation than the convex hull to the area occupied by the input. Frequently used to convert a multi-point into a polygonal area. that contains all the points in the input Geometry.
* 
* A set of points has a sequence of hulls of increasing concaveness, determined by a numeric target parameter. The concave hull is constructed by removing the longest outer edges of the Delaunay Triangulation of the space between the polygons, until the target criterion parameter is reached. This can be expressed as a ratio between the lengths of the longest and shortest edges. 1 produces the convex hull; 0 produces a hull with maximum concaveness
  * @param handle - A pointer to the GEOS context handle.
  * @param geom - A pointer to a GEOSGeometry object.
  * @param ratio - The edge length ratio value, between 0 and 1.
  * @param allowHoles - When non-zero, the polygonal output may contain holes.
  * @returns number - A newly allocated geometry of the concave hull. NULL on exception.
    */
  GEOSConcaveHull_r(handle: number, geom: number, ratio: number, allowHoles: number): number;
  /**
  * Returns a "concave hull" of a geometry. A concave hull is a polygon which contains all the points of the input, but is a better approximation than the convex hull to the area occupied by the input. Frequently used to convert a multi-point into a polygonal area. that contains all the points in the input Geometry.
* 
* A set of points has a sequence of hulls of increasing concaveness, determined by a numeric target parameter. The concave hull is constructed by removing the longest outer edges of the Delaunay Triangulation of the space between the polygons, until the specified maximm edge length is reached. A large value produces the convex hull, 0 produces the hull of maximim concaveness.
  * @param geom - A pointer to a GEOSGeometry object.
  * @param length - The maximum edge length (0 or greater).
  * @param allowHoles - When non-zero, the polygonal output may contain holes.
  * @returns number - A newly allocated geometry of the concave hull. NULL on exception.
    */
  GEOSConcaveHullByLength(geom: number, length: number, allowHoles: number): number;
  /**
  * Returns a "concave hull" of a geometry. A concave hull is a polygon which contains all the points of the input, but is a better approximation than the convex hull to the area occupied by the input. Frequently used to convert a multi-point into a polygonal area. that contains all the points in the input Geometry.
* 
* A set of points has a sequence of hulls of increasing concaveness, determined by a numeric target parameter. The concave hull is constructed by removing the longest outer edges of the Delaunay Triangulation of the space between the polygons, until the specified maximm edge length is reached. A large value produces the convex hull, 0 produces the hull of maximim concaveness.
  * @param handle - A pointer to the GEOS context handle.
  * @param geom - A pointer to a GEOSGeometry object.
  * @param length - The maximum edge length (0 or greater).
  * @param allowHoles - When non-zero, the polygonal output may contain holes.
  * @returns number - A newly allocated geometry of the concave hull. NULL on exception.
    */
  GEOSConcaveHullByLength_r(handle: number, geom: number, length: number, allowHoles: number): number;
  /**
  * Constructs a concave hull of a set of polygons, respecting the polygons as constraints.
* A concave hull is a (possibly) non-convex polygon containing all the input polygons. The computed hull "fills the gap" between the polygons, and does not intersect their interior. A set of polygons has a sequence of hulls of increasing concaveness, determined by a numeric target parameter.
* The concave hull is constructed by removing the longest outer edges of the Delaunay Triangulation of the space between the polygons, until the target criterion parameter is reached. The "Maximum Edge Length" parameter limits the length of the longest edge between polygons to be no larger than this value. This can be expressed as a ratio between the lengths of the longest and shortest edges.
* The input polygons must be a valid MultiPolygon (i.e. they must be non-overlapping).
  * @param geom - A pointer to a GEOSGeometry object.
  * @param lengthRatio - specifies the Maximum Edge Length as a fraction of the difference between the longest and shortest edge lengths between the polygons. This normalizes the Maximum Edge Length to be scale-free. A value of 1 produces the convex hull; a value of 0 produces the original polygons.
  * @param isTight - does the hull follow the outer boundaries of the input polygons.
  * @param isHolesAllowed - is the concave hull allowed to contain holes?
  * @returns number - A newly allocated geometry of the concave hull. NULL on exception.
    */
  GEOSConcaveHullOfPolygons(geom: number, lengthRatio: number, isTight: number, isHolesAllowed: number): number;
  /**
  * Constructs a concave hull of a set of polygons, respecting the polygons as constraints.
* A concave hull is a (possibly) non-convex polygon containing all the input polygons. The computed hull "fills the gap" between the polygons, and does not intersect their interior. A set of polygons has a sequence of hulls of increasing concaveness, determined by a numeric target parameter.
* The concave hull is constructed by removing the longest outer edges of the Delaunay Triangulation of the space between the polygons, until the target criterion parameter is reached. The "Maximum Edge Length" parameter limits the length of the longest edge between polygons to be no larger than this value. This can be expressed as a ratio between the lengths of the longest and shortest edges.
* The input polygons must be a valid MultiPolygon (i.e. they must be non-overlapping).
  * @param handle - A pointer to the GEOS context handle.
  * @param geom - A pointer to a GEOSGeometry object.
  * @param lengthRatio - specifies the Maximum Edge Length as a fraction of the difference between the longest and shortest edge lengths between the polygons. This normalizes the Maximum Edge Length to be scale-free. A value of 1 produces the convex hull; a value of 0 produces the original polygons.
  * @param isTight - does the hull follow the outer boundaries of the input polygons.
  * @param isHolesAllowed - is the concave hull allowed to contain holes?
  * @returns number - A newly allocated geometry of the concave hull. NULL on exception.
    */
  GEOSConcaveHullOfPolygons_r(handle: number, geom: number, lengthRatio: number, isTight: number, isHolesAllowed: number): number;
  /**
  * Computes the minimum rotated rectangle (MRR) of a geometry. The MRR is similar to an envelope, but not necessarily aligned with coordinate axes. It has minimum area among all rectangles enclosing its input geometry.
  * @param geom - A pointer to a GEOSGeometry object.
  * @returns number - A pointer to a new GEOSGeometry object representing the MRR, or NULL on exception.
    */
  GEOSMinimumRotatedRectangle(geom: number): number;
  /**
  * Computes the minimum-area rotated rectangle containing a geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The pointer to the input geometry.
  * @returns number - The pointer to the output geometry, or NULL on exception.
    */
  GEOSMinimumRotatedRectangle_r(handle: number, g: number): number;
  /**
  * Constructs the "maximum inscribed circle" (MIC) for a polygonal geometry, up to a specified tolerance. The MIC is determined by a point in the interior of the area which has the farthest distance from the area boundary, along with a boundary point at that distance. In the context of geography the center of the MIC is known as the "pole of inaccessibility". A cartographic use case is to determine a suitable point to place a map label within a polygon. The radius length of the MIC is a measure of how "narrow" a polygon is. It is the distance at which the negative buffer becomes empty. The class supports polygons with holes and multipolygons. The implementation uses a successive-approximation technique over a grid of square cells covering the area geometry. The grid is refined using a branch-and-bound algorithm. Point containment and distance are computed in a performant way by using spatial indexes. Returns a two-point linestring, with one point at the center of the inscribed circle and the other on the boundary of the inscribed circle.
  * @param g - The pointer to the input geometry.
  * @param tolerance - The distance tolerance for computing the center point.
  * @returns number - The pointer to the output geometry, or NULL on exception.
    */
  GEOSMaximumInscribedCircle(g: number, tolerance: number): number;
  /**
  * Constructs the "maximum inscribed circle" (MIC) for a polygonal geometry, up to a specified tolerance. The MIC is determined by a point in the interior of the area which has the farthest distance from the area boundary, along with a boundary point at that distance. In the context of geography the center of the MIC is known as the "pole of inaccessibility". A cartographic use case is to determine a suitable point to place a map label within a polygon. The radius length of the MIC is a measure of how "narrow" a polygon is. It is the distance at which the negative buffer becomes empty. The class supports polygons with holes and multipolygons. The implementation uses a successive-approximation technique over a grid of square cells covering the area geometry. The grid is refined using a branch-and-bound algorithm. Point containment and distance are computed in a performant way by using spatial indexes. Returns a two-point linestring, with one point at the center of the inscribed circle and the other on the boundary of the inscribed circle.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The pointer to the input geometry.
  * @param tolerance - The distance tolerance for computing the center point.
  * @returns number - The pointer to the output geometry, or NULL on exception.
    */
  GEOSMaximumInscribedCircle_r(handle: number, g: number, tolerance: number): number;
  /**
  * Constructs the "largest empty circle" (LEC) for a set of obstacle geometries and within a polygonal boundary, with accuracy to to a specified distance tolerance. The obstacles may be any collection of points, lines and polygons. The LEC is the largest circle whose interior does not intersect with any obstacle. and which has its center inside the given boundary. If no boundary is provided, the convex hull of the obstacles is used. The LEC center is the point in the interior of the boundary which has the farthest distance from the obstacles (up to the given distance tolerance). The LEC is determined by the center point and a point indicating the circle radius (which will lie on an obstacle).
* To compute an LEC which lies wholly within a polygonal boundary, include the boundary of the polygon(s) as a linear obstacle.
* The implementation uses a successive-approximation technique over a grid of square cells covering the obstacles and boundary. The grid is refined using a branch-and-bound algorithm. Point containment and distance are computed in a performant way by using spatial indexes.
* Returns the LEC radius as a two-point linestring, with the start point at the center of the inscribed circle and the end on the boundary of the circle.
  * @param g - The pointer to the input obstacle geometry.
  * @param boundary - The pointer to the input boundary geometry, or NULL if not provided.
  * @param tolerance - The distance tolerance for computing the center point and radius.
  * @returns number - The pointer to the output geometry, or NULL on exception.
    */
  GEOSLargestEmptyCircle(g: number, boundary: number, tolerance: number): number;
  /**
  * Constructs the "largest empty circle" (LEC) for a set of obstacle geometries and within a polygonal boundary, with accuracy to to a specified distance tolerance. The obstacles may be any collection of points, lines and polygons. The LEC is the largest circle whose interior does not intersect with any obstacle. and which has its center inside the given boundary. If no boundary is provided, the convex hull of the obstacles is used. The LEC center is the point in the interior of the boundary which has the farthest distance from the obstacles (up to the given distance tolerance). The LEC is determined by the center point and a point indicating the circle radius (which will lie on an obstacle).
* To compute an LEC which lies wholly within a polygonal boundary, include the boundary of the polygon(s) as a linear obstacle.
* The implementation uses a successive-approximation technique over a grid of square cells covering the obstacles and boundary. The grid is refined using a branch-and-bound algorithm. Point containment and distance are computed in a performant way by using spatial indexes.
* Returns the LEC radius as a two-point linestring, with the start point at the center of the inscribed circle and the end on the boundary of the circle.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The pointer to the input obstacle geometry.
  * @param boundary - The pointer to the input boundary geometry, or NULL if not provided.
  * @param tolerance - The distance tolerance for computing the center point and radius.
  * @returns number - The pointer to the output geometry, or NULL on exception.
    */
  GEOSLargestEmptyCircle_r(handle: number, g: number, boundary: number, tolerance: number): number;
  /**
  * Computes the minimum width of a geometry.
* The minimum width is the smallest distance between two parallel lines that contain the geometry.
  * @param g - The input geometry pointer.
  * @returns number - A pointer to a LineString containing two points defining the minimum width line segment, or NULL on exception.
    */
  GEOSMinimumWidth(g: number): number;
  /**
  * Computes the minimum width of a geometry in a reentrant way.
* The minimum width is the smallest distance between two parallel lines that contain the geometry.
  * @param handle - The context handle pointer to use for error reporting.
  * @param g - The input geometry pointer.
  * @returns number - A pointer to a LineString containing two points defining the minimum width line segment, or NULL on exception.
    */
  GEOSMinimumWidth_r(handle: number, g: number): number;
  /**
  * Computes the minimum clearance line of a geometry.
* The minimum clearance line is a line connecting two points in the geometry that have the minimum clearance distance between them.
  * @param g - The input geometry pointer.
  * @returns number - A pointer to a LineString containing two points defining the minimum clearance line segment, or NULL on exception or if no such line exists (e.g. for an empty or point geometry).
    */
  GEOSMinimumClearanceLine(g: number): number;
  /**
  * Computes the minimum clearance line of a geometry in a reentrant way.
* The minimum clearance line is a line connecting two points in the geometry that have the minimum clearance distance between them.
  * @param handle - The context handle pointer to use for error reporting.
  * @param g - The input geometry pointer.
  * @returns number - A pointer to a LineString containing two points defining the minimum clearance line segment, or NULL on exception or if no such line exists (e.g. for an empty or point geometry).
    */
  GEOSMinimumClearanceLine_r(handle: number, g: number): number;
  /**
  * Computes the minimum clearance distance of a geometry.
* The minimum clearance distance is the smallest distance between any pair of points in the geometry.
  * @param g - The input geometry pointer.
  * @param dist - A pointer to store the result distance, or NULL if not needed.
  * @returns number - 1 on success, 0 on exception or if no such distance exists (e.g. for an empty or point geometry).
    */
  GEOSMinimumClearance(g: number, dist: number): number;
  /**
  * Computes the minimum clearance of a geometry.
* The minimum clearance is the smallest amount by which
* a vertex of a geometry can be moved to produce an invalid geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry to compute the clearance for.
  * @param d - The output parameter for the clearance value.
  * @returns number - 1 on success, 0 on exception.
  * @see https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in
  */
  GEOSMinimumClearance_r(handle: number, g: number, d: number): number;
  /**
  * Computes the difference between two geometries.
* The difference is the part of the first geometry that does not intersect with the second geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first GEOS geometry.
  * @param g2 - The second GEOS geometry.
  * @returns number - A new GEOS geometry representing the difference, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html
  */
  GEOSDifference(handle: number, g1: number, g2: number): number;
  /**
  * Computes the difference between two geometries using a reentrant API.
* The difference is the part of the first geometry that does not intersect with the second geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first GEOS geometry.
  * @param g2 - The second GEOS geometry.
  * @returns number - A new GEOS geometry representing the difference, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html
  */
  GEOSDifference_r(handle: number, g1: number, g2: number): number;
  /**
  * Computes the difference between two geometries using a precision model grid size.
* The difference is the part of the first geometry that does not intersect with the second geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first GEOS geometry.
  * @param g2 - The second GEOS geometry.
  * @param grid_size - The grid size for snapping vertices to a precision model grid.
  * @returns number - A new GEOS geometry representing the difference, or NULL on exception.
    */
  GEOSDifferencePrec(handle: number, g1: number, g2: number, grid_size: number): number;
  /**
  * Computes the difference between two geometries using a precision model grid size and a reentrant API.
* The difference is the part of the first geometry that does not intersect with the second geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - The first GEOS geometry.
  * @param g2 - The second GEOS geometry.
  * @param grid_size - The grid size for snapping vertices to a precision model grid.
  * @returns number - A new GEOS geometry representing the difference, or NULL on exception.
    */
  GEOSDifferencePrec_r(handle: number, g1: number, g2: number, grid_size: number): number;
  /**
  * Computes the boundary of a geometry.
  * @param g - A pointer to a GEOSGeometry object.
  * @returns number - A pointer to a new GEOSGeometry object representing the boundary, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html
  */
  GEOSBoundary(g: number): number;
  /**
  * Computes the boundary of a geometry in a thread-safe manner.
  * @param handle - A GEOSContextHandle_t object.
  * @param g - A pointer to a GEOSGeometry object.
  * @returns number - A pointer to a new GEOSGeometry object representing the boundary, or NULL on exception.
    */
  GEOSBoundary_r(handle: number, g: number): number;
  /**
  * Computes the symmetric difference of two geometries.
  * @param g1 - A pointer to a GEOSGeometry object.
  * @param g2 - A pointer to another GEOSGeometry object.
  * @returns number - A pointer to a new GEOSGeometry object representing the symmetric difference, or NULL on exception.
    */
  GEOSSymDifference(g1: number, g2: number): number;
  /**
  * Computes the symmetric difference of two geometries in a thread-safe manner.
  * @param handle - A GEOSContextHandle_t object.
  * @param g1 - A pointer to a GEOSGeometry object.
  * @param g2 - A pointer to another GEOSGeometry object.
  * @returns number - A pointer to a new GEOSGeometry object representing the symmetric difference, or NULL on exception.
    */
  GEOSSymDifference_r(handle: number, g1: number, g2: number): number;
  /**
  * Computes the symmetric difference of two geometries with precision control.
  * @param g1 - A pointer to a GEOSGeometry object.
  * @param g2 - A pointer to another GEOSGeometry object.
  * @param gridSize - The size of the grid used for snapping coordinates. Must be positive and non-zero.
  * @returns number - A pointer to a new GEOSGeometry object representing the symmetric difference, or NULL on exception or invalid input.
    */
  GEOSSymDifferencePrec(g1: number, g2: number, gridSize: number): number;
  /**
  * Computes a symmetric difference of two geometries using a specified precision model.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - A pointer to the first geometry.
  * @param g2 - A pointer to the second geometry.
  * @param gridsize - The precision grid size to use.
  * @returns number - A pointer to the resulting geometry, or NULL on exception.
    */
  GEOSSymDifferencePrec_r(handle: number, g1: number, g2: number, gridsize: number): number;
  /**
  * Computes a union of two geometries.
  * @param g1 - A pointer to the first geometry.
  * @param g2 - A pointer to the second geometry.
  * @returns number - A pointer to the resulting geometry, or NULL on exception.
    */
  GEOSUnion(g1: number, g2: number): number;
  /**
  * Computes a union of two geometries using a specified context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - A pointer to the first geometry.
  * @param g2 - A pointer to the second geometry.
  * @returns number - A pointer to the resulting geometry, or NULL on exception.
    */
  GEOSUnion_r(handle: number, g1: number, g2: number): number;
  /**
  * Computes a union of two geometries using a specified precision model.
  * @param g1 - A pointer to the first geometry.
  * @param g2 - A pointer to the second geometry.
  * @param gridsize - The precision grid size to use.
  * @returns number - A pointer to the resulting geometry, or NULL on exception.
    */
  GEOSUnionPrec(g1: number, g2: number, gridsize: number): number;
  /**
  * Computes a union of two geometries using a specified precision model and context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - A pointer to the first geometry.
  * @param g2 - A pointer to the second geometry.
  * @param gridsize - The precision grid size to use.
  * @returns number - A pointer to the resulting geometry, or NULL on exception.
    */
  GEOSUnionPrec_r(handle: number, g1: number, g2: number, gridsize: number): number;
  /**
  * Computes the union of all the input geometries.
  * @param geom - A pointer to a GEOS geometry.
  * @returns number - A pointer to a new GEOS geometry representing the union, or NULL on exception.
  * @see https://postgis.net/docs/ST_UnaryUnion.html
  */
  GEOSUnaryUnion(geom: number): number;
  /**
  * Computes the union of all the input geometries using a GEOS context handle.
  * @param handle - A pointer to a GEOS context handle.
  * @param geom - A pointer to a GEOS geometry.
  * @returns number - A pointer to a new GEOS geometry representing the union, or NULL on exception.
  * @see https://docs.rs/geos-sys/2.0.3/geos_sys/fn.GEOSUnaryUnion_r.html
  */
  GEOSUnaryUnion_r(handle: number, geom: number): number;
  /**
  * Computes the union of all the input geometries with a specified precision grid size.
  * @param geom - A pointer to a GEOS geometry.
  * @param gridSize - The size of the precision grid to use for snapping.
  * @returns number - A pointer to a new GEOS geometry representing the union, or NULL on exception.
  * @see https://postgis.net/docs/ST_UnaryUnion.html
  */
  GEOSUnaryUnionPrec(geom: number, gridSize: number): number;
  /**
  * Computes the union of all the input geometries with a specified precision grid size using a GEOS context handle.
  * @param handle - A pointer to a GEOS context handle.
  * @param geom - A pointer to a GEOS geometry.
  * @param gridSize - The size of the precision grid to use for snapping.
  * @returns number - A pointer to a new GEOS geometry representing the union, or NULL on exception.
  * @see https://docs.rs/geos-sys/2.0.3/geos_sys/fn.GEOSUnaryUnionPrec_r.html
  */
  GEOSUnaryUnionPrec_r(handle: number, geom: number, gridSize: number): number;
  /**
  * Computes the union of all the input geometries using an optimized algorithm for coverages (collections of polygons that do not overlap).
  * @param geom - A pointer to a GEOS geometry.
  * @returns number - A pointer to a new GEOS geometry representing the union, or NULL on exception.
  * @see https://libgeos.org/doxygen/classgeos_1_1operation_1_1geounion_1_1CoverageUnionOp.html
  */
  GEOSCoverageUnion(geom: number): number;
  /**
  * Computes the union of a collection of polygonal geometries that are correctly noded and do not overlap.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The pointer to a GEOS geometry collection.
  * @returns number - The pointer to a GEOS geometry representing the union, or NULL on error.
  * @see https://libgeos.org/doxygen/classgeos_1_1operation_1_1geounion_1_1CascadedPolygonUnion.html
  */
  GEOSCoverageUnion_r(handle: number, g: number): number;
  /**
  * Computes a geometric shape that represents the points shared by two or more geometries.
  * @param g - The pointer to a GEOS geometry.
  * @returns number - The pointer to a GEOS geometry representing the noded version of the input, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html
  */
  GEOSNode(g: number): number;
  /**
  * Computes a geometric shape that represents the points shared by two or more geometries.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The pointer to a GEOS geometry.
  * @returns number - The pointer to a GEOS geometry representing the noded version of the input, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html
  */
  GEOSNode_r(handle: number, g: number): number;
  /**
  * Computes the union of a collection of polygonal geometries using an optimized algorithm.
  * @param g - The pointer to a GEOS geometry collection.
  * @returns number - The pointer to a GEOS geometry representing the union, or NULL on error.
  * @see https://libgeos.org/doxygen/classgeos_1_1operation_1_1geounion_1_1CascadedPolygonUnion.html
  */
  GEOSUnionCascaded(g: number): number;
  /**
  * Computes the union of a collection of polygonal geometries using an optimized algorithm.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The pointer to a GEOS geometry collection.
  * @returns number - The pointer to a GEOS geometry representing the union, or NULL on error.
  * @see https://libgeos.org/doxygen/classgeos_1_1operation_1_1geounion_1_1CascadedPolygonUnion.html
  */
  GEOSUnionCascaded_r(handle: number, g: number): number;
  /**
  * Computes a point that lies on the surface of a geometry.
  * @param g - A pointer to a GEOSGeometry object.
  * @returns number - A pointer to a GEOSGeometry object representing a point on the surface of g, or null if an error occurs.
    */
  GEOSPointOnSurface(g: number): number;
  /**
  * Computes a point that lies on the surface of a geometry, using a thread-safe context handle.
  * @param handle - A pointer to a GEOSContextHandle_t object.
  * @param g - A pointer to a GEOSGeometry object.
  * @returns number - A pointer to a GEOSGeometry object representing a point on the surface of g, or null if an error occurs.
    */
  GEOSPointOnSurface_r(handle: number, g: number): number;
  /**
  * Clips a geometry by a rectangular extent.
  * @param g - A pointer to a GEOSGeometry object.
  * @param xmin - The minimum x-coordinate of the clipping rectangle.
  * @param ymin - The minimum y-coordinate of the clipping rectangle.
  * @param xmax - The maximum x-coordinate of the clipping rectangle.
  * @param ymax - The maximum y-coordinate of the clipping rectangle.
  * @returns number - The clipped geometry or NULL on exception Caller is responsible for freeing with GEOSGeom_destroy().
    */
  GEOSClipByRect(g: number, xmin: number, ymin: number, xmax: number, ymax: number): number;
  /**
  * Clips a geometry by a rectangular extent, using a thread-safe context handle.
  * @param handle - A pointer to a GEOSContextHandle_t object.
  * @param g - A pointer to a GEOSGeometry object.
  * @param xmin - The minimum x-coordinate of the clipping rectangle.
  * @param ymin - The minimum y-coordinate of the clipping rectangle.
  * @param xmax - The maximum x-coordinate of the clipping rectangle.
  * @param ymax - The maximum y-coordinate of the clipping rectangle.
  * @returns number - The clipped geometry or NULL on exception Caller is responsible for freeing with GEOSGeom_destroy().
    */
  GEOSClipByRect_r(handle: number, g: number, xmin: number, ymin: number, xmax: number, ymax: number): number;
  /**
  * Destroys a geometry object and frees its memory.
  * @param g - A pointer to a GEOSGeometry object.
  * @returns void - undefined
    */
  GEOSGeom_destroy(g: number): void;
  /**
  * Destroys a GEOS geometry object.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry object to destroy.
  * @returns void - undefined
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a6a7f0c9b3f9a1d0f4e5c2a7b5b3e6a8d
  */
  GEOSGeom_destroy_r(handle: number, g: number): void;
  /**
  * Returns the number of coordinates in a GEOS geometry object.
  * @param g - The GEOS geometry object.
  * @returns number - The number of coordinates, or -1 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a7b4e0c2f9d3f1a6d4b9a9c1d5e0f3a2e
  */
  GEOSGetNumCoordinates(g: number): number;
  /**
  * Returns the number of coordinates in a GEOS geometry object (thread-safe version).
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry object.
  * @returns number - The number of coordinates, or -1 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a7b4e0c2f9d3f1a6d4b9a9c1d5e0f3a2e
  */
  GEOSGetNumCoordinates_r(handle: number, g: number): number;
  /**
  * Normalizes a GEOS geometry object in place, so that it has a canonical form.
* A normalized geometry has points in a LineString deduplicated, members of a collection sorted by type,
* points in a polygon's rings or a single LinearRing start at the top-left-most point and go clockwise, etc.
* This can be useful for unit testing or geometry deduplication.
  * @param g - The GEOS geometry object to normalize.
  * @returns number - 0 on success, or -1 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a5a7f0c9b3f9a1d0f4e5c2a7b5b3e6a8d
  */
  GEOSNormalize(g: number): number;
  /**
  * Normalizes a GEOS geometry object in place, so that it has a canonical form (thread-safe version).
* A normalized geometry has points in a LineString deduplicated, members of a collection sorted by type,
* points in a polygon's rings or a single LinearRing start at the top-left-most point and go clockwise, etc.
* This can be useful for unit testing or geometry deduplication.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry object to normalize.
  * @returns number - 0 on success, or -1 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a5a7f0c9b3f9a1d0f4e5c2a7b5b3e6a8d
  */
  GEOSNormalize_r(handle: number, g: number): number;
  /**
  * Returns the number of interior rings in the polygon.
  * @param g - A pointer to a GEOSGeometry object of type Polygon.
  * @returns number - The number of interior rings in the polygon, or -1 on exception.
    */
  GEOSGetNumInteriorRings(g: number): number;
  /**
  * Returns the number of interior rings in the polygon (thread-safe version).
  * @param handle - A GEOS context handle created with initGEOS_r().
  * @param g - A pointer to a GEOSGeometry object of type Polygon.
  * @returns number - The number of interior rings in the polygon, or -1 on exception.
    */
  GEOSGetNumInteriorRings_r(handle: number, g: number): number;
  /**
  * Returns the number of geometries in a collection.
  * @param g - A pointer to a GEOSGeometry object of type GeometryCollection or its subclasses (MultiPoint, MultiLineString, MultiPolygon).
  * @returns number - The number of geometries in the collection, or -1 on exception.
    */
  GEOSGetNumGeometries(g: number): number;
  /**
  * Returns the number of geometries in a collection (thread-safe version).
  * @param handle - A GEOS context handle created with initGEOS_r().
  * @param g - A pointer to a GEOSGeometry object of type GeometryCollection or its subclasses (MultiPoint, MultiLineString, MultiPolygon).
  * @returns number - The number of geometries in the collection, or -1 on exception.
    */
  GEOSGetNumGeometries_r(handle: number, g: number): number;
  /**
  * Returns a geometry from a collection by index.
  * @param g - A pointer to a GEOSGeometry object of type GeometryCollection or its subclasses (MultiPoint, MultiLineString, MultiPolygon).
  * @param n - The index of the geometry to return, starting from 0.
  * @returns number - A pointer to a newly allocated GEOSGeometry object corresponding to the geometry at index n. The caller is responsible for destroying it. Returns NULL on exception.
    */
  GEOSGetGeometryN(g: number, n: number): number;
  /**
  * Return the Nth geometry of a collection.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The input geometry, must be a collection.
  * @param n - The index of the geometry to return, starting from 0.
  * @returns number - A geometry object. Caller must free with GEOSGeom_destroy().
NULL on exception or if n is out of range.
    */
  GEOSGetGeometryN_r(handle: number, g: number, n: number): number;
  /**
  * Return the Nth point of a LineString.
  * @param g - The input geometry, must be a LineString.
  * @param n - The index of the point to return, starting from 0.
  * @returns number - A Point geometry. Caller must free with GEOSGeom_destroy().
NULL on exception or if n is out of range.
    */
  GEOSGeomGetPointN(g: number, n: number): number;
  /**
  * Return the Nth point of a LineString (thread-safe version).
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The input geometry, must be a LineString.
  * @param n - The index of the point to return, starting from 0.
  * @returns number - A Point geometry. Caller must free with GEOSGeom_destroy().
NULL on exception or if n is out of range.
    */
  GEOSGeomGetPointN_r(handle: number, g: number, n: number): number;
  /**
  * Return the first point of a LineString.
  * @param g - The input geometry, must be a LineString.
  * @returns number - A Point geometry. Caller must free with GEOSGeom_destroy().
NULL on exception.
    */
  GEOSGeomGetStartPoint(g: number): number;
  /**
  * Returns the first point of a geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry pointer.
  * @returns number - The GEOS geometry pointer of the start point, or null if the input is not a lineal geometry.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a9b7a6c0f5f8a2f2d0e3a4b9e1c3f8d6c|GEOSGeomGetStartPoint_r}
  */
  GEOSGeomGetStartPoint_r(handle: number, g: number): number;
  /**
  * Returns the endpoint of a LineString geometry.
  * @param geom - A pointer to a GEOSGeometry object.
  * @returns number - A pointer to a newly allocated GEOSGeometry object representing the endpoint, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9c6f7f5d3b9a0f1a0d4c1a4e2a0b6e8b
  */
  GEOSGeomGetEndPoint(geom: number): number;
  /**
  * Returns the endpoint of a LineString geometry.
  * @param handle - A GEOS context handle.
  * @param geom - A pointer to a GEOSGeometry object.
  * @returns number - A pointer to a newly allocated GEOSGeometry object representing the endpoint, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9c6f7f5d3b9a0f1a0d4c1a4e2a0b6e8b
  */
  GEOSGeomGetEndPoint_r(handle: number, geom: number): number;
  /**
  * Tests whether the input geometry is closed. A closed geometry is a linestring or multilinestring with the start and end points being the same.
* Returns 1 if the geometry is closed, 0 otherwise.
  * @param geom - A pointer to a GEOSGeometry object.
  * @returns number - 1 or 0, or -1 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a5d3f7f2e7d5e7b2c3d9a3c9f2e8a6b4b
  */
  GEOSisClosed(geom: number): number;
  /**
  * Tests whether the input geometry is closed. A closed geometry is a linestring or multilinestring with the start and end points being the same.
* Returns 1 if the geometry is closed, 0 otherwise.
  * @param handle - A GEOS context handle.
  * @param geom - A pointer to a GEOSGeometry object.
  * @returns number - 1 or 0, or -1 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a5d3f7f2e7d5e7b2c3d9a3c9f2e8a6b4b
  */
  GEOSisClosed_r(handle: number, geom: number): number;
  /**
  * Computes the length of a geometry.
  * @param geom - A pointer to a GEOSGeometry object.
  * @param length - A pointer to a double value that will store the result.
  * @returns number - 1 on success, 0 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a4e9f6a5bbec6edea7adba6ffac72fe48
  */
  GEOSGeomGetLength(geom: number, length: number): number;
  /**
  * Computes the length of a geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry pointer.
  * @param length - The output parameter for the length value.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSGeomGetLength_r(handle: number, g: number, length: number): number;
  /**
  * Returns the number of points in a geometry.
  * @param g - The GEOS geometry pointer.
  * @returns number - The number of points, or -1 on exception.
    */
  GEOSGeomGetNumPoints(g: number): number;
  /**
  * Returns the number of points in a geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry pointer.
  * @returns number - The number of points, or -1 on exception.
    */
  GEOSGeomGetNumPoints_r(handle: number, g: number): number;
  /**
  * Returns the X coordinate of a point geometry.
  * @param g - The GEOS geometry pointer. Must be a point geometry.
  * @param x - The output parameter for the X coordinate value.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSGeomGetX(g: number, x: number): number;
  /**
  * Returns the X coordinate of a point geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry pointer. Must be a point geometry.
  * @param x - The output parameter for the X coordinate value.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSGeomGetX_r(handle: number, g: number, x: number): number;
  /**
  * Returns the Y coordinate of a point geometry.
  * @param g - A pointer to a GEOSGeometry object of type Point.
  * @param value - A pointer to a double where the result will be stored.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSGeomGetY(g: number, value: number): number;
  /**
  * Returns the Y coordinate of a point geometry in a thread-safe manner.
  * @param handle - A GEOS context handle created by GEOS_init_r().
  * @param g - A pointer to a GEOSGeometry object of type Point.
  * @param value - A pointer to a double where the result will be stored.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSGeomGetY_r(handle: number, g: number, value: number): number;
  /**
  * Returns the Z coordinate of a point geometry.
  * @param g - A pointer to a GEOSGeometry object of type Point.
  * @param value - A pointer to a double where the result will be stored.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSGeomGetZ(g: number, value: number): number;
  /**
  * Returns the Z coordinate of a point geometry in a thread-safe manner.
  * @param handle - A GEOS context handle created by GEOS_init_r().
  * @param g - A pointer to a GEOSGeometry object of type Point.
  * @param value - A pointer to a double where the result will be stored.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSGeomGetZ_r(handle: number, g: number, value: number): number;
  /**
  * Returns the exterior ring of a polygon geometry as a linear ring.
  * @param g - A pointer to a GEOSGeometry object of type Polygon.
  * @returns number - A pointer to a newly allocated GEOSGeometry object of type LinearRing, or NULL on exception. The caller is responsible for destroying it with GEOSGeom_destroy().
    */
  GEOSGetExteriorRing(g: number): number;
  /**
  * Returns the exterior ring of a polygon geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The polygon geometry.
  * @returns number - - The exterior ring geometry, or null if an error occurred.
  * @see https://libgeos.org/doxygen/geos__c_8h_source.html
  */
  GEOSGetExteriorRing_r(handle: number, g: number): number;
  /**
  * Returns the Nth interior ring of a polygon geometry.
* Note: Returned object is a pointer to internal storage: it must NOT be destroyed directly.
  * @param g - The polygon geometry.
  * @param n - The index of the interior ring, starting from 0.
  * @returns number - - The interior ring geometry, or null if an error occurred or the index is out of range.
  * @see https://libgeos.org/doxygen/geos__c_8h.html
  */
  GEOSGetInteriorRingN(g: number, n: number): number;
  /**
  * Returns the Nth interior ring of a polygon geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The polygon geometry.
  * @param n - The index of the interior ring, starting from 0.
  * @returns number - - The interior ring geometry, or null if an error occurred or the index is out of range.
  * @see https://libgeos.org/doxygen/geos__c_8h_source.html
  */
  GEOSGetInteriorRingN_r(handle: number, g: number, n: number): number;
  /**
  * Computes the centroid point of a geometry.
  * @param g - The geometry to use.
  * @returns number - - The centroid point geometry, or null if an error occurred or the input is empty.
  * @see https://libgeos.org/doxygen/classgeos_1_1algorithm_1_1Centroid.html
  */
  GEOSGetCentroid(g: number): number;
  /**
  * Computes the centroid point of a geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The geometry to use.
  * @returns number - - The centroid point geometry, or null if an error occurred or the input is empty.
  * @see https://libgeos.org/doxygen/classgeos_1_1algorithm_1_1Centroid.html
  */
  GEOSGetCentroid_r(handle: number, g: number): number;
  /**
  * Returns a geometry which represents the "minimum bounding circle", the smallest circle that contains the input.
  * @param geom - A pointer to a GEOS geometry object.
  * @param radius - A pointer to a double that will store the radius of the circle.
  * @param center - A pointer to a GEOS geometry object that will store the center point of the circle.
  * @returns number - A pointer to a GEOS geometry object representing the minimum bounding circle, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#af09ef325324bec3c1ccb981db52eec4f
  */
  GEOSMinimumBoundingCircle(geom: number, radius: number, center: number): number;
  /**
  * Returns a geometry which represents the "minimum bounding circle", the smallest circle that contains the input.
  * @param handle - A pointer to a GEOS context handle.
  * @param geom - A pointer to a GEOS geometry object.
  * @param radius - A pointer to a double that will store the radius of the circle.
  * @param center - A pointer to a GEOS geometry object that will store the center point of the circle.
  * @returns number - A pointer to a GEOS geometry object representing the minimum bounding circle, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#af09ef325324bec3c1ccb981db52eec4f
  */
  GEOSMinimumBoundingCircle_r(handle: number, geom: number, radius: number, center: number): number;
  /**
  * Creates a geometry collection of a specified type from an array of geometries.
  * @param type - An integer representing the geometry type, such as 1 for POINT, 4 for MULTIPOINT, etc.
  * @param geoms - A pointer to a list of GEOS geometry pointers.
  * @param ngeoms - The number of geometries in the array.
  * @returns number - A pointer to a GEOS geometry object representing the collection, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a871ccb7efc6dd63162b3547fcd76c638
  */
  GEOSGeom_createCollection(type: number, geoms: number, ngeoms: number): number;
  /**
  * Creates a geometry collection of a specified type from an array of geometries using a thread-safe context handle.
  * @param handle - A pointer to a GEOS context handle.
  * @param type - An integer representing the geometry type, such as 1 for POINT, 4 for MULTIPOINT, etc.
  * @param geoms - A pointer to an array of pointers to GEOS geometry objects.
  * @param ngeoms - The number of geometries in the array.
  * @returns number - A pointer to a GEOS geometry object representing the collection, or NULL on error.
  * @see https://geos.osgeo.org/doxygen/classgeos_1_1geom_1_1GeometryFactory.html#a0c8f6a2a9f5b7a3d7e5c7b0f8b6f9c0e
  */
  GEOSGeom_createCollection_r(handle: number, type: number, geoms: number, ngeoms: number): number;
  /**
  * Release the sub-geometries of a collection for management.
* by the caller. The input collection remains as an empty collection,
* that the caller is responsible for destroying. The output geometries
* are also the responsibility of the caller, as is the containing array,
* which must be freed with GEOSFree().
  * @param collection - A pointer to a GEOS geometry collection object.
  * @param ngeoms - A pointer to a variable that will be filled with the size of the output array.
  * @returns number - A newly allocated array of GEOSGeometry pointers.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9a7624d6406f501856cfb1e60a30f4fa
  */
  GEOSGeom_releaseCollection(collection: number, ngeoms: number): number;
  /**
  * Release the sub-geometries of a collection for management.
* by the caller. The input collection remains as an empty collection,
* that the caller is responsible for destroying. The output geometries
* are also the responsibility of the caller, as is the containing array,
* which must be freed with GEOSFree().
  * @param handle - A pointer to a GEOS context handle.
  * @param collection - A pointer to a GEOS geometry collection object.
  * @param ngeoms - A pointer to a variable that will be filled with the size of the output array.
  * @returns number - A newly allocated array of GEOSGeometry pointers.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9a7624d6406f501856cfb1e60a30f4fa
  */
  GEOSGeom_releaseCollection_r(handle: number, collection: number, ngeoms: number): number;
  /**
  * Polygonizes a set of Geometries which contain linework that represents the edges of a planar graph.
* 
* All types of Geometry are accepted as input; the constituent linework is extracted as the edges to be polygonized.
* 
* The edges must be correctly noded; that is, they must only meet at their endpoints and not overlap anywhere. If your edges are not already noded, run them through GEOSUnaryUnion() first. Polygonization will accept incorrectly noded input but will not form polygons from non-noded edges, and reports them as errors.
* 
* The Polygonizer reports the following kinds of errors:
* 
* - Dangles - edges which have one or both ends which are not incident on another edge endpoint
* - Cut Edges - edges which are connected at both ends but which do not form part of a polygon
* - Invalid Ring Lines - edges which form rings which are invalid (e.g. the component lines contain a self-intersection)
* Errors are reported to output parameters "cuts", "dangles" and "invalid" (if not-null). Formed polygons are returned as a collection. NULL is returned on exception. All returned geometries must be destroyed by caller.
* 
* The GEOSPolygonize_valid() variant allows extracting only polygons which form a valid polygonal result. The set of extracted polygons is guaranteed to be edge-disjoint. This is useful when it is known that the input lines form a valid polygonal geometry (which may include holes or nested polygons).
  * @param geoms - A pointer to an array of pointers to GEOS geometry objects representing line segments.
  * @param ngeoms - The number of geometries in the array.
  * @returns number - A pointer to a GEOS geometry object representing the polygonized geometry, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9d98e448d3b846d591c726d1c0000d25
  */
  GEOSPolygonize(geoms: number, ngeoms: number): number;
  /**
  * Polygonizes a set of Geometries which contain linework that represents the edges of a planar graph.
* 
* All types of Geometry are accepted as input; the constituent linework is extracted as the edges to be polygonized.
* 
* The edges must be correctly noded; that is, they must only meet at their endpoints and not overlap anywhere. If your edges are not already noded, run them through GEOSUnaryUnion() first. Polygonization will accept incorrectly noded input but will not form polygons from non-noded edges, and reports them as errors.
* 
* The Polygonizer reports the following kinds of errors:
* 
* - Dangles - edges which have one or both ends which are not incident on another edge endpoint
* - Cut Edges - edges which are connected at both ends but which do not form part of a polygon
* - Invalid Ring Lines - edges which form rings which are invalid (e.g. the component lines contain a self-intersection)
* Errors are reported to output parameters "cuts", "dangles" and "invalid" (if not-null). Formed polygons are returned as a collection. NULL is returned on exception. All returned geometries must be destroyed by caller.
* 
* The GEOSPolygonize_valid() variant allows extracting only polygons which form a valid polygonal result. The set of extracted polygons is guaranteed to be edge-disjoint. This is useful when it is known that the input lines form a valid polygonal geometry (which may include holes or nested polygons).
  * @param handle - A pointer to the GEOS context handle.
  * @param geoms - A pointer to an array of Geometrys containing the linework to polygonize.
  * @param ngeoms - The number of Geometrys in the array.
  * @returns number - A GeometryCollection containing the polygons formed by the polygonization.
    */
  GEOSPolygonize_r(handle: number, geoms: number, ngeoms: number): number;
  /**
  * Polygonizes a set of valid Geometrys which contain linework that
* represents the edges of a planar graph.
* All types of Geometry are accepted as input; the constituent
* linework is extracted as the edges to be polygonized.
* The processed edges must be correctly noded; that is, they must only meet
* at their endpoints. The Polygonizer will run on incorrectly noded input
* but will not form polygons from non-noded edges,
* and will report them as errors.
  * @param geoms - A pointer to an array of valid Geometrys containing the linework to polygonize.
  * @param ngeoms - The number of valid Geometrys in the array.
  * @returns number - A GeometryCollection containing the polygons formed by the polygonization.
    */
  GEOSPolygonize_valid(geoms: number, ngeoms: number): number;
  /**
  * Polygonizes a set of valid Geometrys which contain linework that
* represents the edges of a planar graph, using a GEOS context handle.
* All types of Geometry are accepted as input; the constituent
* linework is extracted as the edges to be polygonized.
* The processed edges must be correctly noded; that is, they must only meet
* at their endpoints. The Polygonizer will run on incorrectly noded input
* but will not form polygons from non-noded edges,
* and will report them as errors.
  * @param handle - A pointer to the GEOS context handle.
  * @param geoms - An array of valid Geometrys containing the linework to polygonize.
  * @param ngeoms - The number of valid Geometrys in the array.
  * @returns string - A GeometryCollection containing the polygons formed by the polygonization.
    */
  GEOSPolygonize_valid_r(handle: number, geoms: string, ngeoms: number): string;
  /**
  * Gets the list of line segments forming the boundary between
* inside and outside portions of a set of Geometrys which contain linework that
* represents the edges of a planar graph.
  * @param geoms - An array of Geometrys containing the linework to process.
  * @param ngeoms - The number of Geometrys in the array.
  * @returns string - A MultiLineString containing the boundary segments.
    */
  GEOSPolygonizer_getCutEdges(geoms: string, ngeoms: number): string;
  /**
  * Gets the list of line segments forming the boundary between
* inside and outside portions of a set of Geometrys which contain linework that
* represents the edges of a planar graph, using a GEOS context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param geoms - An array of Geometrys containing the linework to process.
  * @param ngeoms - The number of Geometrys in the array.
  * @returns string - A MultiLineString containing the boundary segments.
    */
  GEOSPolygonizer_getCutEdges_r(handle: number, geoms: string, ngeoms: number): string;
  /**
  * Polygonizes a set of Geometrys which contain linework that
* represents the edges of a planar graph.
* All types of Geometry are accepted as input; the constituent
* linework is extracted as the edges to be polygonized.
* The processed edges must be correctly noded; that is, they must only meet
* at their endpoints. The Polygonizer will run on incorrectly noded input
* but will not form polygons from non-noded edges, and will report them as errors.
* 
* The Polygonizer reports the follow kinds of errors:
* 
* - Dangles - edges which have one or both ends which are not incident on another edge endpoint
* - Cut Edges - edges which are connected at both ends but which do not form part of a polygon
* - Invalid Ring Lines - edges which form rings which are invalid
*   (e.g. the component lines contain a self-intersection)
* 
* The returned geometry is a collection containing valid polygonal geometry.
  * @param input - a pointer to an array of Geometrys
  * @param cuts - if non-null, will be set to a pointer to a collection containing the cut edges found during polygonization
  * @param dangles - if non-null, will be set to a pointer to a collection containing the dangle edges found during polygonization
  * @param invalidRings - if non-null, will be set to a pointer to a collection containing the invalid ring lines found during polygonization
  * @returns number - a pointer to a collection containing the polygons formed by the polygonization
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a0f0a7f5c9a3b6d7f9b4d1c8e2a0d4f5a
  */
  GEOSPolygonize_full(input: number, cuts: number, dangles: number, invalidRings: number): number;
  /**
  * Polygonizes a set of Geometrys which contain linework that
* represents the edges of a planar graph.
* All types of Geometry are accepted as input; the constituent
* linework is extracted as the edges to be polygonized.
* The processed edges must be correctly noded; that is, they must only meet
* at their endpoints. The Polygonizer will run on incorrectly noded input
* but will not form polygons from non-noded edges, and will report them as errors.
* 
* The Polygonizer reports the follow kinds of errors:
* 
* - Dangles - edges which have one or both ends which are not incident on another edge endpoint
* - Cut Edges - edges which are connected at both ends but which do not form part of a polygon
* - Invalid Ring Lines - edges which form rings which are invalid
*   (e.g. the component lines contain a self-intersection)
* 
* The returned geometry is a collection containing valid polygonal geometry.
  * @param handle - a pointer to an opaque GEOS context handle object
  * @param input - a pointer to an array of Geometrys
  * @param cuts - if non-null, will be set to a pointer to a collection containing the cut edges found during polygonization
  * @param dangles - if non-null, will be set to a pointer to a collection containing the dangle edges found during polygonization
  * @param invalidRings - if non-null, will be set to a pointer to a collection containing the invalid ring lines found during polygonization
  * @returns number - a pointer to a collection containing the polygons formed by the polygonization
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a1b2a8f7c3b2b6d9a3a0b4c6e1d0f9e1f
  */
  GEOSPolygonize_full_r(handle: number, input: number, cuts: number, dangles: number, invalidRings: number): number;
  /**
  * Creates an areal geometry formed by constituent linework of given geometry.
* 
* Returns null on exception or empty result.
  * @param g - a pointer to an input Geometry (linestring or multilinestring)
  * @returns number - a pointer to a Geometry (polygon or multipolygon) representing the input's areal geometry
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a0a8f4f9d7e3b9c1a2b6a0f1b5c4a7d0f
  */
  GEOSBuildArea(g: number): number;
  /**
  * Creates an areal geometry formed by constituent linework of given geometry.
* 
* Returns null on exception or empty result.
  * @param handle - a pointer to an opaque GEOS context handle object
  * @param g - a pointer to an input Geometry (linestring or multilinestring)
  * @returns number - a pointer to a Geometry (polygon or multipolygon) representing the input's areal geometry
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3e6d7e2a4f9c5b6b2e0d1f3e8a9c1b3d
  */
  GEOSBuildArea_r(handle: number, g: number): number;
  /**
  * Returns a geometry which is valid according to the GEOS
* validity rules, and preserves as much as possible of the input geometry's extent,
* dimension, and structure.
* 
* The returned geometry may be a copy of the input, or a new geometry with different structure.
* 
* The returned geometry will be valid, or null if an exception occurs.
  * @param g - a pointer to an input Geometry
  * @returns number - a pointer to a valid Geometry, or null on exception
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a0f0a7f5c9a3b6d7f9b4d1c8e2a0d4f5a
  */
  GEOSMakeValid(g: number): number;
  /**
  * Returns a geometry that is valid according to the GEOS library.
  * @param handle - The GEOS context pointer handle.
  * @param g - The input geometry pointer.
  * @returns number - A pointer to a valid geometry, or NULL on exception.
  * @see https://libgeos.org/doxygen/classgeos_1_1operation_1_1valid_1_1MakeValidOp.html
  */
  GEOSMakeValid_r(handle: number, g: number): number;
  /**
  * Merges a collection of linear components into maximal-length linestrings.
  * @param g - The input geometry pointer, which must be a collection of linestrings or multilinestrings.
  * @returns number - A pointer to a multilinestring containing the merged linestrings, or NULL on exception.
  * @see https://libgeos.org/doxygen/classgeos_1_1operation_1_1linemerge_1_1LineMerger.html
  */
  GEOSLineMerge(g: number): number;
  /**
  * Merges a collection of linear components into maximal-length linestrings, using a GEOS context handle.
  * @param handle - The GEOS context pointer handle.
  * @param g - The input geometry pointer, which must be a collection of linestrings or multilinestrings.
  * @returns number - A pointer to a multilinestring containing the merged linestrings, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h_source.html
  */
  GEOSLineMerge_r(handle: number, g: number): number;
  /**
  * Reverses the order of the coordinates in a geometry.
  * @param g - The input geometry pointer, which must be a point, linestring, linearring or polygon.
  * @returns number - A pointer to a geometry with reversed coordinates, or NULL on exception.
  * @see https://libgeos.org/doxygen/classgeos_1_1geom_1_1Geometry.html
  */
  GEOSReverse(g: number): number;
  /**
  * Reverses the order of the coordinates in a geometry, using a GEOS context handle.
  * @param handle - The GEOS context pointer handle.
  * @param g - The input geometry pointer, which must be a point, linestring, linearring or polygon.
  * @returns number - A pointer to a geometry with reversed coordinates, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h_source.html
  */
  GEOSReverse_r(handle: number, g: number): number;
  /**
  * Returns the Spatial Reference System ID (SRID) of a geometry.
  * @param g - A pointer to a GEOSGeometry object.
  * @returns number - The SRID of the geometry, or 0 if not set.
    */
  GEOSGetSRID(g: number): number;
  /**
  * Returns the Spatial Reference System ID (SRID) of a geometry in a thread-safe manner.
  * @param handle - A pointer to a GEOSContextHandle_t object.
  * @param g - A pointer to a GEOSGeometry object.
  * @returns number - The SRID of the geometry, or 0 if not set.
    */
  GEOSGetSRID_r(handle: number, g: number): number;
  /**
  * Sets the Spatial Reference System ID (SRID) of a geometry.
  * @param g - A pointer to a GEOSGeometry object.
  * @param SRID - The SRID value to set.
  * @returns void - undefined
    */
  GEOSSetSRID(g: number, SRID: number): void;
  /**
  * Sets the Spatial Reference System ID (SRID) of a geometry in a thread-safe manner.
  * @param handle - A pointer to a GEOSContextHandle_t object.
  * @param g - A pointer to a GEOSGeometry object.
  * @param SRID - The SRID value to set.
  * @returns void - undefined
    */
  GEOSSetSRID_r(handle: number, g: number, SRID: number): void;
  /**
  * Returns the user data associated with a geometry as a string.
  * @param g - A pointer to a GEOSGeometry object.
  * @returns string - The user data of the geometry, or null if not set or not a string.
    */
  GEOSGeom_getUserData(g: string): string;
  /**
  * Get the user data pointer associated with a geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry object.
  * @returns string - The user data pointer, or null if none is set.
    */
  GEOSGeom_getUserData_r(handle: number, g: string): string;
  /**
  * Set the user data pointer associated with a geometry.
* The user is responsible for freeing the memory of the user data pointer.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry object.
  * @param userData - The user data pointer to be set.
  * @returns void - undefined
    */
  GEOSGeom_setUserData(handle: number, g: number, userData: string): void;
  /**
  * Set the user data pointer associated with a geometry.
* The user is responsible for freeing the memory of the user data pointer.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry object.
  * @param userData - The user data pointer to be set.
  * @returns void - undefined
    */
  GEOSGeom_setUserData_r(handle: number, g: number, userData: string): void;
  /**
  * Check if a geometry has a Z coordinate dimension.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry object.
  * @returns number - 1 if the geometry has Z dimension, 0 otherwise, or -1 on error.
    */
  GEOSHasZ(handle: number, g: number): number;
  /**
  * Check if a geometry has a Z coordinate dimension.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry object.
  * @returns number - 1 if the geometry has Z dimension, 0 otherwise, or -1 on error.
    */
  GEOSHasZ_r(handle: number, g: number): number;
  /**
  * Get the output dimension of WKB geometries.
  * @returns number - The output dimension (2 or 3).
    */
  GEOS_getWKBOutputDims(): number;
  /**
  * Get the output dimension of WKB geometries in a given context.
  * @param handle - A pointer to the GEOS context handle.
  * @returns number - The output dimension (2 or 3).
    */
  GEOS_getWKBOutputDims_r(handle: number): number;
  /**
  * Set the output dimension of WKB geometries.
  * @param newDims - The new output dimension (2 or 3).
  * @returns number - The previous output dimension.
    */
  GEOS_setWKBOutputDims(newDims: number): number;
  /**
  * Set the output dimension of WKB geometries in a given context.
  * @param handle - A pointer to the GEOS context handle.
  * @param newDims - The new output dimension (2 or 3).
  * @returns number - The previous output dimension.
    */
  GEOS_setWKBOutputDims_r(handle: number, newDims: number): number;
  /**
  * Get the byte order of WKB geometries.
  * @returns number - The byte order (0 for XDR, 1 for NDR).
    */
  GEOS_getWKBByteOrder(): number;
  /**
  * Get the byte order used for WKB output.
  * @param handle - A pointer to the GEOS context handle.
  * @returns number - The byte order (GEOS_WKB_XDR or GEOS_WKB_NDR) or -1 on error.
    */
  GEOS_getWKBByteOrder_r(handle: number): number;
  /**
  * Set the byte order used for WKB output.
  * @param byteOrder - The byte order (GEOS_WKB_XDR or GEOS_WKB_NDR).
  * @returns number - 1 on success, 0 on error.
    */
  GEOS_setWKBByteOrder(byteOrder: number): number;
  /**
  * Set the byte order used for WKB output in a given context.
  * @param handle - A pointer to the GEOS context handle.
  * @param byteOrder - The byte order (GEOS_WKB_XDR or GEOS_WKB_NDR).
  * @returns number - 1 on success, 0 on error.
    */
  GEOS_setWKBByteOrder_r(handle: number, byteOrder: number): number;
  /**
  * Create a coordinate sequence of a given size and dimension.
  * @param size - The number of coordinates in the sequence.
  * @param dims - The dimension of each coordinate (2 or 3).
  * @returns number - A pointer to the created coordinate sequence or NULL on error.
    */
  GEOSCoordSeq_create(size: number, dims: number): number;
  /**
  * Create a coordinate sequence of a given size and dimension in a given context.
  * @param handle - A pointer to the GEOS context handle.
  * @param size - The number of coordinates in the sequence.
  * @param dims - The dimension of each coordinate (2 or 3).
  * @returns number - A pointer to the created coordinate sequence or NULL on error.
    */
  GEOSCoordSeq_create_r(handle: number, size: number, dims: number): number;
  /**
  * Set an ordinate value in a coordinate sequence.
  * @param s - The coordinate sequence handle.
  * @param idx - The index of the coordinate to alter, zero based.
  * @param dim - The dimension to set (0=X, 1=Y, 2=Z, 3=M).
  * @param val - The value to set the ordinate to.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_setOrdinate(s: number, idx: number, dim: number, val: number): number;
  /**
  * Set an ordinate value in a coordinate sequence (thread-safe version).
  * @param handle - A pointer to the GEOS context handle.
  * @param s - The coordinate sequence handle.
  * @param idx - The index of the coordinate to alter, zero based.
  * @param dim - The dimension to set (0=X, 1=Y, 2=Z, 3=M).
  * @param val - The value to set the ordinate to.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_setOrdinate_r(handle: number, s: number, idx: number, dim: number, val: number): number;
  /**
  * Set X ordinate values in a coordinate sequence.
  * @param s - The coordinate sequence handle.
  * @param idx - The index of the coordinate to alter, zero based.
  * @param val - The value to set the ordinate to.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_setX(s: number, idx: number, val: number): number;
  /**
  * Set Y ordinate values in a coordinate sequence.
  * @param s - The coordinate sequence handle.
  * @param idx - The index of the coordinate to alter, zero based.
  * @param val - The value to set the ordinate to.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_setY(s: number, idx: number, val: number): number;
  /**
  * Set Z ordinate values in a coordinate sequence.
  * @param s - The coordinate sequence handle.
  * @param idx - The index of the coordinate to alter, zero based.
  * @param val - The value to set the ordinate to.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_setZ(s: number, idx: number, val: number): number;
  /**
  * Set X and Y ordinate values in a coordinate sequence.
  * @param s - The coordinate sequence pointer
  * @param idx - The index of the coordinate to alter, zero based
  * @param x - The value to set the X ordinate to
  * @param y - The value to set the Y ordinate to
  * @returns number - 0 on exception, 1 otherwise
    */
  GEOSCoordSeq_setXY(s: number, idx: number, x: number, y: number): number;
  /**
  * Set X and Y ordinate values in a coordinate sequence (reentrant).
  * @param handle - The context handle
pointer  * @param {number} s - The coordinate sequence pointer
  * @param idx - The index of the coordinate to alter, zero based
  * @param x - The value to set the X ordinate to
  * @param y - The value to set the Y ordinate to
  * @returns number - 0 on exception, 1 otherwise
    */
  GEOSCoordSeq_setXY_r(handle: number, idx: number, x: number, y: number): number;
  /**
  * Set X, Y and Z ordinate values in a coordinate sequence.
  * @param s - The coordinate sequence pointer
  * @param idx - The index of the coordinate to alter, zero based
  * @param x - The value to set the X ordinate to
  * @param y - The value to set the Y ordinate to
  * @param z - The value to set the Z ordinate to
  * @returns number - 0 on exception, 1 otherwise
    */
  GEOSCoordSeq_setXYZ(s: number, idx: number, x: number, y: number, z: number): number;
  /**
  * Set X, Y and Z ordinate values in a coordinate sequence (reentrant).
  * @param handle - The context handle pointer
  * @param s - The coordinate sequence pointer
  * @param idx - The index of the coordinate to alter, zero based
  * @param x - The value to set the X ordinate to
  * @param y - The value to set the Y ordinate to
  * @param z - The value to set the Z ordinate to
  * @returns number - 0 on exception, 1 otherwise
    */
  GEOSCoordSeq_setXYZ_r(handle: number, s: number, idx: number, x: number, y: number, z: number): number;
  /**
  * Clone a coordinate sequence.
  * @param s - The coordinate sequence to clone pointer
  * @returns number - A pointer to a new coordinate sequence object or null on exception.
    */
  GEOSCoordSeq_clone(s: number): number;
  /**
  * Clone a coordinate sequence.
  * @param handle - A pointer to the GEOS context handle.
  * @param s - The pointer to the coordinate sequence to clone.
  * @returns number - The pointer to the cloned coordinate sequence, or null on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a4f2a6b1f8e9a1c0e3f9c6e7d5a3f6b5a
  */
  GEOSCoordSeq_clone_r(handle: number, s: number): number;
  /**
  * Get an ordinate value from a coordinate sequence.
  * @param s - The pointer to the coordinate sequence.
  * @param idx - The index of the coordinate to get the ordinate from.
  * @param dim - The dimension of the ordinate to get (0 for X, 1 for Y, 2 for Z, 3 for M).
  * @returns number - The ordinate value, or NaN on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9b7d0e4b5c8f2d7e4a9b0d6c3a9f1a7c
  */
  GEOSCoordSeq_getOrdinate(s: number, idx: number, dim: number): number;
  /**
  * Get an ordinate value from a coordinate sequence in a thread-safe way.
  * @param handle - A pointer to the GEOS context handle.
  * @param s - The pointer to the coordinate sequence.
  * @param idx - The index of the coordinate to get the ordinate from.
  * @param dim - The dimension of the ordinate to get (0 for X, 1 for Y, 2 for Z, 3 for M).
  * @param val - The pointer to a double variable to store the ordinate value.
  * @returns number - 1 on success, 0 on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a4f2a6b1f8e9a1c0e3f9c6e7d5a3f6b5a
  */
  GEOSCoordSeq_getOrdinate_r(handle: number, s: number, idx: number, dim: number, val: number): number;
  /**
  * Get the X ordinate value from a coordinate sequence.
  * @param s - The pointer to the coordinate sequence.
  * @param idx - The index of the coordinate to get the X ordinate from.
  * @param x - The pointer to a double variable to store the X ordinate value.
  * @returns number - 1 on success, 0 on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9b7d0e4b5c8f2d7e4a9b0d6c3a9f1a7c
  */
  GEOSCoordSeq_getX(s: number, idx: number, x: number): number;
  /**
  * Get the Y ordinate value from a coordinate sequence.
  * @param s - The pointer to the coordinate sequence.
  * @param idx - The index of the coordinate to get the Y ordinate from.
  * @param y - The pointer to a double variable to store the Y ordinate value.
  * @returns number - 1 on success, 0 on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9b7d0e4b5c8f2d7e4a9b0d6c3a9f1a7c
  */
  GEOSCoordSeq_getY(s: number, idx: number, y: number): number;
  /**
  * Get the Z ordinate value of a coordinate in a sequence.
  * @param s - The coordinate sequence handle.
  * @param idx - The index of the coordinate to get, zero based.
  * @param val - A pointer to a double where the ordinate value will be stored.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getZ(s: number, idx: number, val: number): number;
  /**
  * Get the X and Y ordinate values of a coordinate in a sequence.
  * @param s - The coordinate sequence handle.
  * @param idx - The index of the coordinate to get, zero based.
  * @param x - A pointer to a double where the X ordinate value will be stored.
  * @param y - A pointer to a double where the Y ordinate value will be stored.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getXY(s: number, idx: number, x: number, y: number): number;
  /**
  * Get the X and Y ordinate values of a coordinate in a sequence using a context handle.
  * @param handle - The context handle.
  * @param s - The coordinate sequence handle.
  * @param idx - The index of the coordinate to get, zero based.
  * @param x - A pointer to a double where the X ordinate value will be stored.
  * @param y - A pointer to a double where the Y ordinate value will be stored.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getXY_r(handle: number, s: number, idx: number, x: number, y: number): number;
  /**
  * Get the X, Y and Z ordinate values of a coordinate in a sequence.
  * @param s - The coordinate sequence handle.
  * @param idx - The index of the coordinate to get, zero based.
  * @param x - A pointer to a double where the X ordinate value will be stored.
  * @param y - A pointer to a double where the Y ordinate value will be stored.
  * @param z - A pointer to a double where the Z ordinate value will be stored.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getXYZ(s: number, idx: number, x: number, y: number, z: number): number;
  /**
  * Get the X, Y and Z ordinate values of a coordinate in a sequence using a context handle.
  * @param handle - The context handle.
  * @param s - The coordinate sequence handle.
  * @param idx - The index of the coordinate to get, zero based.
  * @param x - A pointer to a double where the X ordinate value will be stored.
  * @param y - A pointer to a double where the Y ordinate value will be stored.
  * @param z - A pointer to a double where the Z ordinate value will be stored.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getXYZ_r(handle: number, s: number, idx: number, x: number, y: number, z: number): number;
  /**
  * Get the size of a coordinate sequence.
  * @param s - A pointer to a GEOSCoordSequence object.
  * @param size - A pointer to an unsigned int to store the size.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getSize(s: number, size: number): number;
  /**
  * Get the size of a coordinate sequence (thread-safe).
  * @param handle - A GEOS context handle.
  * @param s - A pointer to a GEOSCoordSequence object.
  * @param size - A pointer to an unsigned int to store the size.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getSize_r(handle: number, s: number, size: number): number;
  /**
  * Get the dimensions of a coordinate sequence.
  * @param s - A pointer to a GEOSCoordSequence object.
  * @param dims - A pointer to an unsigned int to store the dimensions.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getDimensions(s: number, dims: number): number;
  /**
  * Get the dimensions of a coordinate sequence (thread-safe).
  * @param handle - A GEOS context handle.
  * @param s - A pointer to a GEOSCoordSequence object.
  * @param dims - A pointer to an unsigned int to store the dimensions.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getDimensions_r(handle: number, s: number, dims: number): number;
  /**
  * Check the orientation of a coordinate sequence. Closure of the sequence is assumed. Invalid (collapsed) or short (fewer than 4 points) sequences return false.
  * @param s - A pointer to a GEOSCoordSequence object.
  * @param is_ccw - A pointer to a char to store the result (1 for counter-clockwise, 0 for clockwise).
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_isCCW(s: number, is_ccw: number): number;
  /**
  * Checks whether a coordinate sequence forms a counter-clockwise ring.
  * @param handle - A pointer to the GEOS context handle.
  * @param s - The coordinate sequence to check.
  * @param is_ccw - A pointer to a char variable to store the result (1 if CCW, 0 otherwise).
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_isCCW_r(handle: number, s: number, is_ccw: number): number;
  /**
  * Destroys a coordinate sequence.
  * @param s - The coordinate sequence to destroy.
  * @returns void - undefined
    */
  GEOSCoordSeq_destroy(s: number): void;
  /**
  * Destroys a coordinate sequence with a context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param s - The coordinate sequence to destroy.
  * @returns void - undefined
    */
  GEOSCoordSeq_destroy_r(handle: number, s: number): void;
  /**
  * Returns the coordinate sequence of a geometry.
  * @param g - The geometry to get the coordinate sequence from.
  * @returns number - A pointer to the coordinate sequence, or NULL on exception.
    */
  GEOSGeom_getCoordSeq(g: number): number;
  /**
  * Returns the coordinate sequence of a geometry with a context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The geometry to get the coordinate sequence from.
  * @returns number - A pointer to the coordinate sequence, or NULL on exception.
    */
  GEOSGeom_getCoordSeq_r(handle: number, g: number): number;
  /**
  * Creates a point geometry from a coordinate sequence.
  * @param s - The pointer to a coordinate sequence to use. Must have exactly one element.
  * @returns number - A pointer to a newly allocated point geometry. NULL on exception.
Caller is responsible for freeing with GEOSGeom_destroy().
    */
  GEOSGeom_createPoint(s: number): number;
  /**
  * Creates a point geometry from a coordinate sequence (reentrant).
  * @param handle - The context handle topointer  use.
  * @param s - The pointer to a coordinate sequence to use. Must have exactly one element.
  * @returns number - A pointer to a newly allocated point geometry. NULL on exception.
Caller is responsible for freeing with GEOSGeom_destroy_r().
    */
  GEOSGeom_createPoint_r(handle: number, s: number): number;
  /**
  * Creates a point geometry from a pair of coordinates.
  * @param x - The X coordinate
  * @param y - The Y coordinate
  * @returns number - A pointer to a newly allocated point geometry. NULL on exception.
Caller is responsible for freeing with GEOSGeom_destroy().
    */
  GEOSGeom_createPointFromXY(x: number, y: number): number;
  /**
  * Creates a point geometry from a pair of coordinates (reentrant).
  * @param handle - The context handle topointer  use.
  * @param x - The X coordinate
  * @param y - The Y coordinate
  * @returns number - A pointer to a newly allocated point geometry. NULL on exception.
Caller is responsible for freeing with GEOSGeom_destroy_r().
    */
  GEOSGeom_createPointFromXY_r(handle: number, x: number, y: number): number;
  /**
  * Creates a linear ring geometry from a coordinate sequence.
* A linear ring is a line string that is both closed and simple.
  * @param s - A pointer to the coordinate sequence to use. Must have at least four elements and be closed.
  * @returns number - A pointer to a newly allocated linear ring geometry. NULL on exception.
Caller is responsible for freeing with GEOSGeom_destroy().
    */
  GEOSGeom_createLinearRing(s: number): number;
  /**
  * Creates a linear ring geometry from a coordinate sequence.
  * @param handle - The handle to the GEOS context.
  * @param coordSeq - The pointer to the coordinate sequence.
  * @returns number - The pointer to the created geometry or NULL on exception.
    */
  GEOSGeom_createLinearRing_r(handle: number, coordSeq: number): number;
  /**
  * Creates a line string geometry from a coordinate sequence.
  * @param coordSeq - The pointer to the coordinate sequence.
  * @returns number - The pointer to the created geometry or NULL on exception.
    */
  GEOSGeom_createLineString(coordSeq: number): number;
  /**
  * Creates a line string geometry from a coordinate sequence.
  * @param handle - The handle to the GEOS context.
  * @param coordSeq - The pointer to the coordinate sequence.
  * @returns number - The pointer to the created geometry or NULL on exception.
    */
  GEOSGeom_createLineString_r(handle: number, coordSeq: number): number;
  /**
  * Creates a polygon geometry from an outer ring and an optional array of inner rings.
  * @param outerRing - The pointer to the outer ring geometry.
  * @param innerRings - The pointer to an array of inner ring geometries.
  * @param numInnerRings - The number of inner rings in the array.
  * @returns number - The pointer to the created geometry or NULL on exception.
    */
  GEOSGeom_createPolygon(outerRing: number, innerRings: number, numInnerRings: number): number;
  /**
  * Creates a polygon geometry from an outer ring and an optional array of inner rings.
  * @param handle - The handle to the GEOS context.
  * @param outerRing - The pointer to the outer ring geometry.
  * @param innerRings - The pointer to an array of inner ring geometries.
  * @param numInnerRings - The number of inner rings in the array.
  * @returns number - The pointer to the created geometry or NULL on exception.
    */
  GEOSGeom_createPolygon_r(handle: number, outerRing: number, innerRings: number, numInnerRings: number): number;
  /**
  * Clone a geometry object.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The pointer to the geometry object to clone.
  * @returns number - The pointer to the cloned geometry object, or null on exception.
    */
  GEOSGeom_clone_r(handle: number, g: number): number;
  /**
  * Clone a geometry object with a different context handle.
  * @param g - The pointer to the geometry object to clone.
  * @returns number - The pointer to the cloned geometry object, or null on exception.
    */
  GEOSGeom_clone(g: number): number;
  /**
  * Set the precision of a geometry object.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The pointer to the geometry object to modify.
  * @param precision - The precision value to set, or 0.0 for full precision.
  * @param flags - The flags to control the behavior of the function. See https://libgeos.org/doxygen/geos__c_8h.html for details.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSGeom_setPrecision_r(handle: number, g: number, precision: number, flags: number): number;
  /**
  * Set the precision of a geometry object with a different context handle.
  * @param g - The pointer to the geometry object to modify.
  * @param precision - The precision value to set, or 0.0 for full precision.
  * @param flags - The flags to control the behavior of the function. See https://libgeos.org/doxygen/geos__c_8h.html for details.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSGeom_setPrecision(g: number, precision: number, flags: number): number;
  /**
  * Get the precision of a geometry object.
  * @param g - The pointer to the geometry object to query.
  * @returns number - The precision value of the geometry object, or NaN on exception.
    */
  GEOSGeom_getPrecision(g: number): number;
  /**
  * Get the precision model of a geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry pointer.
  * @returns number - The precision model value, or 0 on exception.
  * @see {@link https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in#L2239}
  */
  GEOSGeom_getPrecision_r(handle: number, g: number): number;
  /**
  * Get the inherent dimension of a geometry object, which must be less than or equal to the coordinate dimension.
  * @param g - The GEOS geometry pointer.
  * @returns number - The inherent dimension of the geometry, or -1 on exception.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a4a2a3f2b9c7d6f9f0c8e5b4a0d6c1e7d}
  */
  GEOSGeom_getDimensions(g: number): number;
  /**
  * Get the inherent dimension of a geometry object, which must be less than or equal to the coordinate dimension.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry pointer.
  * @returns number - The inherent dimension of the geometry, or -1 on exception.
  * @see {@link https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in#L2245}
  */
  GEOSGeom_getDimensions_r(handle: number, g: number): number;
  /**
  * Get the coordinate dimension of a geometry object.
  * @param g - The GEOS geometry pointer.
  * @returns number - The coordinate dimension of the geometry, or 0 on exception.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a3f7a6b7f8b6a3e5d2e9e4a4f5d8a0b1c}
  */
  GEOSGeom_getCoordinateDimension(g: number): number;
  /**
  * Get the coordinate dimension of a geometry object.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The GEOS geometry pointer.
  * @returns number - The coordinate dimension of the geometry, or 0 on exception.
  * @see {@link https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in#L2251}
  */
  GEOSGeom_getCoordinateDimension_r(handle: number, g: number): number;
  /**
  * Finds the minimum X value in the geometry.
  * @param g - The pointer to the input geometry
  * @param value - The pointer to place the result
  * @returns number - 0 on exception
    */
  GEOSGeom_getXMin(g: number, value: number): number;
  /**
  * Finds the minimum X value in the geometry (thread-safe version).
  * @param handle - The pointer to the GEOS context
  * @param g - The pointer to the input geometry
  * @param value - The pointer to place the result
  * @returns number - 0 on exception
    */
  GEOSGeom_getXMin_r(handle: number, g: number, value: number): number;
  /**
  * Finds the minimum Y value in the geometry.
  * @param g - The pointer to the input geometry
  * @param value - The pointer to place the result
  * @returns number - 0 on exception
    */
  GEOSGeom_getYMin(g: number, value: number): number;
  /**
  * Finds the minimum Y value in the geometry (thread-safe version).
  * @param handle - The pointer to the GEOS context
  * @param g - The pointer to the input geometry
  * @param value - The pointer to place the result
  * @returns number - 0 on exception
    */
  GEOSGeom_getYMin_r(handle: number, g: number, value: number): number;
  /**
  * Finds the maximum X value in the geometry.
  * @param g - The pointer to the input geometry
  * @param value - The pointer to place the result
  * @returns number - 0 on exception
    */
  GEOSGeom_getXMax(g: number, value: number): number;
  /**
  * Returns the maximum X coordinate of the geometry.
  * @param handle - The GEOS context pointer handle.
  * @param g - The input geometry pointer.
  * @param value - The output value.
  * @returns number - 1 on success, 0 on exception.
  * @see https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in#L1652
  */
  GEOSGeom_getXMax_r(handle: number, g: number, value: number): number;
  /**
  * Returns the maximum Y coordinate of the geometry.
  * @param g - The input geometry pointer.
  * @param value - The output value.
  * @returns number - 1 on success, 0 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9c6f3a7b7d4a5f3a9e8f1c6b0b4e2d9f
  */
  GEOSGeom_getYMax(g: number, value: number): number;
  /**
  * Returns the maximum Y coordinate of the geometry.
  * @param handle - The GEOS context pointer handle.
  * @param g - The input geometry pointer.
  * @param value - The output value.
  * @returns number - 1 on success, 0 on exception.
  * @see https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in#L1655
  */
  GEOSGeom_getYMax_r(handle: number, g: number, value: number): number;
  /**
  * Simplifies a Geometry using the standard Douglas-Peucker algorithm.
  * @param g - The input geometry pointer.
  * @param tolerance - The distance tolerance for the simplification.
All vertices in the simplified geometry will be within this distance of the original geometry.
The tolerance value must be non-negative. A tolerance value of zero is effectively a no-op.
  * @returns number - A pointer to a new simplified geometry, or NULL if an exception occurs.
  * @see https://libgeos.org/doxygen/classgeos_1_1simplify_1_1DouglasPeuckerSimplifier.html
  */
  GEOSSimplify(g: number, tolerance: number): number;
  /**
  * Simplifies a Geometry using the standard Douglas-Peucker algorithm.
  * @param handle - The GEOS context pointer handle.
  * @param g - The input geometry pointer.
  * @param tolerance - The distance tolerance for the simplification.
All vertices in the simplified geometry will be within this distance of the original geometry.
The tolerance value must be non-negative. A tolerance value of zero is effectively a no-op.
  * @returns number - A pointer to a new simplified geometry, or NULL if an exception occurs.
  * @see https://docs.rs/geos-sys/1.0.10/geos_sys/fn.GEOSSimplify_r.html
  */
  GEOSSimplify_r(handle: number, g: number, tolerance: number): number;
  /**
  * Simplifies a geometry, ensuring that the result is a valid geometry having the same dimension and number of components as the input.
  * @param geom - A pointer to a GEOSGeometry object.
  * @param tolerance - The distance tolerance for the simplification.
  * @returns number - A pointer to a simplified GEOSGeometry object, or NULL on error.
  * @see https://libgeos.org/doxygen/classgeos_1_1simplify_1_1TopologyPreservingSimplifier.html
  */
  GEOSTopologyPreserveSimplify(geom: number, tolerance: number): number;
  /**
  * Simplifies a geometry, ensuring that the result is a valid geometry having the same dimension and number of components as the input. Thread-safe version.
  * @param handle - A pointer to a GEOSContextHandle object.
  * @param geom - A pointer to a GEOSGeometry object.
  * @param tolerance - The distance tolerance for the simplification.
  * @returns number - A pointer to a simplified GEOSGeometry object, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html
  */
  GEOSTopologyPreserveSimplify_r(handle: number, geom: number, tolerance: number): number;
  /**
  * Creates a new GEOSWKTReader object.
  * @returns number - A pointer to a GEOSWKTReader object, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html
  */
  GEOSWKTReader_create(): number;
  /**
  * Creates a new GEOSWKTReader object. Thread-safe version.
  * @param handle - A pointer to a GEOSContextHandle object.
  * @returns number - A pointer to a GEOSWKTReader object, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html
  */
  GEOSWKTReader_create_r(handle: number): number;
  /**
  * Destroys a GEOSWKTReader object and releases its memory.
  * @param reader - A pointer to a GEOSWKTReader object.
  * @returns void - undefined
    */
  GEOSWKTReader_destroy(reader: number): void;
  /**
  * Destroys a GEOSWKTReader object.
  * @param handle - A pointer to the GEOS context handle.
  * @param reader - The pointer to the GEOSWKTReader object.
  * @returns null - undefined
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a0a0f7c1b9f6a9f7c3c4d1b5a7b6f9e2e|GEOSWKTReader_destroy_r}
  */
  GEOSWKTReader_destroy_r(handle: number, reader: number): null;
  /**
  * Reads a WKT string and returns a GEOSGeometry object.
  * @param reader - The pointer to the GEOSWKTReader object.
  * @param wkt - The WKT string to read.
  * @returns number - The pointer to the GEOSGeometry object or null if an error occurred.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a0a0f7c1b9f6a9f7c3c4d1b5a7b6f9e2e|GEOSWKTReader_read_r}
  */
  GEOSWKTReader_read(reader: number, wkt: string): number;
  /**
  * Reads a WKT string and returns a GEOSGeometry object.
  * @param handle - A pointer to the GEOS context handle.
  * @param reader - The pointer to the GEOSWKTReader object.
  * @param wkt - The WKT string to read.
  * @returns number - The pointer to the GEOSGeometry object or null if an error occurred.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a0a0f7c1b9f6a9f7c3c4d1b5a7b6f9e2e|GEOSWKTReader_read_r}
  */
  GEOSWKTReader_read_r(handle: number, reader: number, wkt: string): number;
  /**
  * Creates a GEOSWKTWriter object.
  * @returns number - The pointer to the GEOSWKTWriter object or null if an error occurred.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a0a0f7c1b9f6a9f7c3c4d1b5a7b6f9e2e|GEOSWKTWriter_create}
  */
  GEOSWKTWriter_create(): number;
  /**
  * Creates a GEOSWKTWriter object.
  * @param handle - A pointer to the GEOS context handle.
  * @returns number - The pointer to the GEOSWKTWriter object or null if an error occurred.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html#a0a0f7c1b9f6a9f7c3c4d1b5a7b6f9e2e|GEOSWKTWriter_create_r}
  */
  GEOSWKTWriter_create_r(handle: number): number;
  /**
  * Destroys a GEOSWKTWriter object.
  * @param writer - The pointer to the GEOSWKTWriter object to be destroyed.
  * @returns void - undefined
    */
  GEOSWKTWriter_destroy(writer: number): void;
  /**
  * Destroys a GEOSWKTWriter object using a context handle.
  * @param handle - The pointer to the GEOSContextHandle_t object.
  * @param writer - The pointer to the GEOSWKTWriter object to be destroyed.
  * @returns void - undefined
    */
  GEOSWKTWriter_destroy_r(handle: number, writer: number): void;
  /**
  * Converts a GEOSGeometry object to a WKT string.
  * @param writer - The pointer to the GEOSWKTWriter object.
  * @param g - The pointer to the GEOSGeometry object to be converted.
  * @returns string - The WKT representation of the geometry, or null on error.
    */
  GEOSWKTWriter_write(writer: number, g: number): string;
  /**
  * Converts a GEOSGeometry object to a WKT string using a context handle.
  * @param handle - The pointer to the GEOSContextHandle_t object.
  * @param writer - The pointer to the GEOSWKTWriter object.
  * @param g - The pointer to the GEOSGeometry object to be converted.
  * @returns string - The WKT representation of the geometry, or null on error.
    */
  GEOSWKTWriter_write_r(handle: number, writer: number, g: number): string;
  /**
  * Sets whether the output WKT string should be trimmed of unnecessary spaces.
  * @param writer - The pointer to the GEOSWKTWriter object.
  * @param trim - A boolean value indicating whether to trim or not. 1 for true, 0 for false.
  * @returns void - undefined
    */
  GEOSWKTWriter_setTrim(writer: number, trim: number): void;
  /**
  * Sets the number trimming option on a GEOSWKTWriter.
* With trim set to 1, the writer will strip trailing 0's from
* the output coordinates. With 0, all coordinates will be
* padded with 0's out to the rounding precision.
  * @param handle - A pointer to the GEOS context handle.
  * @param writer - The GEOSWKTWriter object.
  * @param trim - The trimming option (0 or 1).
  * @returns void - undefined
    */
  GEOSWKTWriter_setTrim_r(handle: number, writer: number, trim: number): void;
  /**
  * Sets the number places after the decimal to output in WKT.
* Default 16.
  * @param writer - The GEOSWKTWriter object.
  * @param precision - The number of decimal places.
  * @returns void - undefined
    */
  GEOSWKTWriter_setRoundingPrecision(writer: number, precision: number): void;
  /**
  * Sets the number places after the decimal to output in WKT.
* Default 16.
  * @param handle - A pointer to the GEOS context handle.
  * @param writer - The GEOSWKTWriter object.
  * @param precision - The number of decimal places.
  * @returns void - undefined
    */
  GEOSWKTWriter_setRoundingPrecision_r(handle: number, writer: number, precision: number): void;
  /**
  * Sets the output dimension on a GEOSWKTWriter.
* If the specified dimension is 2, any input geometries
* that have a higher dimension will be flattened to 2D
* by dropping their z value. If the specified dimension is 3,
* any input geometries that have a lower dimension will be
* promoted to 3D by appending a z value of 0.
  * @param writer - The GEOSWKTWriter object.
  * @param dim - The output dimension (2 or 3).
  * @returns void - undefined
    */
  GEOSWKTWriter_setOutputDimension(writer: number, dim: number): void;
  /**
  * Sets the output dimension on a GEOSWKTWriter.
* If the specified dimension is 2, any input geometries
* that have a higher dimension will be flattened to 2D
* by dropping their z value. If the specified dimension is 3,
* any input geometries that have a lower dimension will be
* promoted to 3D by appending a z value of 0.
  * @param handle - A pointer to the GEOS context handle.
  * @param writer - The GEOSWKTWriter object.
  * @param dim - The output dimension (2 or 3).
  * @returns void - undefined
    */
  GEOSWKTWriter_setOutputDimension_r(handle: number, writer: number, dim: number): void;
  /**
  * Get the output dimension of a GEOSWKTWriter object.
  * @param writer - A pointer to a GEOSWKTWriter object.
  * @returns number - The output dimension, either 2 or 3.
  * @see https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in
  */
  GEOSWKTWriter_getOutputDimension(writer: number): number;
  /**
  * Get the output dimension of a GEOSWKTWriter object (thread-safe version).
  * @param handle - A GEOS context handle.
  * @param writer - A pointer to a GEOSWKTWriter object.
  * @returns number - The output dimension, either 2 or 3.
  * @see https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in
  */
  GEOSWKTWriter_getOutputDimension_r(handle: number, writer: number): number;
  /**
  * Set whether to use the old-style 3D WKT representation (Z inside brackets).
  * @param writer - A pointer to a GEOSWKTWriter object.
  * @param useOld3D - A boolean value indicating whether to use the old-style 3D WKT representation.
  * @returns void - undefined
  * @see https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in
  */
  GEOSWKTWriter_setOld3D(writer: number, useOld3D: number): void;
  /**
  * Set whether to use the old-style 3D WKT representation (Z inside brackets) (thread-safe version).
  * @param handle - A GEOS context handle.
  * @param writer - A pointer to a GEOSWKTWriter object.
  * @param useOld3D - A boolean value indicating whether to use the old-style 3D WKT representation.
  * @returns void - undefined
  * @see https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in
  */
  GEOSWKTWriter_setOld3D_r(handle: number, writer: number, useOld3D: number): void;
  /**
  * Create a new GEOSWKBReader object.
  * @returns number - A pointer to a newly allocated GEOSWKBReader object.
  * @see https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in
  */
  GEOSWKBReader_create(): number;
  /**
  * Creates a new WKB reader object.
  * @param handle - The GEOS context pointer handle.
  * @returns number - - The pointer to the WKB reader object.
    */
  GEOSWKBReader_create_r(handle: number): number;
  /**
  * Destroys a WKB reader object.
  * @param reader - The pointer to the WKB reader object.
  * @returns void - undefined
    */
  GEOSWKBReader_destroy(reader: number): void;
  /**
  * Destroys a WKB reader object using a context handle.
  * @param handle - The GEOS context pointer handle.
  * @param reader - The pointer to the WKB reader object.
  * @returns void - undefined
    */
  GEOSWKBReader_destroy_r(handle: number, reader: number): void;
  /**
  * Reads a geometry from a WKB byte buffer.
  * @param reader - The pointer to the WKB reader object.
  * @param wkb - The WKB byte buffer as a hex string.
  * @param size - The size of the WKB byte buffer in bytes.
  * @returns number - - The pointer to the geometry object, or null if an error occurred.
    */
  GEOSWKBReader_read(reader: number, wkb: string, size: number): number;
  /**
  * Reads a geometry from a WKB byte buffer using a context handle.
  * @param handle - The GEOS context pointer handle.
  * @param reader - The pointer to the WKB reader object.
  * @param wkb - The WKB byte buffer as a hex string.
  * @param size - The size of the WKB byte buffer in bytes.
  * @returns number - - The pointer to the geometry object, or null if an error occurred.
    */
  GEOSWKBReader_read_r(handle: number, reader: number, wkb: string, size: number): number;
  /**
  * Reads a geometry from a hexadecimal WKB representation.
  * @param reader - A pointer to a GEOSWKBReader object.
  * @param hex - A hexadecimal string representing the WKB of the geometry.
  * @param size - The length of the hexadecimal string.
  * @returns number - A pointer to a GEOSGeometry object, or null on error.
    */
  GEOSWKBReader_readHEX(reader: number, hex: string, size: number): number;
  /**
  * Reads a geometry from a hexadecimal WKB representation with a context handle.
  * @param handle - A GEOSContextHandle_t object.
  * @param reader - A pointer to a GEOSWKBReader object.
  * @param hex - A hexadecimal string representing the WKB of the geometry.
  * @param size - The length of the hexadecimal string.
  * @returns number - A pointer to a GEOSGeometry object, or null on error.
    */
  GEOSWKBReader_readHEX_r(handle: number, reader: number, hex: string, size: number): number;
  /**
  * Creates a new GEOSWKBWriter object.
  * @returns number - A pointer to a GEOSWKBWriter object, or null on error.
    */
  GEOSWKBWriter_create(): number;
  /**
  * Creates a new GEOSWKBWriter object with a context handle.
  * @param handle - A GEOSContextHandle_t object.
  * @returns number - A pointer to a GEOSWKBWriter object, or null on error.
    */
  GEOSWKBWriter_create_r(handle: number): number;
  /**
  * Destroys a GEOSWKBWriter object.
  * @param writer - A pointer to a GEOSWKBWriter object.
  * @returns void - undefined
    */
  GEOSWKBWriter_destroy(writer: number): void;
  /**
  * Destroys a GEOSWKBWriter object.
  * @param handle - A pointer to the GEOS context handle.
  * @param writer - The GEOSWKBWriter object to destroy.
  * @returns void - undefined
    */
  GEOSWKBWriter_destroy_r(handle: number, writer: number): void;
  /**
  * Writes a geometry as a WKB byte array.
  * @param writer - The GEOSWKBWriter object to use.
  * @param g - The GEOSGeometry object to write.
  * @returns string - A WKB byte array as a string.
    */
  GEOSWKBWriter_write(writer: number, g: number): string;
  /**
  * Writes a geometry as a WKB byte array with a context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param writer - The GEOSWKBWriter object to use.
  * @param g - The GEOSGeometry object to write.
  * @param size - A pointer to store the size of the output array.
  * @returns string - A WKB byte array as a string.
    */
  GEOSWKBWriter_write_r(handle: number, writer: number, g: number, size: number): string;
  /**
  * Writes a geometry as a WKB hex-encoded byte array.
  * @param writer - The GEOSWKBWriter object to use.
  * @param g - The GEOSGeometry object to write.
  * @param size - A pointer to store the size of the output array.
  * @returns string - A WKB hex-encoded byte array as a string.
    */
  GEOSWKBWriter_writeHEX(writer: number, g: number, size: number): string;
  /**
  * Writes a geometry as a WKB hex-encoded byte array with a context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param writer - The GEOSWKBWriter object to use.
  * @param g - The GEOSGeometry object to write.
  * @param size - A pointer to store the size of the output array.
  * @returns string - A WKB hex-encoded byte array as a string.
    */
  GEOSWKBWriter_writeHEX_r(handle: number, writer: number, g: number, size: number): string;
  /**
  * Get the output dimension of a WKBWriter object.
  * @param writer - A pointer to a GEOSWKBWriter object.
  * @returns number - The output dimension (2 or 3).
    */
  GEOSWKBWriter_getOutputDimension(writer: number): number;
  /**
  * Get the output dimension of a WKBWriter object (thread-safe).
  * @param handle - A pointer to a GEOS context handle.
  * @param writer - A pointer to a GEOSWKBWriter object.
  * @returns number - The output dimension (2 or 3).
    */
  GEOSWKBWriter_getOutputDimension_r(handle: number, writer: number): number;
  /**
  * Set the output dimension of a WKBWriter object.
  * @param writer - A pointer to a GEOSWKBWriter object.
  * @param newDimension - The new output dimension (2 or 3).
  * @returns void - undefined
    */
  GEOSWKBWriter_setOutputDimension(writer: number, newDimension: number): void;
  /**
  * Set the output dimension of a WKBWriter object (thread-safe).
  * @param handle - A pointer to a GEOS context handle.
  * @param writer - A pointer to a GEOSWKBWriter object.
  * @param newDimension - The new output dimension (2 or 3).
  * @returns void - undefined
    */
  GEOSWKBWriter_setOutputDimension_r(handle: number, writer: number, newDimension: number): void;
  /**
  * Get the byte order of a WKBWriter object.
  * @param writer - A pointer to a GEOSWKBWriter object.
  * @returns number - The byte order (0 for big endian, 1 for little endian).
    */
  GEOSWKBWriter_getByteOrder(writer: number): number;
  /**
  * Returns the byte order used by the writer.
  * @param handle - A pointer to the GEOS context handle.
  * @param writer - The GEOSWKBWriter pointer.
  * @returns number - The byte order (GEOS_WKB_XDR or GEOS_WKB_NDR).
    */
  GEOSWKBWriter_getByteOrder_r(handle: number, writer: number): number;
  /**
  * Sets the byte order for the writer.
  * @param writer - The GEOSWKBWriter pointer.
  * @param byteOrder - The byte order to use (GEOS_WKB_XDR or GEOS_WKB_NDR).
  * @returns void - undefined
    */
  GEOSWKBWriter_setByteOrder(writer: number, byteOrder: number): void;
  /**
  * Sets the byte order for the writer with a context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param writer - The GEOSWKBWriter pointer.
  * @param byteOrder - The byte order to use (GEOS_WKB_XDR or GEOS_WKB_NDR).
  * @returns void - undefined
    */
  GEOSWKBWriter_setByteOrder_r(handle: number, writer: number, byteOrder: number): void;
  /**
  * Returns whether the writer includes SRID values in output WKB.
  * @param writer - The GEOSWKBWriter pointer.
  * @returns number - 1 if SRID values are included, 0 otherwise.
    */
  GEOSWKBWriter_getIncludeSRID(writer: number): number;
  /**
  * Returns whether the writer includes SRID values in output WKB with a context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param writer - The GEOSWKBWriter pointer.
  * @returns number - 1 if SRID values are included, 0 otherwise.
    */
  GEOSWKBWriter_getIncludeSRID_r(handle: number, writer: number): number;
  /**
  * Sets whether output WKB should have an SRID header.
  * @param writer - A GEOSWKBWriter pointer.
  * @param writeSRID - A boolean value indicating whether to include SRID or not.
  * @returns void - undefined
    */
  GEOSWKBWriter_setIncludeSRID(writer: number, writeSRID: number): void;
  /**
  * Sets whether output WKB should have an SRID header, with a context handle.
  * @param handle - A GEOSContextHandle pointer.
  * @param writer - A GEOSWKBWriter pointer.
  * @param writeSRID - A boolean value indicating whether to include SRID or not.
  * @returns void - undefined
    */
  GEOSWKBWriter_setIncludeSRID_r(handle: number, writer: number, writeSRID: number): void;
  /**
  * Creates a prepared geometry from a regular geometry.
  * @param g - A GEOSGeometry pointer.
  * @returns number - A GEOSPreparedGeometry pointer, or NULL on exception.
    */
  GEOSPrepare(g: number): number;
  /**
  * Creates a prepared geometry from a regular geometry, with a context handle.
  * @param handle - A GEOSContextHandle pointer.
  * @param g - A GEOSGeometry pointer.
  * @returns number - A GEOSPreparedGeometry pointer, or NULL on exception.
    */
  GEOSPrepare_r(handle: number, g: number): number;
  /**
  * Destroys a prepared geometry.
  * @param g - A GEOSPreparedGeometry pointer.
  * @returns void - undefined
    */
  GEOSPreparedGeom_destroy(g: number): void;
  /**
  * Destroys a prepared geometry object.
  * @param handle - A pointer to the GEOS context handle.
  * @param prepGeom - The pointer to the prepared geometry object to be destroyed.
  * @returns void - undefined
    */
  GEOSPreparedGeom_destroy_r(handle: number, prepGeom: number): void;
  /**
  * Tests whether a prepared geometry contains another geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param prepGeom1 - The pointer to the prepared geometry object that is the potential container.
  * @param geom2 - The pointer to the geometry object that is the potential containee.
  * @returns number - 1 if prepGeom1 contains geom2, 0 if not, or -1 on error.
    */
  GEOSPreparedContains_r(handle: number, prepGeom1: number, geom2: number): number;
  /**
  * Tests whether a prepared geometry contains another geometry.
  * @param prepGeom1 - The pointer to the prepared geometry object that is the potential container.
  * @param geom2 - The pointer to the geometry object that is the potential containee.
  * @returns number - 1 if prepGeom1 contains geom2, 0 if not, or -1 on error.
    */
  GEOSPreparedContains(prepGeom1: number, geom2: number): number;
  /**
  * Tests whether a prepared geometry properly contains another geometry. Proper containment means that the test geometry is contained in the interior of the target geometry, and does not intersect its boundary.
  * @param handle - A pointer to the GEOS context handle.
  * @param prepGeom1 - The pointer to the prepared geometry object that is the potential container.
  * @param geom2 - The pointer to the geometry object that is the potential containee.
  * @returns number - 1 if prepGeom1 properly contains geom2, 0 if not, or -1 on error.
    */
  GEOSPreparedContainsProperly_r(handle: number, prepGeom1: number, geom2: number): number;
  /**
  * Tests whether a prepared geometry properly contains another geometry. Proper containment means that the test geometry is contained in the interior of the target geometry, and does not intersect its boundary.
  * @param prepGeom1 - The pointer to the prepared geometry object that is the potential container.
  * @param geom2 - The pointer to the geometry object that is the potential containee.
  * @returns number - 1 if prepGeom1 properly contains geom2, 0 if not, or -1 on error.
    */
  GEOSPreparedContainsProperly(prepGeom1: number, geom2: number): number;
  /**
  * Tests whether a geometry is covered by another geometry.
  * @param pg1 - A pointer to a prepared geometry.
  * @param g2 - A pointer to a geometry.
  * @returns number - 1 if g2 is covered by pg1, 0 if not, -1 on exception.
    */
  GEOSPreparedCoveredBy(pg1: number, g2: number): number;
  /**
  * Tests whether a geometry is covered by another geometry, using a GEOS context handle.
  * @param handle - A GEOS context handle.
  * @param pg1 - A pointer to a prepared geometry.
  * @param g2 - A pointer to a geometry.
  * @returns number - 1 if g2 is covered by pg1, 0 if not, -1 on exception.
    */
  GEOSPreparedCoveredBy_r(handle: number, pg1: number, g2: number): number;
  /**
  * Tests whether a geometry covers another geometry.
  * @param pg1 - A pointer to a prepared geometry.
  * @param g2 - A pointer to a geometry.
  * @returns number - 1 if pg1 covers g2, 0 if not, -1 on exception.
    */
  GEOSPreparedCovers(pg1: number, g2: number): number;
  /**
  * Tests whether a geometry covers another geometry, using a GEOS context handle.
  * @param handle - A GEOS context handle.
  * @param pg1 - A pointer to a prepared geometry.
  * @param g2 - A pointer to a geometry.
  * @returns number - 1 if pg1 covers g2, 0 if not, -1 on exception.
    */
  GEOSPreparedCovers_r(handle: number, pg1: number, g2: number): number;
  /**
  * Tests whether a geometry crosses another geometry.
  * @param pg1 - A pointer to a prepared geometry.
  * @param g2 - A pointer to a geometry.
  * @returns number - 1 if pg1 crosses g2, 0 if not, -1 on exception.
    */
  GEOSPreparedCrosses(pg1: number, g2: number): number;
  /**
  * Computes whether the given geometries cross.
  * @param handle - A pointer to the GEOS context handle.
  * @param pg1 - The first prepared geometry.
  * @param g2 - The second geometry.
  * @returns number - 1 if the geometries cross, 0 if they do not, or -1 on error.
    */
  GEOSPreparedCrosses_r(handle: number, pg1: number, g2: number): number;
  /**
  * Computes whether the given geometries are disjoint.
  * @param pg1 - The first prepared geometry.
  * @param g2 - The second geometry.
  * @returns number - 1 if the geometries are disjoint, 0 if they are not, or -1 on error.
    */
  GEOSPreparedDisjoint(pg1: number, g2: number): number;
  /**
  * Computes whether the given geometries are disjoint.
  * @param handle - A pointer to the GEOS context handle.
  * @param pg1 - The first prepared geometry.
  * @param g2 - The second geometry.
  * @returns number - 1 if the geometries are disjoint, 0 if they are not, or -1 on error.
    */
  GEOSPreparedDisjoint_r(handle: number, pg1: number, g2: number): number;
  /**
  * Computes whether the given geometries intersect.
  * @param pg1 - The first prepared geometry.
  * @param g2 - The second geometry.
  * @returns number - 1 if the geometries intersect, 0 if they do not, or -1 on error.
    */
  GEOSPreparedIntersects(pg1: number, g2: number): number;
  /**
  * Computes whether the given geometries intersect.
  * @param handle - A pointer to the GEOS context handle.
  * @param pg1 - The first prepared geometry.
  * @param g2 - The second geometry.
  * @returns number - 1 if the geometries intersect, 0 if they do not, or -1 on error.
    */
  GEOSPreparedIntersects_r(handle: number, pg1: number, g2: number): number;
  /**
  * Tests whether two geometries overlap.
  * @param pg1 - A pointer to the first prepared geometry.
  * @param g2 - A pointer to the second geometry.
  * @returns number - 1 if the geometries overlap, 0 if not, -1 on exception.
    */
  GEOSPreparedOverlaps(pg1: number, g2: number): number;
  /**
  * Tests whether two geometries overlap, using a context handle.
  * @param handle - The context handle.pointer
  * @param pg1 - A pointer to the first prepared geometry.
  * @param g2 - A pointer to the second geometry.
  * @returns number - 1 if the geometries overlap, 0 if not, -1 on exception.
    */
  GEOSPreparedOverlaps_r(handle: number, pg1: number, g2: number): number;
  /**
  * Tests whether two geometries touch at one or more points.
  * @param pg1 - A pointer to the first prepared geometry.
  * @param g2 - A pointer to the second geometry.
  * @returns number - 1 if the geometries touch, 0 if not, -1 on exception.
    */
  GEOSPreparedTouches(pg1: number, g2: number): number;
  /**
  * Tests whether two geometries touch at one or more points, using a context handle.
  * @param handle - The context handle.pointer
  * @param pg1 - A pointer to the first prepared geometry.
  * @param g2 - A pointer to the second geometry.
  * @returns number - 1 if the geometries touch, 0 if not, -1 on exception.
    */
  GEOSPreparedTouches_r(handle: number, pg1: number, g2: number): number;
  /**
  * Tests whether the first geometry is within the second geometry.
  * @param pg1 - A pointer to the first prepared geometry.
  * @param g2 - A pointer to the second geometry.
  * @returns number - 1 if the first geometry is within the second geometry, 0 if not, -1 on exception.
    */
  GEOSPreparedWithin(pg1: number, g2: number): number;
  /**
  * Tests whether a geometry is within another geometry.
  * @param handle - The GEOS context pointer handle.
  * @param pg1 - A pointer to the prepared geometry to test.
  * @param g2 - A pointer to the geometry to test against.
  * @returns number - 1 if pg1 is within g2, 0 otherwise, or -1 on error.
    */
  GEOSPreparedWithin_r(handle: number, pg1: number, g2: number): number;
  /**
  * Computes the nearest points of two geometries.
  * @param pg1 - A pointer to the first prepared geometry.
  * @param g2 - A pointer to the second geometry.
  * @returns number - A pointer to a coordinate sequence containing the nearest points, or NULL on error.
    */
  GEOSPreparedNearestPoints(pg1: number, g2: number): number;
  /**
  * Computes the nearest points of two geometries using a GEOS context handle.
  * @param handle - The GEOS context pointer handle.
  * @param pg1 - A pointer to the first prepared geometry.
  * @param g2 - A pointer to the second geometry.
  * @returns number - A pointer to a coordinate sequence containing the nearest points, or NULL on error.
    */
  GEOSPreparedNearestPoints_r(handle: number, pg1: number, g2: number): number;
  /**
  * Computes the distance between two geometries using a prepared geometry for the first argument.
  * @param pg1 - A pointer to the prepared geometry to compute the distance from.
  * @param g2 - A pointer to the geometry to compute the distance to.
  * @param dist - A pointer to a double variable to store the distance result.
  * @returns number - 1 if the distance was computed successfully, 0 otherwise, or -1 on error.
    */
  GEOSPreparedDistance(pg1: number, g2: number, dist: number): number;
  /**
  * Computes the distance between two geometries using a prepared geometry for the first argument and a GEOS context handle.
  * @param handle - The GEOS context pointer handle.
  * @param pg1 - A pointer to the prepared geometry to compute the distance from.
  * @param g2 - A pointer to the geometry to compute the distance to.
  * @param dist - A pointer to a double variable to store the distance result.
  * @returns number - 1 if the distance was computed successfully, 0 otherwise, or -1 on error.
    */
  GEOSPreparedDistance_r(handle: number, pg1: number, g2: number, dist: number): number;
  /**
  * Creates a GEOS STRTree, a spatial index for quickly querying geometries by their bounding boxes.
  * @param nodeCapacity - The maximum number of child nodes that a node may have. The minimum recommended capacity value is 4. If unsure, use a default node capacity of 10.
  * @returns number - A pointer to the GEOS STRTree object.
    */
  GEOSSTRtree_create(nodeCapacity: number): number;
  /**
  * Creates a GEOS STRTree with a context handle, a spatial index for quickly querying geometries by their bounding boxes.
  * @param handle - The context handle to use for error reporting and memory management.
  * @param nodeCapacity - The maximum number of child nodes that a node may have. The minimum recommended capacity value is 4. If unsure, use a default node capacity of 10.
  * @returns number - A pointer to the GEOS STRTree object.
    */
  GEOSSTRtree_create_r(handle: number, nodeCapacity: number): number;
  /**
  * Inserts a geometry into a GEOS STRTree, along with an associated item that can be retrieved later.
  * @param tree - A pointer to the GEOS STRTree object.
  * @param g - A pointer to the GEOS geometry object to insert.
  * @param item - A pointer to the item associated with the geometry. This can be any arbitrary data that can be cast to a void pointer.
  * @returns void - undefined
    */
  GEOSSTRtree_insert(tree: number, g: number, item: number): void;
  /**
  * Inserts a geometry into a GEOS STRTree with a context handle, along with an associated item that can be retrieved later.
  * @param handle - The context handle to use for error reporting and memory management.
  * @param tree - A pointer to the GEOS STRTree object.
  * @param g - A pointer to the GEOS geometry object to insert.
  * @param item - A pointer to the item associated with the geometry. This can be any arbitrary data that can be cast to a void pointer.
  * @returns void - undefined
    */
  GEOSSTRtree_insert_r(handle: number, tree: number, g: number, item: number): void;
  /**
  * Queries a GEOS STRTree for all geometries whose bounding boxes intersect with the given geometry's bounding box, and calls a callback function for each matching pair.
  * @param tree - A pointer to the GEOS STRTree object.
  * @param g - A pointer to the GEOS geometry object to query with.
  * @param callback - A function that takes two void pointers as arguments: one for the item associated with the tree geometry, and one for the user data passed to this function. The callback function is called for each matching pair of geometries in the tree and the query geometry. The callback function should return 1 to continue the query or 0 to stop it.
  * @param userdata - A pointer to any user data that needs to be passed to the callback function. This can be any arbitrary data that can be cast to a void pointer.
  * @returns void - undefined
    */
  GEOSSTRtree_query(tree: number, g: number, callback: function, userdata: any): void;
  /**
  * Queries the tree for all items whose extents intersect the given geometry's envelope and applies an
* item visitor function to them.
  * @param handle - A pointer to the GEOS context handle.
  * @param tree - The GEOSSTRtree to query.
  * @param g - The GEOSGeometry whose envelope is used for querying.
  * @param callback - The GEOSQueryCallback function to apply to the items found.
  * @param userdata - An optional pointer to user data that is passed to the callback function.
  * @returns void - undefined
    */
  GEOSSTRtree_query_r(handle: number, tree: number, g: number, callback: number, userdata: number): void;
  /**
  * Returns the item whose extent is nearest to the given geometry's envelope.
  * @param tree - The GEOSSTRtree to query.
  * @param geom - The GEOSGeometry whose envelope is used for querying.
  * @returns number - A pointer to the nearest item, or NULL if the tree is empty.
    */
  GEOSSTRtree_nearest(tree: number, geom: number): number;
  /**
  * Returns the item whose extent is nearest to the given geometry's envelope, using a custom distance function.
  * @param handle - A pointer to the GEOS context handle.
  * @param tree - The GEOSSTRtree to query.
  * @param geom - The GEOSGeometry whose envelope is used for querying.
  * @returns number - A pointer to the nearest item, or NULL if the tree is empty or an error occurred.
    */
  GEOSSTRtree_nearest_r(handle: number, tree: number, geom: number): number;
  /**
  * Returns the item whose extent is nearest to the given item's envelope, using a custom distance function.
  * @param tree - The GEOSSTRtree to query.
  * @param item - The item whose envelope is used for querying.
  * @param itemEnvelope - The GEOSGeometry representing the envelope of the item.
  * @param distancefn - The GEOSDistanceCallback function to compute the distance between two items.
  * @param userdata - An optional pointer to user data that is passed to the distance function.
  * @returns any - A pointer to the nearest item, or NULL if the tree is empty or an error occurred.
    */
  GEOSSTRtree_nearest_generic(tree: number, item: any, itemEnvelope: number, distancefn: function, userdata: any): any;
  /**
  * Returns the item whose extent is nearest to the given item's envelope, using a custom distance function.
  * @param handle - A pointer to the GEOS context handle.
  * @param tree - The GEOSSTRtree to query.
  * @param item - The item whose envelope is used for querying.
  * @param itemEnvelope - The GEOSGeometry representing the envelope of the item.
  * @param distancefn - The GEOSDistanceCallback function to compute the distance between two items.
  * @param userdata - An optional pointer to user data that is passed to the distance function and the callback function.
  * @returns any - A pointer to the nearest item, or NULL if the tree is empty or an error occurred.
    */
  GEOSSTRtree_nearest_generic_r(handle: number, tree: number, item: any, itemEnvelope: number, distancefn: function, userdata: any): any;
  /**
  * Iterates over every item in an STRtree.
  * @param tree - A pointer to a GEOSSTRtree object.
  * @param callback - A function to be called for each item in the tree.
  * @param userdata - An optional user data to be passed to the callback function.
  * @returns void - undefined
    */
  GEOSSTRtree_iterate(tree: number, callback: function, userdata: any): void;
  /**
  * Iterates over every item in an STRtree using a reentrant context handle.
  * @param handle - A pointer to a GEOSpointer  context handle.
  * @param tree - A pointer to a GEOSSTRtree object.
  * @param callback - A function to be called for each item in the tree.
  * @param userdata - An optional user data to be passed to the callback function.
  * @returns void - undefined
    */
  GEOSSTRtree_iterate_r(handle: number, tree: number, callback: function, userdata: any): void;
  /**
  * Removes an item from an STRtree that has a matching geometry and user data.
  * @param tree - A pointer to a GEOSSTRtree object.
  * @param g - A pointer to a GEOSGeometry object representing the envelope of the item to be removed.
  * @param item - The user data of the item to be removed.
  * @returns number - 1 if the item was found and removed, 0 otherwise.
    */
  GEOSSTRtree_remove(tree: number, g: number, item: any): number;
  /**
  * Removes an item from an STRtree that has a matching geometry and user data using a reentrant context handle.
  * @param handle - A pointer to a GEOSpointer  context handle.
  * @param tree - A pointer to a GEOSSTRtree object.
  * @param g - A pointer to a GEOSGeometry object representing the envelope of the item to be removed.
  * @param item - The user data of the item to be removed.
  * @returns number - 1 if the item was found and removed, 0 otherwise.
    */
  GEOSSTRtree_remove_r(handle: number, tree: number, g: number, item: any): number;
  /**
  * Destroys an STRtree and frees its memory.
  * @param tree - A pointer to a GEOSSTRtree object.
  * @returns void - undefined
    */
  GEOSSTRtree_destroy(tree: number): void;
  /**
  * Destroys a GEOSSTRtree spatial index.
  * @param handle - A pointer to the GEOS context handle.
  * @param tree - The GEOSSTRtree pointer to destroy.
  * @returns void - undefined
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a1323e
  */
  GEOSSTRtree_destroy_r(handle: number, tree: number): void;
  /**
  * Returns the distance of a point projected on a line from the origin of the line.
  * @param line - The GEOSGeometry pointer of the lineal geometry.
  * @param point - The GEOSGeometry pointer of the point geometry.
  * @returns number - The distance of the projected point along the line, or -1 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a557
  */
  GEOSProject(line: number, point: number): number;
  /**
  * Returns the distance of a point projected on a line from the origin of the line, using a reentrant context.
  * @param handle - A pointer to the GEOS context handle.
  * @param line - The GEOSGeometry pointer of the lineal geometry.
  * @param point - The GEOSGeometry pointer of the point geometry.
  * @returns number - The distance of the projected point along the line, or -1 on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h_source.html#l00557
  */
  GEOSProject_r(handle: number, line: number, point: number): number;
  /**
  * Measuring from start of line, return point that is distance the start. Line parameter must be a LineString. The returned point is not guaranteed to intersect the line due to limitations of floating point calculations.
  * @param line - The GEOSGeometry pointer of the lineal geometry.
  * @param distance - The distance along the linestring to interpolate.
  * @returns number - A new GEOSGeometry pointer representing the interpolated point, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a1849
  */
  GEOSInterpolate(line: number, distance: number): number;
  /**
  * Measuring from start of line, return point that is distance the start. Line parameter must be a LineString. The returned point is not guaranteed to intersect the line due to limitations of floating point calculations.
  * @param handle - A pointer to the GEOS context handle.
  * @param line - The GEOSGeometry pointer of the lineal geometry.
  * @param distance - The distance along the linestring to interpolate.
  * @returns number - A new GEOSGeometry pointer representing the interpolated point, or NULL on exception.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a1850
  */
  GEOSInterpolate_r(handle: number, line: number, distance: number): number;
  /**
  * Returns the distance of point p projected on line g from the origin of g,
* which must be a lineal geometry. The distance is normalized to the length of g.
  * @param g - A pointer to a GEOSGeometry object representing a lineal geometry.
  * @param p - A pointer to a GEOSGeometry object representing a point.
  * @returns number - The normalized distance of p projected on g, or -1 on error.
  * @see {@link https://docs.rs/geos-sys/1.0.13/geos_sys/fn.GEOSProjectNormalized.html}
  */
  GEOSProjectNormalized(g: number, p: number): number;
  /**
  * Returns the distance of point p projected on line g from the origin of g,
* which must be a lineal geometry. The distance is normalized to the length of g.
* This is a thread-safe variant of GEOSProjectNormalized that takes a GEOS context handle as an argument.
  * @param handle - A pointer to a GEOSContextHandle_t object.
  * @param g - A pointer to a GEOSGeometry object representing a lineal geometry.
  * @param p - A pointer to a GEOSGeometry object representing a point.
  * @returns number - The normalized distance of p projected on g, or -1 on error.
  * @see {@link https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in}
  */
  GEOSProjectNormalized_r(handle: number, g: number, p: number): number;
  /**
  * Returns the point along line g representing the given distance from the origin along the geometry.
* The distance is normalized to the length of g, which must be a lineal geometry.
  * @param g - A pointer to a GEOSGeometry object representing a lineal geometry.
  * @param d - The normalized distance along the linestring to interpolate.
  * @returns number - A pointer to a new GEOSGeometry object representing the interpolated point, or NULL on error.
  * @see {@link https://rdrr.io/cran/geos/man/geos_project.html}
  */
  GEOSInterpolateNormalized(g: number, d: number): number;
  /**
  * Returns the point along line g representing the given distance from the origin along the geometry.
* The distance is normalized to the length of g, which must be a lineal geometry.
* This is a thread-safe variant of GEOSInterpolateNormalized that takes a GEOS context handle as an argument.
  * @param handle - A pointer to a GEOSContextHandle_t object.
  * @param g - A pointer to a GEOSGeometry object representing a lineal geometry.
  * @param d - The normalized distance along the linestring to interpolate.
  * @returns number - A pointer to a new GEOSGeometry object representing the interpolated point, or NULL on error.
  * @see {@link https://github.com/libgeos/geos/blob/main/capi/geos_c.h.in}
  */
  GEOSInterpolateNormalized_r(handle: number, g: number, d: number): number;
  /**
  * Returns a multipoint containing all distinct vertices of input geometry g,
* which can be any kind of geometry.
  * @param g - A pointer to a GEOSGeometry object representing any kind of geometry.
  * @returns number - A pointer to a new GEOSGeometry object representing a multipoint, or NULL on error.
    */
  GEOSGeom_extractUniquePoints(g: number): number;
  /**
  * Extracts every unique point from a geometry.
  * @param handle - A pointer to the GEOS context handle.
  * @param g - The geometry to extract points from.
  * @returns number - A geometry containing the unique points, or null on exception.
    */
  GEOSGeom_extractUniquePoints_r(handle: number, g: number): number;
  /**
  * Creates an empty geometry collection of a given type.
  * @param type - The geometry type, one of GEOS_GEOMETRYCOLLECTION, GEOS_MULTIPOINT, GEOS_MULTILINESTRING, GEOS_MULTIPOLYGON.
  * @returns number - An empty geometry collection, or null on exception.
    */
  GEOSGeom_createEmptyCollection(type: number): number;
  /**
  * Creates an empty geometry collection of a given type with a context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param type - The geometry type, one of GEOS_GEOMETRYCOLLECTION, GEOS_MULTIPOINT, GEOS_MULTILINESTRING, GEOS_MULTIPOLYGON.
  * @returns number - An empty geometry collection, or null on exception.
    */
  GEOSGeom_createEmptyCollection_r(handle: number, type: number): number;
  /**
  * Creates an empty point geometry.
  * @returns number - An empty point geometry, or null on exception.
    */
  GEOSGeom_createEmptyPoint(): number;
  /**
  * Creates an empty point geometry with a context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @returns number - An empty point geometry, or null on exception.
    */
  GEOSGeom_createEmptyPoint_r(handle: number): number;
  /**
  * Creates an empty line string geometry.
  * @returns number - A pointer to a newly allocated empty line string geometry. NULL on exception.
Caller is responsible for freeing with GEOSGeom_destroy().
  * @see https://libgeos.org/usage/c_api/#geosgeom_createemptylinestring
  */
  GEOSGeom_createEmptyLineString(): number;
  /**
  * Creates an empty line string geometry using a reentrant context handle.
  * @param handle - A GEOSContextHandle_t created by GEOS_init_r().
  * @returns number - A pointer to a newly allocated empty line string geometry. NULL on exception.
Caller is responsible for freeing with GEOSGeom_destroy_r().
  * @see https://libgeos.org/usage/c_api/#geosgeom_createemptylinestring_r
  */
  GEOSGeom_createEmptyLineString_r(handle: number): number;
  /**
  * Creates an empty polygon geometry.
  * @returns number - A pointer to a newly allocated empty polygon geometry. NULL on exception.
Caller is responsible for freeing with GEOSGeom_destroy().
  * @see https://libgeos.org/usage/c_api/#geosgeom_createemptypolygon
  */
  GEOSGeom_createEmptyPolygon(): number;
  /**
  * Creates an empty polygon geometry using a reentrant context handle.
  * @param handle - A GEOSContextHandle_t created by GEOS_init_r().
  * @returns number - A pointer to a newly allocated empty polygon geometry. NULL on exception.
Caller is responsible for freeing with GEOSGeom_destroy_r().
  * @see https://libgeos.org/usage/c_api/#geosgeom_createemptypolygon_r
  */
  GEOSGeom_createEmptyPolygon_r(handle: number): number;
  /**
  * Computes the orientation index of the direction of the point q relative to a directed infinite line specified by p1-p2.
* The index indicates whether the point lies to the LEFT or RIGHT of the line, or lies on it COLLINEAR.
* The index also indicates the orientation of the triangle formed by the three points (COUNTERCLOCKWISE, CLOCKWISE, or STRAIGHT).
  * @param ax - The x coordinate of p1.
  * @param ay - The y coordinate of p1.
  * @param bx - The x coordinate of p2.
  * @param by - The y coordinate of p2.
  * @param px - The x coordinate of q.
  * @param py - The y coordinate of q.
  * @returns number - The orientation index: 1 (LEFT), 0 (COLLINEAR), or -1 (RIGHT).
  * @see https://libgeos.org/doxygen/classgeos_1_1algorithm_1_1Orientation.html#a5b9f8c4a7c6f0a3f3a4e7c8d6b9a5d7b
  */
  GEOSOrientationIndex(ax: number, ay: number, bx: number, by: number, px: number, py: number): number;
  /**
  * Computes the orientation index of the direction of the point q relative to
* the directed line segment p1-p2.
  * @param handle - A pointer to the GEOS context handle.
  * @param Ax - The x-coordinate of the first point of the line segment.
  * @param Ay - The y-coordinate of the first point of the line segment.
  * @param Bx - The x-coordinate of the second point of the line segment.
  * @param By - The y-coordinate of the second point of the line segment.
  * @param Px - The x-coordinate of the point to test.
  * @param Py - The y-coordinate of the point to test.
  * @returns number - -1 if q lies to the left of p1-p2, 0 if q is collinear with p1-p2, 1 if q lies to the right of p1-p2.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html}
  */
  GEOSOrientationIndex_r(handle: number, Ax: number, Ay: number, Bx: number, By: number, Px: number, Py: number): number;
  /**
  * Computes a geometry representing the shared paths between two linear geometries.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - A pointer to the first geometry. Must be linear.
  * @param g2 - A pointer to the second geometry. Must be linear.
  * @returns number - A pointer to the geometry representing the shared paths between g1 and g2, or NULL on exception.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html}
  */
  GEOSSharedPaths(handle: number, g1: number, g2: number): number;
  /**
  * Computes a geometry representing the shared paths between two linear geometries using a GEOS context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - A pointer to the first geometry. Must be linear.
  * @param g2 - A pointer to the second geometry. Must be linear.
  * @returns number - A pointer to the geometry representing the shared paths between g1 and g2, or NULL on exception.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html}
  */
  GEOSSharedPaths_r(handle: number, g1: number, g2: number): number;
  /**
  * Snaps two geometries together with a given tolerance.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - A pointer to the first geometry to snap.
  * @param g2 - A pointer to the second geometry to snap.
  * @param tolerance - The snapping tolerance to use, in same units as input geometries.
  * @returns number - A pointer to a new snapped geometry, or NULL on exception.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html}
  */
  GEOSSnap(handle: number, g1: number, g2: number, tolerance: number): number;
  /**
  * Snaps two geometries together with a given tolerance using a GEOS context handle.
  * @param handle - A pointer to the GEOS context handle.
  * @param g1 - A pointer to the first geometry to snap.
  * @param g2 - A pointer to the second geometry to snap.
  * @param tolerance - The snapping tolerance to use, in same units as input geometries.
  * @returns number - A pointer to a new snapped geometry, or NULL on exception.
  * @see {@link https://libgeos.org/doxygen/geos__c_8h.html}
  */
  GEOSSnap_r(handle: number, g1: number, g2: number, tolerance: number): number;
  /**
  * Create a new GEOSBufferParams object
  * @returns number - A pointer to the newly allocated GEOSBufferParams object
    */
  GEOSBufferParams_create(): number;
  /**
  * Create a new GEOSBufferParams object with a context handle
  * @param handle - The context handle
  * @returns number - A pointer to the newly allocated GEOSBufferParams object
    */
  GEOSBufferParams_create_r(handle: number): number;
  /**
  * Destroy a GEOSBufferParams object
  * @param params - A pointer to the GEOSBufferParams object to destroy
  * @returns void - undefined
    */
  GEOSBufferParams_destroy(params: number): void;
  /**
  * Destroy a GEOSBufferParams object with a context handle
  * @param handle - The context handle
  * @param params - A pointer to the GEOSBufferParams object to destroy
  * @returns void - undefined
    */
  GEOSBufferParams_destroy_r(handle: number, params: number): void;
  /**
  * Set the end cap style for a GEOSBufferParams object
  * @param params - A pointer to the GEOSBufferParams object to modify
  * @param style - The end cap style, one of **GEOSBUF_CAP_ROUND**, **GEOSBUF_CAP_FLAT**, or **GEOSBUF_CAP_SQUARE**
  * @returns number - 1 on success, 0 on failure
    */
  GEOSBufferParams_setEndCapStyle(params: number, style: number): number;
  /**
  * Sets the end cap style of the buffer parameters.
  * @param handle - A pointer to the GEOS context handle.
  * @param p - The pointer to the buffer parameters object.
  * @param style - The end cap style code. Possible values are:
 - 1: round
 - 2: flat
 - 3: square
  * @returns null - undefined
    */
  GEOSBufferParams_setEndCapStyle_r(handle: number, p: number, style: number): null;
  /**
  * Sets the join style of the buffer parameters.
  * @param p - The pointer to the buffer parameters object.
  * @param joinStyle - The join style code. Possible values are:
 - 1: round
 - 2: mitre
 - 3: bevel
  * @returns null - undefined
    */
  GEOSBufferParams_setJoinStyle(p: number, joinStyle: number): null;
  /**
  * Sets the join style of the buffer parameters.
  * @param handle - A pointer to the GEOS context handle.
  * @param p - The pointer to the buffer parameters object.
  * @param joinStyle - The join style code. Possible values are:
 - 1: round
 - 2: mitre
 - 3: bevel
  * @returns null - undefined
    */
  GEOSBufferParams_setJoinStyle_r(handle: number, p: number, joinStyle: number): null;
  /**
  * Sets the mitre limit of the buffer parameters.
  * @param p - The pointer to the buffer parameters object.
  * @param mitreLimit - The mitre limit value. This is used for very sharp corners.
The mitre limit is the ratio of the distance from the corner to the end of the mitred offset corner.
When two line segments meet at a sharp angle, a miter join will extend the original geometry.
To prevent unreasonable geometry, the mitre limit allows controlling how far the miter join can extend from the original geometry.
If the limit is exceeded, the join will be beveled.
  * @returns null - undefined
    */
  GEOSBufferParams_setMitreLimit(p: number, mitreLimit: number): null;
  /**
  * Sets the mitre limit of the buffer parameters.
  * @param handle - A pointer to the GEOS context handle.
  * @param p - The pointer to the buffer parameters object.
  * @param mitreLimit - The mitre limit value. This is used for very sharp corners.
The mitre limit is the ratio of the distance from the corner to the end of the mitred offset corner.
When two line segments meet at a sharp angle, a miter join will extend the original geometry.
To prevent unreasonable geometry, the mitre limit allows controlling how far the miter join can extend from the original geometry.
If the limit is exceeded, the join will be beveled.
  * @returns null - undefined
    */
  GEOSBufferParams_setMitreLimit_r(handle: number, p: number, mitreLimit: number): null;
  /**
  * Sets the number of line segments used to approximate an angle fillet.
  * @param p - A pointer to a GEOSBufferParams object.
  * @param quadSegs - The number of segments in a fillet for a quadrant, must be >= 1.
  * @returns null - - No return value.
    */
  GEOSBufferParams_setQuadrantSegments(p: number, quadSegs: number): null;
  /**
  * Sets the number of line segments used to approximate an angle fillet, with a context handle.
  * @param handle - A pointer to a GEOSContextHandle object.
  * @param p - A pointer to a GEOSBufferParams object.
  * @param quadSegs - The number of segments in a fillet for a quadrant, must be >= 1.
  * @returns null - - No return value.
    */
  GEOSBufferParams_setQuadrantSegments_r(handle: number, p: number, quadSegs: number): null;
  /**
  * Sets whether the computed buffer should be single-sided.
* A single-sided buffer is constructed on only one side of each input line.
* The side is determined by the sign of the buffer distance:
* - A positive distance indicates the left-hand side
* - A negative distance indicates the right-hand side
* The single-sided buffer of point geometries is the same as the regular buffer.
* The End Cap Style for single-sided buffers is always ignored, and forced to the equivalent of CAP_FLAT.
  * @param p - A pointer to a GEOSBufferParams object.
  * @param singleSided - A boolean value indicating whether the buffer is single-sided or not.
  * @returns null - - No return value.
    */
  GEOSBufferParams_setSingleSided(p: number, singleSided: number): null;
  /**
  * Sets whether the computed buffer should be single-sided, with a context handle.
* A single-sided buffer is constructed on only one side of each input line.
* The side is determined by the sign of the buffer distance:
* - A positive distance indicates the left-hand side
* - A negative distance indicates the right-hand side
* The single-sided buffer of point geometries is the same as the regular buffer.
* The End Cap Style for single-sided buffers is always ignored, and forced to the equivalent of CAP_FLAT.
  * @param handle - A pointer to a GEOSContextHandle object.
  * @param p - A pointer to a GEOSBufferParams object.
  * @param singleSided - A boolean value indicating whether the buffer is single-sided or not.
  * @returns null - - No return value.
    */
  GEOSBufferParams_setSingleSided_r(handle: number, p: number, singleSided: number): null;
  /**
  * Returns the buffer computed for a geometry for a given buffer distance and parameters.
  * @param g - A pointer to a GEOSGeometry object representing the input geometry.
  * @param p - A pointer to a GEOSBufferParams object containing the buffer parameters.
  * @param width - The buffer distance.
  * @returns number - - A pointer to a GEOSGeometry object representing the output geometry, or NULL if an exception occurred.
    */
  GEOSBufferWithParams(g: number, p: number, width: number): number;
  /**
  * Creates a buffer area around this geometry having the given width and with a specified accuracy of approximation for circular arcs.
  * @param handle - The GEOS context pointer handle.
  * @param g - The geometry pointer to buffer.
  * @param p - A pointer to the buffer parameters.
  * @param width - The width of the buffer area.
  * @returns number - A pointer to the geometry representing the buffer area. Returns NULL on exception.
    */
  GEOSBufferWithParams_r(handle: number, g: number, p: number, width: number): number;
  /**
  * Computes a Delaunay triangulation of the vertices of the given geometry.
  * @param g - The geometry pointer whose vertices to triangulate.
  * @param tolerance - The snapping tolerance to use for improved robustness. A tolerance of 0.0 specifies that no snapping will take place.
  * @param onlyEdges - If non-zero, only the edges of the triangulation will be returned, forming a multilinestring. If zero, the triangles will be returned as a collection of polygons.
  * @returns number - A pointer to the geometry representing the Delaunay triangulation. Returns NULL on exception.
    */
  GEOSDelaunayTriangulation(g: number, tolerance: number, onlyEdges: number): number;
  /**
  * Computes a Delaunay triangulation of the vertices of the given geometry using a GEOS context handle.
  * @param handle - The GEOS context pointer handle.
  * @param g - The geometry pointer whose vertices to triangulate.
  * @param tolerance - The snapping tolerance to use for improved robustness. A tolerance of 0.0 specifies that no snapping will take place.
  * @param onlyEdges - If non-zero, only the edges of the triangulation will be returned, forming a multilinestring. If zero, the triangles will be returned as a collection of polygons.
  * @returns number - A pointer to the geometry representing the Delaunay triangulation. Returns NULL on exception.
    */
  GEOSDelaunayTriangulation_r(handle: number, g: number, tolerance: number, onlyEdges: number): number;
  /**
  * Computes a Voronoi diagram from the vertices of the given geometry.
  * @param g - The geometry pointer whose vertices to use as sites for Voronoi diagram generation.
  * @param env - A pointer to an optional clipping envelope (GEOSGeometry). If provided, only cells which lie in this envelope will be included in the result. This can be used to avoid constructing very large cells (such as those which are far away from all sites). If NULL, no clipping envelope is used.
  * @param tolerance - The snapping tolerance to use for improved robustness. A tolerance of 0.0 specifies that no snapping will take place.
  * @param onlyEdges - If non-zero, only the edges of the Voronoi diagram will be returned, forming a multilinestring. If zero, the cells will be returned as a collection of polygons.
  * @returns number - A pointer to the geometry representing the Voronoi diagram. Returns NULL on exception.
    */
  GEOSVoronoiDiagram(g: number, env: number, tolerance: number, onlyEdges: number): number;
  /**
  * Computes a Voronoi diagram from the vertices of the given geometry using a GEOS context handle.
  * @param handle - The GEOS context pointer handle.
  * @param g - The geometry pointer whose vertices to use as sites for Voronoi diagram generation.
  * @param env - A pointer to an optional clipping envelope (GEOSGeometry). If provided, only cells which lie in this envelope will be included in the result. This can be used to avoid constructing very large cells (such as those which are far away from all sites). If NULL, no clipping envelope is used.
  * @param tolerance - The snapping tolerance to use for improved robustness. A tolerance of 0.0 specifies that no snapping will take place.
  * @param onlyEdges - If non-zero, only the edges of the Voronoi diagram will be returned, forming a multilinestring. If zero, the cells will be returned as a collection of polygons.
  * @returns number - A pointer to the geometry representing the Voronoi diagram. Returns NULL on exception.
    */
  GEOSVoronoiDiagram_r(handle: number, g: number, env: number, tolerance: number, onlyEdges: number): number;
  /**
  * Computes the intersection point of two line segments, if there is one.
  * @param p0x - The x-coordinate of the first endpoint of the first segment.
  * @param p0y - The y-coordinate of the first endpoint of the first segment.
  * @param p1x - The x-coordinate of the second endpoint of the first segment.
  * @param p1y - The y-coordinate of the second endpoint of the first segment.
  * @param q0x - The x-coordinate of the first endpoint of the second segment.
  * @param q0y - The y-coordinate of the first endpoint of the second segment.
  * @param q1x - The x-coordinate of the second endpoint of the second segment.
  * @param q1y - The y-coordinate of the second endpoint of the second segment.
  * @param rx - A pointer to a double where the x-coordinate of the intersection point will be stored, if any.
  * @param ry - A pointer to a double where the y-coordinate of the intersection point will be stored, if any.
  * @returns number - 1 if an intersection point is found, 0 if no intersection point is found, or -1 on error.
    */
  GEOSSegmentIntersection(p0x: number, p0y: number, p1x: number, p1y: number, q0x: number, q0y: number, q1x: number, q1y: number, rx: number, ry: number): number;
  /**
  * Computes the intersection point of two line segments, if there is one. This is a reentrant version that takes a GEOS context handle as an argument.
  * @param handle - A pointer to the GEOS context handle.
  * @param p0x - The x-coordinate of the first endpoint of the first segment.
  * @param p0y - The y-coordinate of the first endpoint of the first segment.
  * @param p1x - The x-coordinate of the second endpoint of the first segment.
  * @param p1y - The y-coordinate of the second endpoint of the first segment.
  * @param q0x - The x-coordinate of the first endpoint of the second segment.
  * @param q0y - The y-coordinate of the first endpoint of the second segment.
  * @param q1x - The x-coordinate of the second endpoint of the second segment.
  * @param q1y - The y-coordinate of the second endpoint of the second segment.
  * @param rx - A pointer to a double where the x-coordinate of the intersection point will be stored, if any.
  * @param ry - A pointer to a double where the y-coordinate of the intersection point will be stored, if any.
  * @returns number - 1 if an intersection point is found, 0 if no intersection point is found, or -1 on error.
    */
  GEOSSegmentIntersection_r(handle: number, p0x: number, p0y: number, p1x: number, p1y: number, q0x: number, q0y: number, q1x: number, q1y: number, rx: number, ry: number): number;
  /**
  * Creates and returns a new GEOS context handle. This function must be called before any other GEOS function in order to initialize thread-local storage and error handlers. Each thread must have its own GEOS context handle.
  * @returns number - A GEOS context handle or null on error.
    */
  GEOS_init_r(): number;
  /**
  * Sets a notice message handler for a given GEOS context. A notice message handler is a function that takes a string argument and does something with it, such as printing it to stdout or logging it somewhere. Notice messages are informative messages that do not indicate an error condition.
  * @param handle - A pointer to the GEOS context handle.
  * @param nf - A pointer to a notice message handler function.
  * @param userData - A pointer to user data that will be passed to the notice message handler function.
  * @returns void - undefined
    */
  GEOSContext_setNoticeMessageHandler_r(handle: number, nf: function, userData: any): void;
  /**
  * Sets an error message handler for a given GEOS context. An error message handler is a function that takes a string argument and does something with it, such as printing it to stderr or logging it somewhere. Error messages indicate that something went wrong during a GEOS operation.
  * @param handle - A pointer to the GEOS context handle.
  * @param ef - A pointer to an error message handler function.
  * @param userData - A pointer to user data that will be passed to the error message handler function.
  * @returns void - undefined
    */
  GEOSContext_setErrorMessageHandler_r(handle: number, ef: function, userData: any): void;
  /**
  * Finish the use of GEOS library
  * @param handle - The context returned by GEOS_init_r
  * @returns void - undefined
    */
  GEOS_finish_r(handle: number): void;
  /**
  * Return the version of GEOS as a string
  * @returns string - The version of GEOS
    */
  GEOSversion(): string;
  /**
  * Return the JTS port of GEOS as a string
  * @returns string - The JTS port of GEOS
    */
  GEOSjtsport(): string;
  /**
  * Set the X coordinate of a point in a CoordinateSequence
  * @param handle - The context returned by GEOS_init_r
  * @param s - The CoordinateSequence object
  * @param idx - The index of the point to set
  * @param val - The X coordinate value to set
  * @returns number - 1 on success, 0 on exception
    */
  GEOSCoordSeq_setX_r(handle: number, s: number, idx: number, val: number): number;
  /**
  * Set the Y coordinate of a point in a CoordinateSequence
  * @param handle - The context returned by GEOS_init_r
  * @param s - The CoordinateSequence object
  * @param idx - The index of the point to set
  * @param val - The Y coordinate value to set
  * @returns number - 1 on success, 0 on exception
    */
  GEOSCoordSeq_setY_r(handle: number, s: number, idx: number, val: number): number;
  /**
  * Sets the Z ordinate value for a given coordinate in a GEOSCoordSequence.
  * @param handle - The GEOS context pointer handle.
  * @param s - A pointer to the GEOSCoordSequence to modify.
  * @param idx - The coordinate index.
  * @param val - The Z ordinate value to set.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_setZ_r(handle: number, s: number, idx: number, val: number): number;
  /**
  * Gets the X ordinate value for a given coordinate in a GEOSCoordSequence.
  * @param handle - The GEOS context pointer handle.
  * @param s - A pointer to the GEOSCoordSequence to query.
  * @param idx - The coordinate index.
  * @param val - A pointer to a double to store the X ordinate value.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getX_r(handle: number, s: number, idx: number, val: number): number;
  /**
  * Gets the Y ordinate value for a given coordinate in a GEOSCoordSequence.
  * @param handle - The GEOS context pointer handle.
  * @param s - A pointer to the GEOSCoordSequence to query.
  * @param idx - The coordinate index.
  * @param val - A pointer to a double to store the Y ordinate value.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getY_r(handle: number, s: number, idx: number, val: number): number;
  /**
  * Gets the Z ordinate value for a given coordinate in a GEOSCoordSequence. This is an alias for GEOSCoordSeq_getZ_r without the context handle parameter.
  * @param s - A pointer to the GEOSCoordSequence to query.
  * @param idx - The coordinate index.
  * @param val - A pointer to a double to store the Z ordinate value.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getZ(s: number, idx: number, val: number): number;
  /**
  * Gets the Z ordinate value for a given coordinate in a GEOSCoordSequence. This is equivalent to GEOSCoordSeq_getZ_ with an additional context handle parameter.
  * @param handle - The GEOS context pointer handle.
  * @param s - A pointer to the GEOSCoordSequence to query.
  * @param idx - The coordinate index.
  * @param val - A pointer to a double to store the Z ordinate value.
  * @returns number - 1 on success, 0 on exception.
    */
  GEOSCoordSeq_getZ_r(handle: number, s: number, idx: number, val: number): number;
  /**
  * Computes the Constrained Delaunay Triangulation of a polygon geometry.
  * @param g - A pointer to a GEOS geometry object representing the input polygon.
  * @returns number - A pointer to a GEOS geometry object representing the triangulation, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a0f6b9f4c2e8a3b1f7c2a5d0e3c6a7d9f
  */
  GEOSConstrainedDelaunayTriangulation(g: number): number;
  /**
  * Computes the Constrained Delaunay Triangulation of a polygon.
  * @param handle - A pointer to a GEOS context handle.
  * @param g - A pointer to a GEOS geometry object representing the polygon.
  * @returns number - A pointer to a GEOS geometry object representing the triangulation, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a4a9f3c0bfbf1f9c6b1a7b0e7a8d4d5c6
  */
  GEOSConstrainedDelaunayTriangulation_r(handle: number, g: number): number;
  /**
  * Copies the coordinate values from the given arrays to a coordinate sequence.
  * @param x - A pointer to an array of x values.
  * @param y - A pointer to an array of y values.
  * @param z - A pointer to an array of z values, or NULL if not needed.
  * @param m - A pointer to an array of m values, or NULL if not needed.
  * @param size - The number of coordinates in the arrays.
  * @returns number - A pointer to a GEOS coordinate sequence object, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a4f7a1f9d0c2b0c119e6f9e2f60b112a9
  */
  GEOSCoordSeq_copyFromArrays(x: number, y: number, z: number, m: number, size: number): number;
  /**
  * Copies the ordinates of a coordinate sequence from arrays of x, y and optionally z values.
  * @param ctx - A GEOS context handle.
  * @param x - A pointer to an array of x values.
  * @param y - A pointer to an array of y values.
  * @param z - A pointer to an array of z values, or NULL if not needed.
  * @param m - A pointer to an array of m values, or NULL if not needed.
  * @param size - The number of coordinates in the arrays.
  * @returns number - A pointer to a GEOS coordinate sequence object, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9f4c6a5a7f2b9c0b1f6c0e1d0a1e4f5a
  */
  GEOSCoordSeq_copyFromArrays_r(ctx: number, x: number, y: number, z: number, m: number, size: number): number;
  /**
  * Copies a coordinate sequence from a buffer of x,y(,z) values.
  * @param buf - A pointer to a buffer of double values, in x, y order.
  * @param size - The number of coordinates in the sequence.
  * @param hasZ - Does buffer have Z values?
  * @param hasM - Does buffer have M values?
  * @returns number - A pointer to a GEOS coordinate sequence object, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a6f4b0a3f8e9c7a9b5f3b2c1f0c3a1d6f
  */
  GEOSCoordSeq_copyFromBuffer(buf: number, size: number, hasZ: number, hasM: number): number;
  /**
  * Copies a coordinate sequence from a buffer of x,y(,z) values.
  * @param ctx - A GEOS context handle.
  * @param buf - A pointer to a buffer of double values, in x, y order.
  * @param size - The number of coordinates in the sequence.
  * @param hasZ - Does buffer have Z values?
  * @param hasM - Does buffer have M values?
  * @returns number - A pointer to a GEOS coordinate sequence object, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a7a9f0b9c5a4f3b8f1d6b2a0c6e8d7c7f
  */
  GEOSCoordSeq_copyFromBuffer_r(ctx: number, buf: number, size: number, hasZ: number, hasM: number): number;
  /**
  * Copies the coordinates of a coordinate sequence to two arrays.
  * @param cs - A pointer to a GEOS coordinate sequence object.
  * @param x - A pointer to an array of doubles to store the x-coordinates.
  * @param y - A pointer to an array of doubles to store the y-coordinates.
  * @param z - A pointer to an array of doubles to store the z-coordinates, or NULL.
  * @param m - A pointer to an array of doubles to store the m-coordinates, or NULL.
  * @returns number - 1 on success, 0 on failure.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3f7a4b9a6d0e2b7c8f6f9e5b1f6d0a5c
  */
  GEOSCoordSeq_copyToArrays(cs: number, x: number, y: number, z: number, m: number): number;
  /**
  * Copies the coordinates of a coordinate sequence to two arrays of doubles.
  * @param ctx - A GEOS context handle.
  * @param seq - A pointer to a GEOS coordinate sequence object.
  * @param x - A pointer to an array of doubles to store the x coordinates.
  * @param y - A pointer to an array of doubles to store the y coordinates.
  * @returns number - 1 on success, 0 on failure.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a5f7a6d4f3f9a9c5e2b7a0b8f6c4e3d0c
  */
  GEOSCoordSeq_copyToArrays_r(ctx: number, seq: number, x: number, y: number): number;
  /**
  * Copies the coordinates of a coordinate sequence to a user-supplied buffer.
  * @param cs - A pointer to a GEOS coordinate sequence object.
  * @param buf - A pointer to a buffer of double values, allocated by the caller.
  * @param stride - The number of doubles per coordinate in the buffer.
  * @returns number - 1 on success, 0 on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9a6f2c8b0a3f4e4f2c9a7b1d5c2e9b7a
  */
  GEOSCoordSeq_copyToBuffer(cs: number, buf: number, stride: number): number;
  /**
  * Copies the coordinates of a coordinate sequence to user-supplied buffers.
  * @param ctx - A GEOS context handle.
  * @param seq - A pointer to a GEOS coordinate sequence object.
  * @param x - A pointer to a buffer for the x coordinates.
  * @param y - A pointer to a buffer for the y coordinates.
  * @param z - A pointer to a buffer for the z coordinates (optional).
  * @returns number - 1 on success, 0 on failure.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9f4a9f6b7a5d8a6c0f2b5e7e6a3f0c7d
  */
  GEOSCoordSeq_copyToBuffer_r(ctx: number, seq: number, x: number, y: number, z: number): number;
  /**
  * Returns a geometry with vertices added to make the lines no longer than a given distance.
  * @param g - A pointer to a GEOS geometry object.
  * @param tolerance - The maximum distance between vertices in the output geometry.
  * @returns number - A pointer to a new GEOS geometry object, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a6f0b2f9f1a0c7e4b8a5f2d3c1a7e5c9b
  */
  GEOSDensify(g: number, tolerance: number): number;
  /**
  * Densifies a geometry by adding points along the segments so that the distance between any two consecutive points is less than or equal to a given tolerance.
  * @param ctx - A GEOS context handle.
  * @param geom - A pointer to a GEOS geometry object.
  * @param tolerance - A positive distance value.
  * @returns number - A pointer to a new GEOS geometry object representing the densified geometry, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a6a9f4b9a8d1f2c0f0b3e5f6b7a0c5a2c
  */
  GEOSDensify_r(ctx: number, geom: number, tolerance: number): number;
  /**
  * Tests whether two geometries are within a specified distance of each other.
  * @param g1 - A pointer to a GEOS geometry object.
  * @param g2 - A pointer to another GEOS geometry object.
  * @param distance - The distance threshold to test.
  * @returns number - 1 if the geometries are within the distance of each other, 0 if not, or -1 on error.
  * @see [GEOS C header file](https://libgeos.org/doxygen/geos__c_8h_source.html)
  */
  GEOSDistanceWithin(g1: number, g2: number, distance: number): number;
  /**
  * Checks if two geometries are within a given distance of each other.
  * @param handle - A pointer to a GEOS context handle.
  * @param g1 - A pointer to the first GEOS geometry object.
  * @param g2 - A pointer to the second GEOS geometry object.
  * @param distance - The distance threshold in units of the coordinate reference system.
  * @returns number - A char value that is 1 if the geometries are within distance, 0 if not, or -1 on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a7a9b0f6a3d3f9f8a5d4e9e7b6f4a5c0d
  */
  GEOSDistanceWithin_r(handle: number, g1: number, g2: number, distance: number): number;
  /**
  * Creates a GeoJSON reader object.
  * @returns number - A pointer to a GEOSGeoJSONReader object, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a6a9f3f0b6c7a4a1d1f5c0e2b2a8f9b4d
  */
  GEOSGeoJSONReader_create(): number;
  /**
  * Creates a GeoJSON reader object.
  * @param ctx - A pointer to a GEOS context handle.
  * @returns number - A pointer to a GEOS GeoJSON reader object, or NULL on error.
  * @see [GEOS C header file](https://libgeos.org/doxygen/geos__c_8h_source.html)
  */
  GEOSGeoJSONReader_create_r(ctx: number): number;
  /**
  * Destroys a GeoJSON reader object and releases any allocated resources.
  * @param reader - A pointer to a GEOSGeoJSONReader object.
  * @returns void - undefined
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a0f7a9a6c0b4f1e5b7f2c3a1d3e4d8a9f
  */
  GEOSGeoJSONReader_destroy(reader: number): void;
  /**
  * Destroys a GEOSGeoJSONReader object.
  * @param ctx - A pointer to a GEOS context handle.
  * @param reader - A pointer to a GEOSGeoJSONReader object.
  * @returns void - undefined
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a7f9a4b0e2c9b3a1f1d7a9c6d3f0b5e8a
  */
  GEOSGeoJSONReader_destroy_r(ctx: number, reader: number): void;
  /**
  * Reads a geometry from a GeoJSON string.
  * @param reader - A pointer to a GEOSGeoJSONReader object.
  * @param geojson - A string containing a valid GeoJSON geometry object.
  * @returns number - A pointer to a GEOS geometry object representing the GeoJSON geometry, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a0f2a9b4a3b2b6d1c7f0a5e7f1c8f9d5a
  */
  GEOSGeoJSONReader_readGeometry(reader: number, geojson: string): number;
  /**
  * Reads a geometry from a GeoJSON string.
  * @param ctx - A pointer to a GEOS context handle.
  * @param reader - A pointer to a GEOSGeoJSONReader object.
  * @param geojson - A GeoJSON string representing a geometry.
  * @returns number - A pointer to a GEOS geometry object, or NULL on error or empty input.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9f6a1c0e8f5a3b4c7b3d0f7f2c9a4e8f
  */
  GEOSGeoJSONReader_readGeometry_r(ctx: number, reader: number, geojson: string): number;
  /**
  * Creates a GeoJSON writer object.
  * @returns number - A pointer to a GEOSGeoJSONWriter object, or NULL on error.
  * @see [GEOS C header file](https://libgeos.org/doxygen/geos__c_8h_source.html)
  */
  GEOSGeoJSONWriter_create(): number;
  /**
  * Creates a GeoJSON writer object.
  * @param handle - A pointer to a GEOS context handle.
  * @returns number - A pointer to a GEOS GeoJSON writer object, or NULL on error.
  * @see [GEOS C header file](https://libgeos.org/doxygen/geos__c_8h_source.html)
  */
  GEOSGeoJSONWriter_create_r(handle: number): number;
  /**
  * Destroys a GeoJSON writer object and releases any allocated resources.
  * @param writer - A pointer to a GEOS GeoJSON writer object.
  * @returns void - undefined
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a2f0a7c9f0a4a6b3e1f5d6d8c7a3e1b4c
  */
  GEOSGeoJSONWriter_destroy(writer: number): void;
  /**
  * Destroys a GEOSGeoJSONWriter object and releases the memory associated with it.
  * @param ctx - A pointer to a GEOS context handle.
  * @param writer - A pointer to a GEOSGeoJSONWriter object.
  * @returns void - No return value.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3f0f5c2b0a6d1a9e7f9a4b6b1c4f3d5a
  */
  GEOSGeoJSONWriter_destroy_r(ctx: number, writer: number): void;
  /**
  * Writes a geometry object to a GeoJSON string.
  * @param writer - A pointer to a GEOSGeoJSONWriter object.
  * @param g - A pointer to a GEOS geometry object.
  * @returns string - A GeoJSON string representation of the geometry, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a0a3f1f6b2a1f7d0b5a9c4d2c6b7e9e0a
  */
  GEOSGeoJSONWriter_writeGeometry(writer: number, g: number): string;
  /**
  * Writes a geometry to a GeoJSON string.
  * @param ctx - A GEOS context handle.
  * @param writer - A pointer to a GEOSGeoJSONWriter object.
  * @param geom - A pointer to a GEOS geometry object.
  * @returns string - A GeoJSON string representation of the geometry, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9a6f9f0b3a4c1a2e7c1d0e4e1b3a2c5d
  */
  GEOSGeoJSONWriter_writeGeometry_r(ctx: number, writer: number, geom: number): string;
  /**
  * Creates a GEOSMakeValidParams object, which can be used to control the behavior of the GEOSMakeValid function.
  * @returns number - A pointer to a GEOSMakeValidParams object, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9a1f0e0b8a3f7d2b9c6f3f4d5a0b1c8e
  */
  GEOSMakeValidParams_create(): number;
  /**
  * Creates a GEOSMakeValidParams object, which can be used to control the behaviour of the GEOSMakeValid function.
  * @param ctx - A pointer to a GEOS context handle.
  * @returns number - A pointer to a GEOSMakeValidParams object, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a5f0c2a4a7b6f9a1e9b3f0a0c8d9d0f1d
  */
  GEOSMakeValidParams_create_r(ctx: number): number;
  /**
  * Destroys a GEOSMakeValidParams object.
  * @param params - A pointer to a GEOSMakeValidParams object.
  * @returns void - undefined
  * @see [12](https://github.com/libgeos/geos)
  */
  GEOSMakeValidParams_destroy(params: number): void;
  /**
  * Destroys a GEOSMakeValidParams object.
  * @param ctx - A pointer to a GEOS context handle.
  * @param params - A pointer to a GEOSMakeValidParams object.
  * @returns void - undefined
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a0c9f9a6f5c5e3b1f4b4a0d7f8b7c2a3c
  */
  GEOSMakeValidParams_destroy_r(ctx: number, params: number): void;
  /**
  * Attempts to make an invalid geometry valid without losing any of the input vertices.
  * @param ctx - A GEOS context handle.
  * @param params - A pointer to a GEOSMakeValidParams object.
  * @param g - A pointer to a GEOS geometry object.
  * @returns number - A pointer to a new GEOS geometry object representing the valid geometry, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a4f1a0c3f6f0a7b9b9e2d9c2a6f7b0e8e
  */
  GEOSMakeValidParams_setKeepCollapsed_r(ctx: number, params: number, g: number): number;
  /**
  * Sets the method used to make a geometry valid.
  * @param params - A pointer to a GEOSMakeValidParams object.
  * @param method - An integer representing the method, such as 1 for MAKE_VALID_STRUCTURE_PRESERVING, 2 for MAKE_VALID_BUFFERED, etc.
  * @returns number - A boolean value indicating success or failure.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a7a0f4e0c3b6a9f9e1d5c7b9a1d6f0c4e
  */
  GEOSMakeValidParams_setMethod(params: number, method: number): number;
  /**
  * Sets the method to use for making a geometry valid.
  * @param ctx - A pointer to a GEOS context handle.
  * @param params - A pointer to a GEOSMakeValidParams object.
  * @param method - An integer representing the method, such as 1 for MAKE_VALID_STRUCTURE_PRESERVING, 2 for MAKE_VALID_BUFFERED, etc.
  * @returns number - A boolean value indicating success or failure.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9b0f7a7a3c5f0f8e1d4b6a9c3c3f1e9b
  */
  GEOSMakeValidParams_setMethod_r(ctx: number, params: number, method: number): number;
  /**
  * Attempts to make an invalid geometry valid without losing any of the input vertices.
  * @param g - A pointer to a GEOS geometry object.
  * @param params - A pointer to a GEOSMakeValidParams object that specifies the options for the operation.
  * @returns number - A pointer to a new GEOS geometry object representing the valid geometry, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3a1c0b9f6a5f7a4e0b7c9e9b1f3d8f7c
  */
  GEOSMakeValidWithParams(g: number, params: number): number;
  /**
  * Attempts to make an invalid geometry valid without losing any of the input vertices.
  * @param ctx - A GEOS context handle.
  * @param params - A pointer to a GEOSMakeValidParams object.
  * @param g - A pointer to a GEOS geometry object.
  * @returns number - A pointer to a new GEOS geometry object representing the valid geometry, or NULL on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a6f0e9a3b8f2c5b1b4a0d3c7a7f6f5a9d
  */
  GEOSMakeValidWithParams_r(ctx: number, params: number, g: number): number;
  /**
  * Tests whether the distance between a prepared geometry and another geometry is within a specified distance.
  * @param prep - A pointer to a GEOS prepared geometry object.
  * @param g2 - A pointer to a GEOS geometry object.
  * @param dist - A double value representing the distance threshold.
  * @returns number - 1 if the geometries are within the distance, 0 if not, or -1 on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a6f5a0b7f6f9e3a9c4b0b8a3c1a4d4e6d
  */
  GEOSPreparedDistanceWithin(prep: number, g2: number, dist: number): number;
  /**
  * Computes the distance between a prepared geometry and another geometry, and checks if it is within a given tolerance.
  * @param ctx - A GEOS context handle.
  * @param prep - A pointer to a prepared GEOS geometry object.
  * @param g - A pointer to a GEOS geometry object.
  * @param tolerance - A non-negative distance value.
  * @returns number - 1 if the geometries are within the given distance of each other, 0 if not, or -1 on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3a9f7c0b6f9b2c6a0e3f4f5a7d8b9e5d
  */
  GEOSPreparedDistanceWithin_r(ctx: number, prep: number, g: number, tolerance: number): number;
  /**
  * Returns the output format of a WKB writer.
  * @param writer - A pointer to a GEOS WKB writer object.
  * @returns number - An integer representing the output format, such as 0 for WKB_ISO, 1 for WKB_SFSQL, or 2 for WKB_HEX.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a3f7a4c9b8b4f8e2f6d3a0e5c5b7b0a6f
  */
  GEOSWKBWriter_getFlavor(writer: number): number;
  /**
  * Returns the output format of a WKB writer.
  * @param handle - A GEOS context handle.
  * @param writer - A pointer to a GEOS WKB writer object.
  * @returns number - An integer representing the output format, such as 0 for WKB_ISO, 1 for WKB_SFSQL, or 2 for WKB_HEX. Returns -1 on error.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a0a7f9c4a6f5b8e0f3c8b9a2d1b6b4e3a
  */
  GEOSWKBWriter_getFlavor_r(handle: number, writer: number): number;
  /**
  * Sets the output format of a WKB writer to either WKB or EWKB.
  * @param writer - A pointer to a GEOS WKB writer object.
  * @param flavor - An integer representing the output format, 1 for WKB, 2 for EWKB.
  * @returns void - undefined
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9a0e1f9fa62b0d657c4b18e01a276a94
  */
  GEOSWKBWriter_setFlavor(writer: number, flavor: number): void;
  /**
  * Sets the output format of a WKB writer to either WKB or EWKB.
  * @param handle - A GEOS context handle.
  * @param writer - A pointer to a GEOS WKB writer object.
  * @param flavor - An integer representing the output format, 1 for WKB, 2 for EWKB.
  * @returns number - 1 on success, 0 on failure.
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a7f9a6d3a5f2b0f1a4c6b4c2e3b9b8a9e
  */
  GEOSWKBWriter_setFlavor_r(handle: number, writer: number, flavor: number): number;
  /**
  * Geometry type number, used by functions returning or
* consuming geometry types.
* 
* | Name             | Value     |
* | :--------------- | :----------  |
* | GEOS_POINT   | 0         |
* | GEOS_LINESTRING   | 1         |
* | GEOS_LINEARRING   | 2         |
* | GEOS_POLYGON   | 3         |
* | GEOS_MULTIPOINT   | 4         |
* | GEOS_MULTILINESTRING   | 5         |
* | GEOS_MULTIPOLYGON   | 6         |
* | GEOS_GEOMETRYCOLLECTION   | 7         |
  * @returns void - undefined
  * @see GEOSGeomType
  */
  GEOSGeomTypes(): void;
  /**
  * Well-known binary byte orders used when
* writing to WKB.
* 
* | Name             | Value     |
* | :--------------- | :----------  |
* | GEOS_WKB_XDR   | 0         |
* | GEOS_WKB_NDR   | 1         |
  * @returns void - undefined
  * @see GEOSWKBWriter_setByteOrder
  */
  GEOSWKBByteOrders(): void;
  /**
  * Well-known binary flavors to use
* when writing to WKB. ISO flavour is
* more standard. Extended flavour supports
* 3D and SRID embedding. GEOS reads both
* transparently.
* 
* | Name             | Value     |
* | :--------------- | :----------  |
* | GEOS_WKB_ISO   | 1         |
* | GEOS_WKB_EXTENDED   | 2         |
  * @returns void - undefined
  * @see GEOSWKBWriter_setFlavor
  */
  GEOSWKBFlavors(): void;
  /**
  * Cap styles control the ends of buffered lines.
* 
* | Name             | Value     |
* | :--------------- | :----------  |
* | GEOSBUF_CAP_ROUND   | 1         |
* | GEOSBUF_CAP_FLAT   | 2         |
* | GEOSBUF_CAP_SQUARE   | 3         |
  * @returns void - undefined
  * @see GEOSBuffer
  */
  GEOSBufCapStyles(): void;
  /**
  * Join styles control the buffer shape at bends in a line.
* 
* | Name             | Value     |
* | :--------------- | :----------  |
* | GEOSBUF_JOIN_ROUND   | 1         |
* | GEOSBUF_JOIN_MITRE   | 2         |
* | GEOSBUF_JOIN_BEVEL   | 3         |
  * @returns void - undefined
  * @see GEOSBuffer
  */
  GEOSBufJoinStyles(): void;
  /**
  * Controls the behavior of the result of GEOSRelate when returning
* DE9IM results for two geometries.
* 
* | Name             | Value     |
* | :--------------- | :----------  |
* | GEOSRELATE_BNR_MOD2   | 1         |
* | GEOSRELATE_BNR_OGC   | 1         |
* | GEOSRELATE_BNR_ENDPOINT   | 2         |
* | GEOSRELATE_BNR_MULTIVALENT_ENDPOINT   | 3         |
* | GEOSRELATE_BNR_MONOVALENT_ENDPOINT   | 4         |
  * @returns void - undefined
    */
  GEOSRelateBoundaryNodeRules(): void;
  /**
  * Change behaviour of validity testing in \ref GEOSisValidDetail
* 
* | Name             | Value     |
* | :--------------- | :----------  |
* | GEOSVALID_ALLOW_SELFTOUCHING_RING_FORMING_HOLE   | 1         |
  * @returns void - undefined
    */
  GEOSValidFlags(): void;
  /**
  * Controls the behavior of GEOSGeom_setPrecision()
* when altering the precision of a geometry.
* 
* | Name             | Value     | Description |
* | :--------------- | :----------  | :----------  |
* | GEOS_PREC_VALID_OUTPUT   | 0         | The output is always valid. Collapsed geometry elements (including both polygons and lines) are removed.|
* | GEOS_PREC_NO_TOPO   | 1         | Precision reduction is performed pointwise. Output geometry may be invalid due to collapse or self-intersection. (This might be better called "GEOS_PREC_POINTWISE" - the current name is historical.) |
* | GEOS_PREC_KEEP_COLLAPSED   | 2         | Like the default mode, except that collapsed linear geometry elements are preserved. Collapsed polygonal input elements are removed. |
  * @returns void - undefined
    */
  GEOSPrecisionRules(): void;
  /**
  * Controls the behavior of the GEOSPolygonHullSimplify parameter.
* 
* | Name             | Value     | Description |
* | :--------------- | :----------  | :----------  |
* | GEOSHULL_PARAM_VERTEX_RATIO   | 1         | Fraction of input vertices retained |
* | GEOSHULL_PARAM_AREA_RATIO   | 2         | Ratio of simplified hull area to input area |
  * @returns void - undefined
    */
  GEOSPolygonHullParameterModes(): void;
  /**
  * Change behaviour of @ref GEOSVoronoiDiagram
* 
* | Name             | Value     | Description |
* | :--------------- | :----------  | :----------  |
* | GEOS_VORONOI_ONLY_EDGES   | 1         | Return only edges of the Voronoi cells, as a MultiLineString |
* | GEOS_VORONOI_PRESERVE_ORDER   | 2         | Preserve order of inputs, such that the nth cell in the result corresponds to the nth vertex in the input. If this cannot be done, such as for inputs that contain repeated points, @ref GEOSVoronoiDiagram will return NULL. |
  * @returns void - undefined
    */
  GEOSVoronoiFlags(): void;
  /**
  * Cleans up GEOS and releases any allocated resources.
  * @returns void - undefined
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9a6f9f0b3a4c1a2e7c1d0e4e1b3a2c5d
  */
  finishGEOS(): void;
  /**
  * Initializes GEOS and allocates any required resources.
  * @returns void - undefined
  * @see https://libgeos.org/doxygen/geos__c_8h.html#a9a6f9f0b3a4c1a2e7c1d0e4e1b3a2c5d
  */
  initGEOS(): void;
  Module: Module;
}

interface Config {
  /** 
    * If autoInit is true, the GEOS library will be initialized when the geos object is created.
    * @default true
  */ 
  autoInit?: boolean;
  errorHandler?: (message: string) => void;
  noticeHandler?: (message: string) => void;
}

declare module 'geos-wasm' {
  export default function initGeosJs(config?: Config): Promise<geos>;
}